/*
 * linked list routines
 * by Matthew Luckie
 *
 * Copyright (C) 2004 Matthew Luckie. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Matthew Luckie ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Matthew Luckie BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: mjl_list.c,v 1.15 2005/03/03 01:20:44 mjl Exp $
 *
 */

#include <stdlib.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "mjl_list.h"

struct slist_node
{
  void              *item;
  struct slist_node *next;
};

struct dlist_node
{
  void              *item;
  struct dlist_node *prev;
  struct dlist_node *next;
};

struct clist_node
{
  void              *item;
  struct clist_node *prev;
  struct clist_node *next;
};

struct slist
{
  slist_node_t *head;
  slist_node_t *tail;
  int           length;
};

struct dlist
{
  dlist_node_t *head;
  dlist_node_t *tail;
  int           length;
};

struct clist
{
  clist_node_t *head;
  clist_node_t *tail;
  int           length;
};

slist_t *slist_alloc()
{
  slist_t *list;

  if((list = (slist_t *)malloc(sizeof(slist_t))) != NULL)
    {
      list->head   = NULL;
      list->tail   = NULL;
      list->length = 0;
    }

  return list;
}

static slist_node_t *slist_node_alloc(void *item, slist_node_t *next)
{
  slist_node_t *node;

  if((node = malloc(sizeof(slist_node_t))) != NULL)
    {
      node->item = item;
      node->next = next;
    }

  return node;
}

/*
 * slist_dup
 *
 * make a copy of the list that points to the same items.
 */
slist_t *slist_dup(slist_t *oldlist)
{
  slist_t      *list = (slist_t *)malloc(sizeof(slist_t));
  slist_node_t *oldnode, *node;

  if((list = (slist_t *)malloc(sizeof(slist_t))) != NULL)
    {
      list->head = NULL;
      list->tail = NULL;
      list->length = 0;
    }
  else return NULL;

  if(oldlist->head != NULL)
    {
      if((node = slist_node_alloc(oldlist->head->item, NULL)) == NULL)
	{
	  goto err;
	}

      list->length = oldlist->length;
      list->head = node;
      oldnode = oldlist->head->next;
    }
  else return list;

  while(oldnode != NULL)
    {
      if((node->next = slist_node_alloc(oldnode->item, NULL)) == NULL)
	{
	  goto err;
	}

      oldnode = oldnode->next;
      node = node->next;
    }

  list->tail = node;

  return list;

 err:
  slist_free(list);
  return NULL;
}

void slist_concat(slist_t *first, slist_t *second)
{
  if(first->tail != NULL)
    {
      first->tail->next = second->head;
      first->length += second->length;
      first->tail = second->tail;
    }
  else
    {
      first->head = second->head;
      first->tail = second->tail;
      first->length = second->length;
    }
  free(second);
  return;    
}

void slist_free(slist_t *list)
{
  slist_node_t *node = list->head;
  slist_node_t *next;

  while(node != NULL)
    {
      next = node->next;
      free(node);
      node = next;
    }

  free(list);

  return;
}

void *slist_head_push(slist_t *list, void *item)
{
  slist_node_t *node;

  if((node = slist_node_alloc(item, list->head)) != NULL)
    {
      if(list->head == NULL)
	{
	  list->head = list->tail = node;
	}

      list->length++;
    }

  return node;
}

void *slist_tail_push(slist_t *list, void *item)
{
  slist_node_t *node;

  if((node = slist_node_alloc(item, NULL)) != NULL)
    {
      if(list->tail != NULL)
	{
	  list->tail->next = node;
	  list->tail = node;
	}
      else
	{
	  list->head = list->tail = node;
	}

      list->length++;
    }

  return node;
}

void *slist_head_pop(slist_t *list)
{
  slist_node_t *node;
  void         *item;

  if(list->head == NULL) return NULL;

  node = list->head;
  item = node->item;

  list->head = node->next;

  if(list->tail == node)
    {
      list->tail = NULL;
    }

  free(node);

  list->length--;

  return item;
}

void *slist_head_get(const slist_t *list)
{
  if(list->head == NULL) return NULL;
  return list->head->item;
}

slist_node_t *slist_head_node(const slist_t *list)
{
  return list->head;
}

void *slist_node_item(const slist_node_t *node)
{
  return node->item;
}

slist_node_t *slist_node_next(const slist_node_t *node)
{
  return node->next;
}

int slist_foreach(const slist_t *list, const void *param,
		  const slist_foreach_t func)
{
  slist_node_t *node = list->head, *next;

  while(node != NULL)
    {
      next = node->next;
      if(func(param, node->item) == -1) return -1;
      node = next;
    }

  return 0;
}

int slist_count(const slist_t *list)
{
  return list->length;
}

dlist_t *dlist_alloc()
{
  dlist_t *list = (dlist_t *)malloc(sizeof(dlist_t));
  list->head   = NULL;
  list->tail   = NULL;
  list->length = 0;

  return list;
}

void dlist_free(dlist_t *list)
{
  dlist_node_t *node = list->head;
  dlist_node_t *next;

  while(node != NULL)
    {
      next = node->next;
      free(node);
      node = next;
    }

  free(list);

  return;
}

void *dlist_head_push(dlist_t *list, void *item)
{
  dlist_node_t *node;

  if((node = (dlist_node_t *)malloc(sizeof(dlist_node_t))) == NULL)
    {
      return NULL;
    }

  node->item = item;
  node->next = list->head;
  node->prev = NULL;

  /* if we don't have a head node, we don't have a tail node set either */
  if(list->head == NULL)
    {
      list->tail = node;
    }
  else
    {
      list->head->prev = node;
    }

  list->head = node;
  list->length++;

  return node;
}

void *dlist_tail_push(dlist_t *list, void *item)
{
  dlist_node_t *node;

  if((node = (dlist_node_t *)malloc(sizeof(dlist_node_t))) == NULL)
    {
      return NULL;
    }

  node->item = item;
  node->next = NULL;
  node->prev = list->tail;

  /* if we don't have a tail node, we don't have a head node set either */
  if(list->tail == NULL)
    {
      list->head = node;
    }
  else
    {
      list->tail->next = node;
    }

  list->tail = node;
  list->length++;

  return node;
}

void *dlist_head_pop(dlist_t *list)
{
  dlist_node_t *node;
  void         *item;

  if(list->head == NULL) return NULL;

  node = list->head;
  item = node->item;

  /*
   * if we have a non-null node to replace the head with, null its prev
   * pointer as the node is now at the head of the list
   */
  if((list->head = node->next) != NULL)
    {
      list->head->prev = NULL;
    }

  if(list->tail == node)
    {
      list->tail = NULL;
    }

  free(node);

  list->length--;

  return item;
}

void *dlist_tail_pop(dlist_t *list)
{
  dlist_node_t *node;
  void         *item;

  if(list->head == NULL) return NULL;

  node = list->tail;
  item = node->item;

  list->tail = node->prev;

  if(list->tail != NULL)
    {
      list->tail->next = NULL;
    }

  if(list->head == node)
    {
      list->head = NULL;
    }

  free(node);

  list->length--;

  return item;
}

void *dlist_node_pop(dlist_t *list, dlist_node_t *node)
{
  void         *item;

  if(list->head == node)
    {
      list->head = node->next;
    }

  if(list->tail == node)
    {
      list->tail = node->prev;
    }

  if(node->prev != NULL) node->prev->next = node->next;
  if(node->next != NULL) node->next->prev = node->prev;
  item = node->item;

  free(node);

  list->length--;

  return item;
}

void *dlist_node_item(const dlist_node_t *node)
{
  return node->item;
}

dlist_node_t *dlist_node_next(const dlist_node_t *node)
{
  return node->next;
}

dlist_node_t *dlist_node_prev(const dlist_node_t *node)
{
  return node->prev;
}

void *dlist_head_get(const dlist_t *list)
{
  if(list->head == NULL) return NULL;
  return list->head->item;
}

dlist_node_t *dlist_head_node(const dlist_t *list)
{
  return list->head;
}

void *dlist_tail_get(const dlist_t *list)
{
  if(list->tail == NULL) return NULL;
  return list->tail->item;
}

int dlist_foreach(const dlist_t *list, const void *param,
		  const dlist_foreach_t func)
{
  dlist_node_t *node = list->head, *next;

  while(node != NULL)
    {
      next = node->next;
      if(func(param, node->item) == -1) return -1;
      node = next;
    }

  return 0;
}

int dlist_count(const dlist_t *list)
{
  return list->length;
}

clist_t *clist_alloc()
{
  clist_t *list = (clist_t *)malloc(sizeof(clist_t));
  list->head   = NULL;
  list->tail   = NULL;
  list->length = 0;

  return list;
}

void clist_free(clist_t *list)
{
  clist_node_t *node = list->head;
  clist_node_t *next;

  /* break the circle */
  if(list->tail != NULL)
    {
      list->tail->next = NULL;
    }

  /* delete all the nodes */
  while(node != NULL)
    {
      next = node->next;
      free(node);
      node = next;
    }

  free(list);
  return;
}

/*
 * clist_push
 *
 * push a new node into the circular list by putting it at the tail of
 * the list before it wraps back onto the head
 */
void *clist_push(clist_t *list, void *item)
{
  clist_node_t *node;

  if((node = (clist_node_t *)malloc(sizeof(clist_node_t))) == NULL)
    {
      return NULL;
    }

  node->item = item;
  node->next = list->head;
  node->prev = list->tail;

  /* if we don't have a tail node, we don't have a head node set either */
  if(list->tail == NULL)
    {
      list->head = node;
      node->prev = node->next = node;
    }
  else
    {
      list->tail->next = node;
      list->head->prev = node;
    }

  list->tail = node;
  list->length++;

  return node;  
}

/*
 * clist_pop
 *
 * take the tail node out of the circle list.
 */
void *clist_pop(clist_t *list)
{
  clist_node_t *node;
  void         *item;

  if(list->head == NULL) return NULL;

  node = list->tail;
  item = node->item;

  /* if there are still nodes on the clist, fix up the tail */
  if(node != node->prev)
    {
      list->tail->next = list->head;
      list->head->prev = list->tail;
    }
  else
    {
      list->tail = NULL;
      list->head = NULL;
    }

  free(node);

  list->length--;

  return item;
}

void *clist_node_pop(clist_t *list, clist_node_t *node)
{
  void         *item;

  item = node->item;

  if(node == node->prev)
    {
      list->head = NULL;
      list->tail = NULL;
    }
  else if(list->head == node)
    {
      list->head = node->next;
      list->tail->next = list->head;
    }
  else if(list->tail == node)
    {
      list->tail = node->prev;
      list->head->prev = list->tail;
    }
  else
    {
      node->prev->next = node->next;
      node->next->prev = node->prev;
    }

  free(node);

  list->length--;

  return item;

}

void *clist_node_item(const clist_node_t *node)
{
  return node->item;
}

clist_node_t *clist_node_next(const clist_node_t *node)
{
  return node->next;
}

int clist_foreach(const clist_t *list, const void *param,
		  const clist_foreach_t func)
{
  clist_node_t *node = list->head, *next;

  if(node == NULL)
    {
      return 0;
    }

  for(;;)
    {
      next = node->next;
      if(func(param, node->item) == -1) return -1;

      if(next != list->head)
	{
	  node = next;
	}
      else break;
    }

  return 0;
}

int clist_count(const clist_t *list)
{
  return list->length;
}
