/*
 * splay tree routines
 * By Matthew Luckie
 * U of Waikato 0657.317b 1999
 *
 * Copyright (C) 1999-2005 Matthew Luckie. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Matthew Luckie ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Matthew Luckie BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: mjl_splaytree.c,v 1.8 2005/05/26 21:27:02 mjl Exp $
 *
 */

#include <stdlib.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "mjl_splaytree.h"

/*
 * the splay tree algorithm needs a simple stack to do the work.
 * the implementations of these functions is found at the bottom of this
 * file.
 */
typedef struct splaytree_stack
{
  void    **items;
  int       i;
  int       c;
} splaytree_stack_t;

/*
 * splay tree node data structure
 * conveniently hidden from users of the splay tree.
 */
typedef struct splaytree_node
{
  void                  *ptr;
  struct splaytree_node *left;
  struct splaytree_node *right;
} splaytree_node_t;

struct splaytree
{
  splaytree_node_t  *head;
  int                size;
  splaytree_cmp_t    cmp;
  splaytree_stack_t *stack;
};

static splaytree_stack_t *stack_create();
static void               stack_destroy(splaytree_stack_t *s);
static int                stack_push(splaytree_stack_t *s, void *ptr);
static void              *stack_pop(splaytree_stack_t *s);
static void               stack_clean(splaytree_stack_t *s);

/*
Function: splaytree_rotate
Purpose:  To perform the generic treenode-rotate algorithm
Pre:      above and below must be valid treenode pointers.
          above must point to the below node
Post:     the below node is now further up the tree, and pointing to above
Returns:  the node that was rotated to the higher position
*/
static splaytree_node_t *splaytree_rotate(splaytree_node_t *above,
					  splaytree_node_t *below)
{
  splaytree_node_t *temp;
	
  /* check for valid pointers */
  if(above == NULL || below == NULL) return NULL;

  /*
   * check to see if the below node is to the left of the above or to
   * the right
   */
  if(above->left == below)
    {
      temp = below->right;
      below->right = above;
      above->left = temp;
    }
  else if(above->right == below)
    {
      temp = below->left;
      below->left = above;
      above->right = temp;
    }
  
  return below;
}

/*
Function:  splaytree_splay2
Purpose:   To appropriately splay the treenodes passed in so that the child
           is moved higher than the other nodes passed in
Pre:       grandparent points to parent, parent points to child
Post:      the grandparent, parent, and child nodes are appropriately splayed
Returns:   the child node, as it is now at the root of this subtree */
splaytree_node_t *splaytree_splay2(splaytree_node_t *child,
				   splaytree_node_t *parent,
				   splaytree_node_t *grandparent)
{
  if(child == NULL) return NULL;
  
  /* case 0: access node is root */
  if(parent == NULL) return child;

  /* case 1: parent is root */
  else if(grandparent == NULL)
    {
      splaytree_rotate(parent, child);
    }

  /*
   * case 2: zig zig - p is not the root and the child and the parent are both
   * left (right) children
   */
  else if((parent->left  == child && grandparent->left  == parent) || 
	  (parent->right == child && grandparent->right == parent))
    {
      splaytree_rotate(grandparent, parent);
      splaytree_rotate(parent, child);
    }

  /*
   * case 3: zig zag - p is not the root and the child is a left(right) child
   * and parent is a right(left) child
   */
  else if((parent->left  == child && grandparent->right == parent) ||
	  (parent->right == child && grandparent->left  == parent))
    {
      if(grandparent->left == parent)
	{
	  grandparent->left = splaytree_rotate(parent, child);
	  splaytree_rotate(grandparent, child);
	}
      else
	{
	  grandparent->right = splaytree_rotate(parent, child);
	  splaytree_rotate(grandparent, child);
	}
    }

  return child;
}

/*
Function:  splaytree_splay
Purpose:   To coordinate the calls to mjl_tree_splay
Pre:       The stack contains, in order, the path to the child so that the
           nodes can be splayed
Post:      The node at the tail of the stack is now the root of the tree,
           having been splayed upwards
Returns:   The new root of the tree
*/
static splaytree_node_t *splaytree_splay(splaytree_stack_t *stack)
{
  splaytree_node_t *child, *parent, *grandparent, *keep;

  child       = (splaytree_node_t *)stack_pop(stack);
  parent      = (splaytree_node_t *)stack_pop(stack);
  grandparent = (splaytree_node_t *)stack_pop(stack);

  /* sanity check */
  if(child == NULL) return NULL;

  /* any need to splay? (aka is the node at the root?) */
  if(parent == NULL) return child;

  /* splay the node */
  splaytree_splay2(child, parent, grandparent);

  /*
   * if it was a simple swap at the root, we can save the hassle and return
   * now
   */
  if(grandparent == NULL) return child;

  /*
   * remember the grandparent so that we can figure out where to relink the
   * splayed child to
   */
  keep = grandparent;

  /* just loop and we will break out when we need to */
  while(1)
    {
      /* get the parent nodes to the child */
      parent      = (splaytree_node_t *)stack_pop(stack);
      grandparent = (splaytree_node_t *)stack_pop(stack);

      /* 
       * if the child node is now at the root, break out as the splay is
       * complete
       */
      if(parent == NULL) break;

      /* 
       * figure out where to relink the child to
       * (as the grandparent in keep is now down the tree)
       */
      if(parent->left == keep)
	{
	  parent->left = child;
	}
      else if(parent->right == keep)
	{
	  parent->right = child;
	}
      else
	{
	  return NULL;
	}
      
      /* splay now */
      splaytree_splay2(child, parent, grandparent);
      if(grandparent == NULL) break;
      keep = grandparent;
    }
  
  /* return the new root of the tree */
  return child;
}

/*
 * splaytree_createstruct
 * creates/mallocs a node and initialises the contents of the node ready to
 * insert to the tree
 *
 */
static splaytree_node_t *splaytree_node_create(const void *ptr)
{
  splaytree_node_t *node;

  node = (splaytree_node_t *)malloc(sizeof(splaytree_node_t));
  if(node == NULL) return NULL;

  node->left    = NULL;
  node->right   = NULL;
  node->ptr     = (void *)ptr;

  return node;
}

/*
Function:  splaytree_insert2
Purpose:   To insert the data into the sorted tree
Pre:       the compare func address is valid, the stack data struct has been
           allocated for use
Post:      The stack contains the path this recursive function took to insert
           the node
Returns:   a value saying if the node was inserted or if it was found
*/
static int splaytree_insert2(splaytree_t *tree, const void *ptr,
			     splaytree_node_t *parent)
{
  splaytree_node_t *node;
  int               i;

  /*
   * see whether the data belongs to the left, right, or at this node
   * (ie a duplicate)
   */
  i = tree->cmp(parent->ptr, ptr);
  if(i < 0)
    {
      /*
       * if the data belongs here, create a node and push the parent to the
       * stack
       */
      if(parent->left == NULL)
	{
	  stack_push(tree->stack, parent);
	  node = parent->left = splaytree_node_create(ptr);
	  tree->size++;
	  i = 1;
	}
      /* put the node into the insert path and try the next level */
      else
	{
	  stack_push(tree->stack, parent);
	  return splaytree_insert2(tree, ptr, parent->left);
	}
    }
  else if(i > 0)
    {
      /*
       * if the data belongs here, create a node and push the parent to the
       * stack
       */
      if(parent->right == NULL)
	{
	  stack_push(tree->stack, parent);
	  node = parent->right = splaytree_node_create(ptr);
	  tree->size++;
	  i = 1;
	}
      /* put the node into the insert path and try the next level */
      else
	{
	  stack_push(tree->stack, parent);
	  return splaytree_insert2(tree, ptr, parent->right);
	}
    }
  else
    {
      /* the data already exists in the tree: do not add it */
      node = parent;
      i = 0;
    }
  
  /*
   * push the newly created/referenced node to the tail of the list and
   * return it 
   */
  stack_push(tree->stack, node);

  return i;
}

/*
 * splaytree_insert
 *
 * insert a value into the splay tree, and return with the tree splayed on
 * that value
 *
 * returns -1 on error.
 * returns 0 if the value was already in the tree
 * returns 1 if the value was successfully inserted
 */
int splaytree_insert(splaytree_t *tree, const void *ptr)
{
  int i;

  /*
   * if the tree actually has something in it, then we need to
   * find the place to insert the node and splay on that.
   */
  if(tree->head != NULL)
    {
      stack_clean(tree->stack);
      i = splaytree_insert2(tree, ptr, tree->head);

      if((tree->head = splaytree_splay(tree->stack)) == NULL)
	{
	  return -1;
	}

      return i;
    }

  tree->head = splaytree_node_create(ptr);
  if(tree->head == NULL) return -1;
  tree->size = 1;

  return 1;
}

/*
Function:  splaytree_find2
Purpose:   To find the data requested in the tree by the user
Returns:   The appropriate node that has been found
Pre:       the comparefunc is valid, and the stack data structure has been
           allocated
Post:      the stack contains the search path
*/
static splaytree_node_t *splaytree_find2(splaytree_t *tree, const void *ptr,
					 splaytree_node_t *tn)
{
  int i;

  /* sanity check */
  if(tn == NULL) return NULL;
	
  /* see whether to look to the left/right for the node */
  i = tree->cmp(tn->ptr, ptr);

  /*
   * try and push the node onto the stack.
   * if we don't then we can't splay the node to the top of the tree, so
   * we fail.
   */
  if(stack_push(tree->stack, tn) == -1)
    {
      return NULL;
    }

  /* look left */
  if(i < 0) return splaytree_find2(tree, ptr, tn->left);

  /* look right */
  if(i > 0) return splaytree_find2(tree, ptr, tn->right);

  /* we found it ! */
  return tn;
}

/*
 * splaytree_find
 * finds a value in a tree, splaying on that value
 */
void *splaytree_find(splaytree_t *tree, const void *ptr)
{
  if(tree == NULL || tree->head == NULL) return NULL;

  stack_clean(tree->stack);
  if(splaytree_find2(tree, ptr, tree->head) == NULL)
    {
      return NULL;
    }

  /* XXX: do we lose the whole tree just like that ? */
  if((tree->head = splaytree_splay(tree->stack)) == NULL)
    {
      return NULL;
    }

  return tree->head->ptr;
}


/*
 * splaytree_remove
 *
 */
int splaytree_remove(splaytree_t *tree, const void *ptr)
{
  splaytree_node_t *node;  /* we store the tree node to be deleted here */
  splaytree_node_t *l, *r; /* left and right of the node to be deleted */
  splaytree_node_t *temp;

  /*
   * find the node that we are supposed to delete.
   * if we can't find it, then the remove operation has failed.
   */
  stack_clean(tree->stack);
  if(splaytree_find2(tree, ptr, tree->head) == NULL)
    {
      return -1;
    }

  /*
   * now that we've found it, splay the tree to bring the node we are to
   * delete to the top of the tree
   */
  tree->head = node = splaytree_splay(tree->stack);

  l = node->left;
  r = node->right;

  /*
   * search for the right most node in the left tree
   * if there are no nodes on the left hand side of the tree, then we just
   * need to shift the head of the tree to whatever is there on the right
   * of it.
   */
  if(l != NULL)
    {
      stack_clean(tree->stack);
      stack_push(tree->stack, l);

      temp = l;
      while(temp->right != NULL)
	{
	  stack_push(tree->stack, temp->right);
	  temp = temp->right;
	}

      /* bring this node to the top of the tree with a splay operation */
      if((temp = splaytree_splay(tree->stack)) == NULL)
	{
	  return -1;
	}
      tree->head = temp;

      /*
       * as the right most node on the left branch has no nodes on the right 
       * branch, we connect the right hand branch to it
       */
      tree->head->right = r;
    }
  else
    {
      tree->head = r;
    }

  free(node);

  return 0;
}

/*
 * function: splaytree_findclosest
 * purpose:  find a value in the tree as close to the specified one as poss
 * comments:
 */
void *splaytree_findclosest(splaytree_t *tree, const void *ptr,
			    splaytree_diff_t diff)
{
  splaytree_node_t *ret;
  splaytree_node_t *first, *second;
  int               first_diff, second_diff;
  
  if(tree == NULL || tree->head == NULL) return NULL;

  stack_clean(tree->stack);

  ret = splaytree_find2(tree, ptr, tree->head);

  /* wow, the value we are looking for is actually in the tree! */
  if(ret != NULL)
    {
      tree->head = splaytree_splay(tree->stack);
      if(tree->head == NULL) return NULL;
      return tree->head->ptr;
    }

  /*
   * we need to get the last two items off the stack and figure out which
   * one of the two is the closest to the one we are looking for
   */
  first  = (splaytree_node_t *)stack_pop(tree->stack);
  second = (splaytree_node_t *)stack_pop(tree->stack);  

  /* if we havent got any items in the stack (i.e. the tree is empty) */
  if(first == NULL) return NULL;

  /* if there is only one item in the stack, splay? on it and return it */
  if(second == NULL)
    {
      stack_push(tree->stack, first);
      tree->head = splaytree_splay(tree->stack);
      if(tree->head == NULL) return NULL;
      return tree->head->ptr;
    }

  /* work out which one is closer to the value we are looking for */
  first_diff  = abs(diff(first->ptr,  ptr));
  second_diff = abs(diff(second->ptr, ptr));

  /*
   * if the first item is closer than the second, put the first back on the
   * stack and the splay on that
   * else put them both back on and splay on that
   */
  if(second_diff > first_diff)
    {
      stack_push(tree->stack, second);
    }
  else
    {
      stack_push(tree->stack, second);
      stack_push(tree->stack, first);
    }

  tree->head = splaytree_splay(tree->stack);
  stack_clean(tree->stack);
  if(tree->head == NULL) return NULL;

  return tree->head->ptr;
}

static int splaytree_depth2(splaytree_node_t *tn)
{
  int left = 0;
  int right = 0;

  if(tn == NULL) return 0;

  if(tn->left != NULL)
    {
      left = splaytree_depth2(tn->left) + 1;
    }
  if(tn->right != NULL)
    {
      right = splaytree_depth2(tn->right) + 1;
    }
  
  return (left > right) ? left : right;
}

/*
 * splaytree_depth
 * returns the longest path (the depth) of the splay tree
 */
int splaytree_depth(splaytree_t *tree)
{
  if(tree == NULL) return -1;
  if(tree->head == NULL) return 0;
  return splaytree_depth2(tree->head) + 1;
}

static void splaytree_destroy2(splaytree_node_t *tn, splaytree_free_t free_ptr)
{
  if(tn == NULL) return;

  splaytree_destroy2(tn->left, free_ptr);
  splaytree_destroy2(tn->right, free_ptr);
  if(free_ptr != NULL) free_ptr(tn->ptr);
  free(tn);

  return;
}

void splaytree_destroy(splaytree_t *tree, splaytree_free_t free_ptr)
{
  if(tree == NULL) return;
  splaytree_destroy2(tree->head, free_ptr);
  stack_destroy(tree->stack);
  free(tree);
  return;
}

void *splaytree_getrmlb(splaytree_t *tree)
{
  splaytree_node_t *tn;

  if(tree == NULL) return NULL;
  tn = tree->head;
  if(tn == NULL || tn->left == NULL) return NULL;
  tn = tn->left;

  while(tn->right != NULL)
    {
      tn = tn->right;
    }

  return tn->ptr;
}

void *splaytree_getlmrb(splaytree_t *tree)
{
  splaytree_node_t *tn;

  if(tree == NULL) return NULL;
  tn = tree->head;
  if(tn == NULL || tn->right == NULL) return NULL;
  tn = tn->right;

  while(tn->left != NULL)
    {
      tn = tn->left;
    }

  return tn->ptr;
}

static void splaytree_display2(splaytree_node_t *tn,
			       splaytree_display_t disp,
			       int pad)
{
  if(tn == NULL || disp == NULL)
    {
      return;
    }

  splaytree_display2(tn->left, disp, pad+1);
  disp(tn->ptr, pad);
  splaytree_display2(tn->right, disp, pad+1);

  return;
}

void splaytree_display(splaytree_t *tree, splaytree_display_t disp)
{
  splaytree_display2(tree->head, disp, 1);
  return;
}

static void splaytree_inorder2(splaytree_node_t *node,
			       splaytree_inorder_t func, void *in)
{
  if(node == NULL) return;

  splaytree_inorder2(node->left, func, in);
  func(in, node->ptr);
  splaytree_inorder2(node->right, func, in);

  return;
}

void splaytree_inorder(splaytree_t *tree, splaytree_inorder_t func, void *in)
{
  splaytree_inorder2(tree->head, func, in);
  return;
}

splaytree_t *splaytree_create(splaytree_cmp_t cmp)
{
  splaytree_t *tree;

  if((tree = (splaytree_t *)malloc(sizeof(splaytree_t))) == NULL)
    {
      return NULL;
    }

  tree->head  = NULL;
  tree->size  = 0;
  tree->cmp   = cmp;
  tree->stack = stack_create();

  return tree;
}

static int splaytree_count2(struct splaytree_node *node)
{
  int i=1;
  if(node->left != NULL)  i += splaytree_count2(node->left);
  if(node->right != NULL) i += splaytree_count2(node->right);
  return i;
}

int splaytree_count(splaytree_t *tree)
{
  if(tree == NULL) return -1;
  if(tree->head == NULL) return 0;
  return tree->size;
}

static splaytree_stack_t *stack_create()
{
  splaytree_stack_t *s;

  if((s = (splaytree_stack_t *)malloc(sizeof(splaytree_stack_t))) == NULL)
    {
      return NULL;
    }

  s->i = -1;
  s->c = 128;
  if((s->items = malloc(sizeof(void *) * s->c)) == NULL)
    {
      free(s);
      return NULL;
    }

  return s;
}

static void stack_clean(splaytree_stack_t *s)
{
  s->i = -1;
  return;
}

static void stack_destroy(splaytree_stack_t *s)
{
  free(s->items);
  free(s);
  return;
}

static int stack_push(splaytree_stack_t *s, void *item)
{
  void **items;

  if(s->i+1 == s->c)
    {
      if((items = malloc(sizeof(void *) * (s->c * 2))) == NULL)
	{
	  return -1;
	}

      s->c *= 2;
      s->items = items;
    }

  s->items[++s->i] = item;
  return 0;
}

static void *stack_pop(splaytree_stack_t *s)
{
  if(s->i == -1) return NULL;
  return s->items[s->i--];
}
