
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include "scamper_list.h"
#include "scamper_addr.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_file.h"

#include "mjl_splaytree.h"

#define OPT_SKIP          0x00001
#define OPT_DEBUG         0x00002
#define OPT_DSTEND        0x00004
#define OPT_OLDFORMAT     0x00008
#define OPT_HIDECOMMENTS  0x00010
#define OPT_HIDESRC       0x00020
#define OPT_HIDEDST       0x00040
#define OPT_HIDELIST      0x00080
#define OPT_HIDECYCLE     0x00100
#define OPT_HIDETIME      0x00200
#define OPT_HIDEREPLY     0x00400
#define OPT_HIDEHALT      0x00800
#define OPT_HIDEPATH      0x01000
#define OPT_HIDEIRTT      0x02000
#define OPT_HELP          0x04000

static uint32_t options = 0;

static int skip_numlines = 0;
static int debug_numlines = 0;

/* the input warts files */
static char **filelist = NULL;
static int    filelist_len = 0;

/* where the output goes.  stdout by default */
static FILE *out = NULL;

#ifdef USE_NETACUITY_
#define OPT_GEO           0x08000
#define OPT_GEOSERV       0x10000
static char               *geo_serv = NULL;
static struct splaytree_t *geo_seen = NULL;
#endif

static void usage(const uint32_t opt_mask)
{
  fprintf(stderr,
	  "usage: sc_analysis_dump [-oeCsdlctrHpig] [-S skip trace count]\n"
	  "                        [-D debug trace count] [-G geo server]\n"
	  "                        [file1 file2 ... fileN]\n");

  if(opt_mask == 0) return;

  return;
}

static int check_options(int argc, char *argv[])
{
  char opts[48];
  char ch;
  snprintf(opts, sizeof(opts), "oeCsdlctrHpiS:D:h");

#ifdef USE_NETACUITY
  strcat(opts, "gG:");
#endif

  while((ch = getopt(argc, argv, opts)) != -1)
    {
      switch(ch)
	{
	case 'S':
	  options |= OPT_SKIP;
	  skip_numlines = atoi(optarg);
	  break;

	case 'D':
	  options |= OPT_DEBUG;
	  debug_numlines = atoi(optarg);
	  break;

	case 'e':
	  options |= OPT_DSTEND;
	  break;

	case 'o':
	  options |= OPT_OLDFORMAT;
	  break;

	case 'C':
	  options |= OPT_HIDECOMMENTS;
	  break;

	case 's':
	  options |= OPT_HIDESRC;
	  break;

	case 'd':
	  options |= OPT_HIDEDST;
	  break;

	case 'l':
	  options |= OPT_HIDELIST;
	  break;

	case 'c':
	  options |= OPT_HIDECYCLE;
	  break;

	case 't':
	  options |= OPT_HIDETIME;
	  break;

	case 'r':
	  options |= OPT_HIDEREPLY;
	  break;

	case 'H':
	  options |= OPT_HIDEHALT;
	  break;

	case 'p':
	  options |= OPT_HIDEPATH;
	  break;

	case 'i':
	  options |= OPT_HIDEIRTT;
	  break;

#ifdef USE_NETACUITY_
	case 'g':
	  options |= OPT_GEO;
	  break;

	case 'G':
	  options |= OPT_GEOSERV;
	  geo_serv = optarg;
	  break;
#endif

	case 'h':
	  options |= OPT_HELP;
	  break;

	default:
	  usage(0);
	  return -1;
	}
    }

  filelist = argv + optind;
  filelist_len = argc - optind;

  return 0;
}

static char *rtt_tostr(char *str, const size_t len, const struct timeval *rtt)
{
  if(rtt != NULL)
    {
      snprintf(str, len, "%ld.%03ld",
	       (rtt->tv_sec * 1000) + (rtt->tv_usec / 1000),
	       rtt->tv_usec % 1000);
    }
  else
    {
      str[0] = '\0';
    }

  return str;
}

static void print_header_fields(const scamper_trace_t *trace)
{
  char  buf[256];

  if((options & OPT_HIDESRC) == 0)
    {
      fprintf(out, "\t%s", scamper_addr_tostr(trace->src, buf, sizeof(buf)));
    }

  if((options & OPT_HIDEDST) == 0)
    {
      fprintf(out, "\t%s", scamper_addr_tostr(trace->dst, buf, sizeof(buf)));
    }

  if((options & OPT_HIDELIST) == 0)
    {
      fprintf(out, "\t%d", (trace->list != NULL) ? trace->list->id : 0);
    }

  if((options & OPT_HIDECYCLE) == 0)
    {
      fprintf(out, "\t%d", (trace->cycle != NULL) ? trace->cycle->id : 0);
    }

  if((options & OPT_HIDETIME) == 0)
    {
      fprintf(out, "\t%lu", trace->start.tv_sec);
    }

  return;
}

static void print_reply_fields(const scamper_trace_hop_t *hop)
{
  char rtt[64];

  if(hop != NULL)
    {
      rtt_tostr(rtt, sizeof(rtt), &hop->rtt);
      fprintf(out, "\tR\t%s\t%d\t%d", rtt, hop->probe_ttl, hop->reply_ttl);
    }
  else
    {
      fprintf(out, "\tN\t0\t0\t0");
    }

  return;
}

static void print_halt_fields(const scamper_trace_t *trace)
{
  switch(trace->stop_reason)
    {
    case SCAMPER_TRACE_STOP_COMPLETED:
      fprintf(out, "\tS\t0");
      break;

    case SCAMPER_TRACE_STOP_UNREACH:
      fprintf(out, "\tU\t%d", trace->stop_data);
      break;

    case SCAMPER_TRACE_STOP_LOOP:
      fprintf(out, "\tL\t%d", trace->stop_data);
      break;

    case SCAMPER_TRACE_STOP_DEAD:
      fprintf(out, "\tG\t%d", trace->stop_data);
      break;

    default:
      fprintf(out, "\t?\t0");
      break;
    }
  return;
}

static void print_old_fields(const scamper_trace_t *trace,
			     const scamper_trace_hop_t *hop)
{
  char src[256], dst[256], rtt[256];

  fprintf(out, " %s %s %ld %s %d",
	  scamper_addr_tostr(trace->src, src, sizeof(src)),
	  scamper_addr_tostr(trace->dst, dst, sizeof(dst)),
	  trace->start.tv_sec,
	  rtt_tostr(rtt, sizeof(rtt), (hop != NULL) ? &hop->rtt : NULL),
	  trace->hop_count);

  return;
}

static void print_path_fields(const scamper_trace_t *trace,
			      const scamper_trace_hop_t *dst)
{
  scamper_trace_hop_t *hop;
  char path_complete;
  char rtt[128], addr[128];
  size_t size;
  int i;
  int *num_replies_per_ttl = NULL;
  int *num_probes_per_ttl = NULL;

  if(trace->hop_count == 0)
    {
      return;
    }

  /*
   * allocate an array to record how many probes per ttl were made, and
   * another to record how many replies per ttl the trace has,
   */
  size = sizeof(int) * trace->hop_count;
  if((num_probes_per_ttl = malloc(size)) == NULL)
    {
      goto cleanup;
    }
  memset(num_probes_per_ttl, 0, size);
  if((num_replies_per_ttl = malloc(size)) == NULL)
    {
      goto cleanup;
    }
  memset(num_replies_per_ttl, 0, size);

  for(i=0; i<trace->hop_count; i++)
    {
      if((hop = trace->hops[i]) != NULL)
	{
	  do
	    {
	      num_replies_per_ttl[i]++;

	      if(hop->probe_id + 1 > num_probes_per_ttl[i])
		{
		  num_probes_per_ttl[i] = hop->probe_id + 1;
		}

#ifdef USE_NETACUITY_
	      if(options & OPT_GEO)
		{
		  print_geo_info(hop->addr);
		}
#endif
	    }
	  while((hop = hop->next) != NULL);
	}
      else
	{
	  num_probes_per_ttl[i] = trace->attempts;
	}
    }

  /*
   * decide what the path_complete flag should be set to.  if we reached
   * the destination then the path_complete flag == 'C' (for complete).
   * otherwise the path_complete flag == 'I' (incomplete) or 'N' if
   * using the old sk_analysis_dump output format.
   */
  path_complete = 'I';
  if(dst != NULL)
    {
      for(i=0; i<trace->hop_count; i++)
	{
	  if(trace->hops[i] == NULL)
	    {
	      break;
	    }
	}

      if(i == trace->hop_count)
	{
	  path_complete = 'C';
	}
      else
	{
	  path_complete = 'I';
	}
    }
  else if(options & OPT_OLDFORMAT)
    {
      path_complete = 'N';
    }

  /*
   * actually output the path complete flag, and some extra old fields
   * if requested
   */
  if((options & OPT_OLDFORMAT) == 0)
    {
      fprintf(out, "\t%c", path_complete);
    }
  else
    {
      fprintf(out, "%c", path_complete);
      print_old_fields(trace, dst);
    }

  if(options & OPT_DSTEND)
    {
      fprintf(out, "\t%s", scamper_addr_tostr(trace->src, addr, sizeof(addr)));
      if((options & OPT_HIDEIRTT) == 0)
	{
	  fprintf(out, ",0.000,0");
	}
    }

  for(i=0; i<trace->hop_count; i++)
    {
      if((options & OPT_OLDFORMAT) == 0)
	{
	  fprintf(out, "\t");
	}
      else
	{
	  fprintf(out, " ");
	}

      if((hop = trace->hops[i]) != NULL)
	{
	  for(;;)
	    {
	      if((options & (OPT_HIDEIRTT | OPT_OLDFORMAT)) == 0)
		{
		  fprintf(out, "%s,%s,%d",
			  scamper_addr_tostr(hop->addr, addr, sizeof(addr)),
			  rtt_tostr(rtt, sizeof(rtt), &hop->rtt),
			  num_probes_per_ttl[i]);
		}

	      if((hop = hop->next) != NULL)
		{
		  if((options & OPT_OLDFORMAT) == 0)
		    {
		      fprintf(out, ";");
		    }
		  else
		    {
		      fprintf(out, ",");
		    }
		}
	      else break;
	    }
	}
      else
	{
	  fprintf(out, "q");
	}
    }

 cleanup:
  if(num_probes_per_ttl != NULL) free(num_probes_per_ttl);
  if(num_replies_per_ttl != NULL) free(num_replies_per_ttl);

  return;
}

#ifdef USE_NETACUITY_
static int print_geo_info(scamper_addr_t *addr)
{
  na_geo_struct answer;
  char buf[256];

  if(splaytree_find(geo_seen, addr) != NULL)
    {
      return 0;
    }

  if(splaytree_insert(geo_seen, scamper_addr_use(addr)) != 1)
    {
      return -1;
    }

  if(scamper_addr_tostr(addr, buf, sizeof(buf)) == NULL)
    {
      return -1;
    }

  if(na_query_geo(buf, &answer))
    {
      fprintf(out, "G\t%s\t%s=%d\t%s=%d\t%s=%d\t%s\t%d\t%.3f\t%.3f\n",
	      buf,
	      answer.country, answer.country_c,
	      answer.region, answer.region_c,
	      answer.city, answer.city_c,
	      answer.speed,
	      answer.metro_code,
	      answer.latitude, answer.longitude);
    }
  else
    {
      fprintf(stderr, "Error in na_query_geo(%s)\n", buf);
    }

  return 0;
}

static void print_path_geo_info(const scamper_trace_t *trace)
{
  scamper_trace_hop_t *hop;
  int i;

  for(i=0; i<trace->hop_count; i++)
    {
      for(hop = trace->hops[i]; hop != NULL; hop = hop->next)
	{
	  print_geo_info(hop->addr);
	}
    }

  return;
}

static int setup_netacuity_server(char *server)
{
  struct addrinfo hints, *res, *res0;
  char buf[256];
  int set = 0;

  memset(&hints, 0, sizeof(hints));

  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_protocol = IPPROTO_UDP;
  hints.ai_family   = AF_INET;

  if((error = getaddrinfo(ipstr, NULL, &hints, &res0)) != 0 || res0 == NULL)
    {
      fprintf(stderr, "could not resolve %s: %s", server, gai_strerror(error));
      return -1;
    }

  for(res = res0; res != NULL; res = res->ai_next)
    {
      if(res->ai_family == PF_INET)
	{
	  inet_ntop(res->ai_family,
		    &((struct sockaddr_in *)ai_list->ai_addr)->sin_addr,
		    buf, sizeof(buf));

	  if(na_api_set_server_addr(buf))
	    {
	      set = 1;
	      break;
	    }
	  else
	    {
	      fprintf(stderr, "Error in setting server addr %s", buf);
	    }
	}
    }

  freeaddrinfo(res0);

  if(set == 1)
    {
      geo_seen = splaytree_create((splaytree_cmp_t)scamper_addr_cmp);
      if(geo_seen == NULL)
	{
	  return -1;
	}

      return 0;
    }

  return -1;
}

#endif /* USE_NETACUITY_ */

static void print_trace(const scamper_trace_t *trace)
{
  scamper_trace_hop_t *dst = NULL, *hop;
  int i;

  if(trace->hop_count == 0)
    {
      return;
    }

  /* try and determine the hop that corresponds to the destination */
  if(trace->stop_reason != SCAMPER_TRACE_STOP_NONE &&
     trace->stop_reason != SCAMPER_TRACE_STOP_DEAD &&
     trace->stop_reason != SCAMPER_TRACE_STOP_ERROR)
    {
      for(i=0; i<trace->hop_count && dst == NULL; i++)
	{
	  for(hop = trace->hops[i]; hop != NULL; hop = hop->next)
	    {
	      if(hop->addr == trace->dst || 
		 SCAMPER_TRACE_HOP_IS_ICMP_UNREACH_PORT(hop) ||
		 SCAMPER_TRACE_HOP_IS_ICMP_ECHO_REPLY(hop))
		{
		  dst = hop;
		  break;
		}
	    }
	}
    }

#ifdef USE_NETACUITY_
  if(options & OPT_GEO)
    {
      if((options & OPT_HIDESRC) == 0)
	{
	  print_geo_info(trace->src);
	}
      if((options & OPT_HIDEDST) == 0)
	{
	  print_geo_info(trace->dst);
	}
      if((options & OPT_HIDEPATH) == 0)
	{
	  print_path_geo_info(trace);
	}
    }
#endif

  if((options & OPT_OLDFORMAT) == 0)
    {
      fprintf(out, "T");
      print_header_fields(trace);

      if((options & OPT_HIDEREPLY) == 0)
	{
	  print_reply_fields(dst);
	}

      if((options & OPT_HIDEHALT) == 0)
	{
	  print_halt_fields(trace);
	}
    }

  if((options & OPT_HIDEPATH) == 0 || (options & OPT_OLDFORMAT))
    {
      print_path_fields(trace, dst);
    }

  fprintf(out, "\n");

  return;
}

int main(int argc, char *argv[])
{
  scamper_file_t *file;
  scamper_trace_t *trace;
  int i, n;

  out = stdout;

  if(check_options(argc, argv) == -1)
    {
      return -1;
    }

  if(filelist_len != 0)
    {
      for(i=0; i<filelist_len; i++)
	{
	  if((file = scamper_file_open(filelist[i], 'r', NULL)) == NULL)
	    {
	      fprintf(stderr, "unable to open %s\n", filelist[i]);
	      if((options & OPT_HIDECOMMENTS) == 0)
		{
		  fprintf(out, "# unable to open %s\n", filelist[i]);
		}

	      continue;
	    }

	  n = 0;

	  while((trace = scamper_file_read_trace(file)) != NULL)
	    {
	      if((options & OPT_DEBUG) && n == debug_numlines)
		{
		  scamper_trace_free(trace);
		  break;
		}

	      n++;

	      if(n > skip_numlines)
		{
		  print_trace(trace);
		}

	      scamper_trace_free(trace);
	    }

	  scamper_file_close(file);
	}
    }

  return 0;
}
