/*
 * scamper: a skitter clone
 *
 * $Id: scamper.c,v 1.125 2005/05/27 02:38:53 mjl Exp $
 *
 *          Matthew Luckie, University of Waikato / NLANR MNA
 *          January 2003
 *          mjl@luckie.org.nz
 *
 * scamper is my attempt at cloning CAIDA's skitter, giving it IPv6 support
 * in addition to IPv4 support.  it was written using traceroute's code as
 * documentation as to how i should approach a traceroute.  scamper also
 * does PMTU discovery.
 *
 * Copyright (C) 2003-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>

#include <netinet/in.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#include <assert.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper.h"
#include "scamper_debug.h"
#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_dl.h"
#include "scamper_probe.h"
#include "utils.h"

#include "scamper_file.h"
#include "scamper_outfiles.h"
#include "scamper_task.h"
#include "scamper_target.h"
#include "scamper_queue.h"
#include "scamper_addresslist.h"
#include "scamper_getsrc.h"
#include "scamper_addr2mac.h"
#include "scamper_icmp_resp.h"
#include "scamper_icmp4.h"
#include "scamper_icmp6.h"
#include "scamper_udp4.h"
#include "scamper_udp6.h"
#include "scamper_rtsock.h"
#include "scamper_fds.h"
#include "scamper_privsep.h"
#include "scamper_control.h"
#include "scamper_bgp.h"
#include "scamper_do_trace.h"
#include "scamper_do_ping.h"

static uint32_t options = 0;
#define OPT_PPS         0x00000001 /* p: */
#define OPT_ATTEMPTS    0x00000002 /* a: */
#define OPT_HOPLIMIT    0x00000004 /* h: */
#define OPT_DPORT       0x00000008 /* d: */
#define OPT_TIMEOUT     0x00000010 /* t: */
#define OPT_OUTFILE     0x00000020 /* o: */
#define OPT_OUTTYPE     0x00000040 /* O: */
#define OPT_IPV4	0x00000080 /* 4: */
#define OPT_IPV6	0x00000100 /* 6: */
#define OPT_IGNORELOOPS 0x00000200 /* l: */
#define OPT_VERSION     0x00000800 /* v: */
#define OPT_PMTU        0x00001000 /* m: */
#define OPT_TCP         0x00002000 /* T: */
#define OPT_BGP         0x00004000 /* b: */
#define OPT_RTT         0x00008000 /* r: */
#define OPT_ALLATTEMPTS 0x00010000 /* A: */
#define OPT_HOLDTIME    0x00020000 /* H: */
#define OPT_DAEMON      0x00040000 /* D: */
#define OPT_IPLIST      0x00080000 /* i: */
#define OPT_SPORT       0x00100000 /* s: */
#define OPT_DL          0x00200000 /* P: */
#define OPT_HELP        0x80000000 /* ?: */

/* the cache of addresses that scamper is dealing with */
scamper_addrcache_t *addrcache = NULL;

/*
 * sport:
 * the first 15 bits of the process id with the upper bit set to make the
 * source port unlikely to be in use at the scamper host
 */
static int sport;
static int dport = SCAMPER_DPORT_DEF;

/*
 * attempts: the number of attempts to make on each hop before we give up
 * hoplimit: the maximum number of hops to probe on each path
 * timeout:  how long to wait for a response 
 */
static int attempts = SCAMPER_ATTEMPTS_DEF;
static int hoplimit = SCAMPER_HOPLIMIT_DEF;
static int timeout  = SCAMPER_TIMEOUT_DEF;

/*
 * pps: how many probe packets to send per second.
 * wait_between: specifies the wait between sending probes required to fill
 *  the packets-per-second rate.  by default it is 50 msec between probes
 * probe_window: specifies how far the probe_window is allowed to extend
 *  before being truncated.  by default that is 250ms
 *
 * at 20 pps / 50 ms between probes, the data stream is something like 0.8kB/s
 */
static int pps          = SCAMPER_PPS_DEF;
static int wait_between = 1000000 / SCAMPER_PPS_DEF;
static int probe_window = 250000;

/*
 * the time to hold a trace in a queue after declaring a trace 'done' so
 * that any other responses coming back can be collected for that trace
 */
static int holdtime = 0;

static char       *outfile = "-";
static char       *outtype = "traceroute";
scamper_outfile_t *file    = NULL;

static int daemon_port = 12579;

/* an array of ip addresses or address lists passed on the command line */
static char **arglist     = NULL;
static int    arglist_len = 0;

/*
 * a flag that specifies if scamper should exit when it is done with
 * the current active window of traces
 */
static int    exit_when_done = 1;

static scamper_list_t *list;

static void usage_str(char c, char *str)
{
  fprintf(stderr, "            -%c %s\n", c, str);
  return;
}

static void usage(uint32_t opt_mask)
{
  char buf[256];

  fprintf(stderr,
	  "usage: scamper [-?46AlmPv] [-H holdtime]\n"
	  "               [-a attempts] [-h hoplimit] [-p pps] [-t timeout]\n"
	  "               [-d dport] [-s sport] [-o outfile] [-O outtype]\n"
	  "               [-D port | -i addr 1..N | listfile]\n");

  if(opt_mask == 0) return;

  fprintf(stderr, "\n");

  if(opt_mask & OPT_HELP)
    usage_str('?', "give an overview of the usage of scamper");

  if(opt_mask & OPT_IPV4)
    usage_str('4', "only open IPv4 sockets");

  if(opt_mask & OPT_IPV6)
    usage_str('6', "only open IPv6 sockets");

  if(opt_mask & OPT_ATTEMPTS)
    {
      snprintf(buf, sizeof(buf),
	       "no of attempts (%d < attempts < %d) to make on each hop",
	       SCAMPER_ATTEMPTS_MIN, SCAMPER_ATTEMPTS_MAX);

      usage_str('a', buf);
    }

  if(opt_mask & OPT_ALLATTEMPTS)
    usage_str('A', "send all allotted attempts per hop");

  if(opt_mask & OPT_DPORT)
    {
      snprintf(buf, sizeof(buf),
	       "base destination port (%d < dport < %d) to probe",
	       SCAMPER_DPORT_MIN, SCAMPER_DPORT_MAX);

      usage_str('d', buf);
    }

  if(opt_mask & OPT_DAEMON)
    usage_str('D', "start as a daemon listening for commands on a port");

  if(opt_mask & OPT_HOPLIMIT)
    {
      snprintf(buf, sizeof(buf),
	       "number of hops (%d < hops < %d) to probe",
	       SCAMPER_HOPLIMIT_MIN, SCAMPER_HOPLIMIT_MAX);

      usage_str('h', buf);
    }

  if(opt_mask & OPT_HOLDTIME)
    {
      snprintf(buf, sizeof(buf),
	       "time to hold trace for delayed responses (%d < holdtime %d)",
	       SCAMPER_HOLDTIME_MIN, SCAMPER_HOLDTIME_MAX);

      usage_str('H', buf);
    }

  if(opt_mask & OPT_IPLIST)
    usage_str('i', "IP addresses to trace provided on the command line");

  if(opt_mask & OPT_IGNORELOOPS)
    usage_str('l', "do not stop on detection of a routing loop");
 
  if(opt_mask & OPT_PMTU)
    usage_str('m', "probe for path MTU");

  if(opt_mask & OPT_OUTFILE)
    usage_str('o', "specify the file to write output to");

  if(opt_mask & OPT_OUTTYPE)
    usage_str('O', "specify the type of output [warts | traceroute]");

  if(opt_mask & OPT_PPS)
    {
      snprintf(buf, sizeof(buf),
	       "number of packets per second to send (%d <= pps <= %d)",
	       SCAMPER_PPS_MIN, SCAMPER_PPS_MAX);

      usage_str('p', buf);
    }

  if(opt_mask & OPT_DL)
    usage_str('P', "use a datalink to get tx timestamps for outgoing probes");

  if(opt_mask & OPT_SPORT)
    usage_str('s', "source port to use");

  if(opt_mask & OPT_TIMEOUT)
    {
      snprintf(buf, sizeof(buf),
	       "number of seconds (%d < seconds < %d) to timeout for response",
	       SCAMPER_TIMEOUT_MIN, SCAMPER_TIMEOUT_MAX);

      usage_str('t', buf);
    }

  if(opt_mask & OPT_VERSION)
    usage_str('v', "output the version of scamper this binary is");

  return;
}

static int set_opt(int opt, char *str, int (*setfunc)(int), int min, int max)
{
  long l;

  if(string_isnumber(str) == 0 || string_tolong(str, &l) == -1 ||
     l < min || l > max)
    {
      usage(opt);
      return -1;
    }

  return setfunc(l);
}

static int check_options(int argc, char *argv[])
{
  char  ch;
  long  lo;
  char *opts = "46a:Ad:D:h:H:ilmo:O:p:Ps:t:w:v?";
  char *opt_attempts = NULL, *opt_dport = NULL;
  char *opt_daemon = NULL, *opt_hoplimit = NULL, *opt_holdtime = NULL;
  char *opt_pps = NULL, *opt_sport = NULL, *opt_timeout = NULL;

  while((ch = getopt(argc, argv, opts)) != -1)
    {
      switch(ch)
	{
        case '4':
          options |= OPT_IPV4;
          break;

        case '6':
          options |= OPT_IPV6;
          break;

        case 'a':
          options |= OPT_ATTEMPTS;
          opt_attempts = optarg;
          break;

	case 'A':
	  options |= OPT_ALLATTEMPTS;
	  break;

	case 'd':
	  options |= OPT_DPORT;
          opt_dport = optarg;
	  break;

	case 'D':
	  options |= OPT_DAEMON;
	  opt_daemon = optarg;
	  break;

        case 'h':
          options |= OPT_HOPLIMIT;
          opt_hoplimit = optarg;
          break;

	case 'H':
	  options |= OPT_HOLDTIME;
	  opt_holdtime = optarg;
	  break;

	case 'i':
	  options |= OPT_IPLIST;
	  break;

	case 'l':
	  options |= OPT_IGNORELOOPS;
	  break;

	case 'm':
	  options |= OPT_PMTU;
	  break;

        case 'o':
          options |= OPT_OUTFILE; 
          outfile = optarg;
          break;

	case 'O':
	  options |= OPT_OUTTYPE;
	  outtype = optarg;
	  break;

	case 'p':
	  options |= OPT_PPS;
	  opt_pps = optarg;
	  break;

	case 'P':
	  options |= OPT_DL;
	  break;

	case 's':
	  options |= OPT_SPORT;
          opt_sport = optarg;
	  break;

        case 't':
          options |= OPT_TIMEOUT;
          opt_timeout = optarg;
          break;

	case 'v':
	  options |= OPT_VERSION;
	  break;

	case '?':
	  options |= OPT_HELP;
	  usage(0xffffffff);
	  return -1;

	default:
	  return -1;
	}
    }

  if(options & OPT_VERSION)
    {
      fprintf(stderr, "scamper version %s\n", SCAMPER_VERSION);
      return -1;
    }

  if(options & OPT_PPS &&
     set_opt(OPT_PPS, opt_pps, scamper_pps_set,
	     SCAMPER_PPS_MIN, SCAMPER_PPS_MAX) == -1)
    {
      return -1;
    }

  if(options & OPT_ATTEMPTS &&
     set_opt(OPT_ATTEMPTS, opt_attempts, scamper_attempts_set,
	     SCAMPER_ATTEMPTS_MIN, SCAMPER_ATTEMPTS_MAX) == -1)
    {
      return -1;
    }

  if(options & OPT_DPORT &&
     set_opt(OPT_DPORT, opt_dport, scamper_dport_set,
	     SCAMPER_DPORT_MIN, SCAMPER_DPORT_MAX) == -1)
    {
      return -1;
    }

  if(options & OPT_HOLDTIME &&
     set_opt(OPT_HOLDTIME, opt_holdtime, scamper_holdtime_set,
	     SCAMPER_HOLDTIME_MIN, SCAMPER_HOLDTIME_MAX) == -1)
    {
      return -1;
    }

  if(options & OPT_HOPLIMIT &&
     set_opt(OPT_HOPLIMIT, opt_hoplimit, scamper_hoplimit_set,
	     SCAMPER_HOPLIMIT_MIN, SCAMPER_HOPLIMIT_MAX) == -1)
    {
      return -1;
    }

  if(options & OPT_SPORT &&
     set_opt(OPT_SPORT, opt_sport, scamper_sport_set, 1, 65535) == -1)
    {
      return -1;
    }

  if(options & OPT_TIMEOUT &&
     set_opt(OPT_TIMEOUT, opt_timeout, scamper_timeout_set,
	     SCAMPER_TIMEOUT_MIN, SCAMPER_TIMEOUT_MAX) == -1)
    {
      return -1;
    }

  if(options & OPT_IPV4 && options & OPT_IPV6)
    {
      usage(OPT_IPV4 | OPT_IPV6);
      return -1;
    }

  if(options & OPT_IPLIST && options & OPT_DAEMON)
    {
      usage(OPT_IPLIST | OPT_DAEMON);
      return -1;
    }

  if(strcasecmp(outtype, "traceroute") != 0 &&
     strcasecmp(outtype, "warts") != 0)
    {
      usage(OPT_OUTTYPE);
      return -1;
    }

  /*
   * if started in daemonised mode, then there should be no leftover
   * arguments.  if it is not to be run as a daemon, then there should be
   * a single argument which is the list to run
   */
  if(options & OPT_DAEMON)
    {
      if(string_isnumber(opt_daemon) == 0 ||
	 string_tolong(opt_daemon, &lo) == -1 ||
	 lo < 1 || lo > 65535)
	{
	  usage(OPT_DAEMON);
	  return -1;
	}

      daemon_port = lo;

      if(argc - optind != 0)
	{
	  usage(0);
	  return -1;
	}
    }
  else
    {
      arglist     = argv + optind;
      arglist_len = argc - optind;
      if(arglist_len < 1)
	{
	  usage(0);
	  return -1;
	}
    }

  return 0;
}

void scamper_exitwhendone(int on)
{
  if(on == 1 || on == 0)
    {
      exit_when_done = on;
    }
  return;
}

int scamper_sport_get()
{
  return sport;
}

int scamper_sport_set(const int sp)
{
  if(sp >= SCAMPER_SPORT_MIN && sp <= SCAMPER_SPORT_MAX)
    {
      sport = sp;
      return 0;
    }

  return -1;
}

int scamper_dport_get()
{
  return dport;
}

int scamper_dport_set(const int dp)
{
  if(dp >= SCAMPER_DPORT_MIN && dp <= SCAMPER_DPORT_MAX)
    {
      dport = dp;
      return 0;
    }

  return -1;
}

int scamper_timeout_get()
{
  return timeout;
}

int scamper_timeout_set(const int to)
{
  if(to >= SCAMPER_TIMEOUT_MIN && to <= SCAMPER_TIMEOUT_MAX)
    {
      timeout = to;
      return 0;
    }

  return -1;
}

int scamper_holdtime_get()
{
  return holdtime;
}

int scamper_holdtime_set(const int ht)
{
  if(ht >= SCAMPER_HOLDTIME_MIN && ht <= SCAMPER_HOLDTIME_MAX)
    {
      holdtime = ht;
      return 0;
    }

  return -1;
}

int scamper_attempts_get()
{
  return attempts;
}

int scamper_attempts_set(const int a)
{
  if(a >= SCAMPER_ATTEMPTS_MIN && a <= SCAMPER_ATTEMPTS_MAX)
    {
      attempts = a;
      return 0;
    }

  return -1;
}

int scamper_pps_get()
{
  return pps;
}

int scamper_pps_set(const int p)
{
  if(p >= SCAMPER_PPS_MIN && p <= SCAMPER_PPS_MAX)
    {
      /*
       * reset the pps scamper is operating at.  re-calculate the inter-probe
       * delay, and the maximum size of the probe window.
       */

      pps = p;
      wait_between = 1000000 / pps;
      probe_window = (wait_between < 250000 ? 250000 : wait_between + 250000);

      return 0;
    }

  return -1;
}

int scamper_hoplimit_get()
{
  return hoplimit;
}

int scamper_hoplimit_set(const int hl)
{
  if(hl >= SCAMPER_HOPLIMIT_MIN && hl <= SCAMPER_HOPLIMIT_MAX)
    {
      hoplimit = hl;
      return 0;
    }

  return -1;
}

int scamper_option_allattempts()
{
  if(options & OPT_ALLATTEMPTS)
    {
      return 1;
    }
  return 0;
}

int scamper_option_detectloops()
{
  if(options & OPT_IGNORELOOPS) return 0;
  return 1;
}

int scamper_option_dl()
{
  if(options & OPT_DL) return 1;
  return 0;
}

int scamper_option_ipv4()
{
  if(options & OPT_IPV4) return 1;
  return 0;
}

int scamper_option_ipv6()
{
  if(options & OPT_IPV6) return 1;
  return 0;
}

int scamper_option_pmtu()
{
  if(options & OPT_PMTU) return 1;
  return 0;
}

static void scamper_hup(int sig)
{
  return;
}

static void scamper_chld(int sig)
{
  pid_t pid;
  int   status;

  for(;;)
    {
      if((pid = waitpid(-1, &status, WNOHANG)) == -1)
	{
	  break;
	}
    }

  return;
}

/*
 * handle_icmp:
 *
 * given an response, figure out which task was responsible for
 * sending the probe, and pass the response to it.
 */
static int handle_icmp(scamper_icmp_resp_t *ir, int type)
{
  scamper_target_t  *target;
  scamper_task_t    *task;
  scamper_addr_t     addr;

  /* the target address of the probe is embedded in the response */
  if(SCAMPER_ICMP_RESP_IS_TTL_EXP(ir) ||
     SCAMPER_ICMP_RESP_IS_UNREACH(ir) ||
     SCAMPER_ICMP_RESP_IS_PACKET_TOO_BIG(ir))
    {
      /* not able to handle this response, drop it */
      if(!SCAMPER_ICMP_RESP_INNER_IS_SET(ir))
	{
	  return 0;
	}

      addr.addr = scamper_icmp_resp_inner_dst(ir);
    }
  else if(SCAMPER_ICMP_RESP_IS_ECHO_REPLY(ir))
    {
      addr.addr = scamper_icmp_resp_src(ir);
    }
  else
    {
      return 0;
    }

  addr.type = type;

  if((target = scamper_target_find(&addr)) == NULL)
    {
      return 0;
    }

  task = target->task;

  if(task->funcs->handle_icmp != NULL)
    {
      task->funcs->handle_icmp(task, ir);
    }

  return 0;
}

void scamper_icmp6_read_cb(const int fd, void *param)
{
  scamper_icmp_resp_t *resp = param;
  int error;

  if((error = scamper_icmp6_recv(resp)) == 0)
    {
      handle_icmp(resp, SCAMPER_ADDR_TYPE_IPV6);
    }

  scamper_icmp_resp_clean(resp);
  return;
}

void scamper_icmp4_read_cb(const int fd, void *param)
{
  scamper_icmp_resp_t *resp = param;
  int error;

  if((error = scamper_icmp4_recv(resp)) == 0)
    {
      handle_icmp(resp, SCAMPER_ADDR_TYPE_IPV4);
    }

  scamper_icmp_resp_clean(resp);
  return;
}

/*
 * scamper:
 * this bit of code contains most of the logic for driving the parallel
 * traceroute process.
 */
static int scamper()
{
  struct timeval        tv;
  struct timeval        lastprobe;
  struct timeval        nextprobe;
  struct timeval       *timeout;
  scamper_target_t     *target;
  scamper_task_t       *task;
  scamper_icmp_resp_t   resp;
  int64_t               diff;
  int                   icmp4 = -1;
  int                   icmp6 = -1;
  int                   i;

  if((list = scamper_list_alloc(0, "default", "default")) == NULL)
    {
      return -1;
    }

  /* allocate the cache of addresses for scamper to keep track of */
  if((addrcache = scamper_addrcache_alloc()) == NULL)
    {
      return -1;
    }

  /*
   * setup the file descriptor monitor, as scamper will be adding icmp
   * file descriptors next.
   */
  if(scamper_fds_init() == -1)
    {
      return -1;
    }

  /*
   * if we have not been restricted to opening IPv4 sockets only, then
   * open the IPv6 sockets
   */
  if((options & OPT_IPV4) == 0)
    {
      if((icmp6 = scamper_icmp6_open()) == -1)
	{
	  return -1;
	}

      /* put the icmp6 socket into the list of fds to monitor */
      scamper_fds_read_add(icmp6, scamper_icmp6_read_cb, &resp);

      if(scamper_udp6_open(sport) == -1)
	{
	  return -1;
	}
    }

  /*
   * if we have not been restricted to opening IPv6 sockets only, then
   * open the IPv4 sockets
   */
  if((options & OPT_IPV6) == 0)
    {
      if((icmp4 = scamper_icmp4_open()) == -1)
	{
	  return -1;
	}

      /* put the icmp4 socket into the list of fds to monitor */
      scamper_fds_read_add(icmp4, scamper_icmp4_read_cb, &resp);

      if(scamper_udp4_open(sport) == -1)
	{
	  return -1;
	}
    }

  /*
   * this has to be done before priviledge separation, as if scamper is
   * running on a BPF system it has to open a BPF fd to establish
   * version compatibility
   */
  if(scamper_dl_init() == -1)
    {
      return -1;
    }

  /*
   * if we have been told to open a control socket and daemonise, then do
   * that now.
   */
  if(options & OPT_DAEMON)
    {
      if(scamper_control_init(daemon_port) == -1)
	{
	  return -1;
	}

      /*
       * scamper should wait for more tasks when it has finished with the
       * active window
       */
      exit_when_done = 0;
    }

#ifndef WITHOUT_PRIVSEP
  /* revoke the root priviledges we started with */
  if(scamper_privsep_init() == -1)
    {
      return -1;
    }
#endif

  /* initialise the subsystem responsible for obtaining source addresses */
  if(scamper_getsrc_init() == -1)
    {
      return -1;
    }

  /* initialise the subsystem responsible for recording mac addresses */
  if(scamper_addr2mac_init() == -1)
    {
      return -1;
    }

  scamper_debug(NULL, "scamper is bound to UDP port %d", sport);

  /* if scamper is going to need a route socket, get one now */
  if(options & OPT_DL || options & OPT_PMTU)
    {
      if(scamper_rtsock_init() == -1)
	{
	  return -1;
	}
    }

  /* initialise the structures necessary to keep track of addresses to probe */
  if(scamper_addresslist_init() == -1)
    {
      return -1;
    }

  /*
   * if we have an address list of some description on the command line,
   * read the addresses now
   */
  if(options & OPT_IPLIST)
    {
      if(scamper_addresslist_addarray(NULL, arglist, arglist_len) == -1)
	{
	  return -1;
	}
    }
  else if((options & OPT_DAEMON) == 0)
    {
      for(i=0; i<arglist_len; i++)
	{
	  if(scamper_addresslist_addfile(NULL, arglist[i]) == -1)
	    {
	      printerror(errno, strerror, __func__,
			 "could not parse %s", arglist[i]);
	      return -1;
	    }
	}
    }

  /*
   * initialise the data structures necessary to keep track of target
   * addresses currently being probed
   */
  if(scamper_targets_init() == -1)
    {
      return -1;
    }

  /* initialise the queues that hold the current tasks */
  if(scamper_queue_init() == -1)
    {
      return -1;
    }

  /*
   * initialise the data structures necessary to keep track of output files
   * currently being written to
   */
  if(scamper_outfiles_init() == -1)
    {
      return -1;
    }

  if((file = scamper_outfile_opendef(outfile, outtype)) == NULL)
    {
      return -1;
    }

  /* initialise scamper so it is ready to traceroute and ping */
  if(scamper_do_trace_init() == -1 || scamper_do_ping_init() == -1)
    {
      return -1;
    }

  /* initialise the ICMP response struct */
  memset(&resp, 0, sizeof(resp));

  gettimeofday_wrap(&lastprobe);

  for(;;)
    {
      if(scamper_queue_readycount() > 0 || scamper_addresslist_isready() == 1)
	{
	  /*
	   * if there is something ready to be probed right now, then set the
	   * timeout to go off when it is time to send the next probe
	   */
	  timeval_cpy(&nextprobe, &lastprobe);
	  timeval_add_usec(&nextprobe, wait_between);
	  timeout = &tv;
	}
      else if(scamper_queue_count() > 0)
	{
	  /*
	   * if there isn't anything ready to go right now, but we are
	   * waiting on a response from an earlier probe, then set the timer
	   * to go off when that probe expires.
	   */
	  scamper_queue_waittime(&nextprobe);
	  timeout = &tv;
	}
      else
	{
	  /*
	   * there is nothing to do, so block in select until a file
	   * descriptor supplies an address to probe.
	   */
	  timeout = NULL;
	}

      if(timeout != NULL)
	{
	  gettimeofday_wrap(&tv);
	  diff = timeval_diff_usec(&nextprobe, &tv);

	  memset(&tv, 0, sizeof(tv));
	  if(diff > (int64_t)0) timeval_add_usec(&tv, diff);
	  timeout = &tv;
	}
      else
	{
	  if(exit_when_done != 0 && scamper_addresslist_isempty() == 1)
	    {
	      break;
	    }
	  timeout = NULL;
	}

      /* listen until it is time to send the next probe */
      if(scamper_fds_poll(timeout) == -1)
	{
	  return -1;
	}

      /* take any 'done' traces and output them now */
      while((task = scamper_queue_getdone()) != NULL)
	{
	  if(task->outfile == NULL) task->outfile = file;

	  ((scamper_trace_t *)task->data)->list = scamper_list_use(list);

	  /* write the trace out */
	  task->funcs->write(task);

	  target = scamper_target_find(task->dst);
	  scamper_target_free(target);

	  /* cleanup the task */
	  scamper_task_free(task);
	}

      /*
       * if there is something waiting to be probed, then find out if it is
       * time to probe yet
       */
      if(scamper_queue_readycount() > 0 || scamper_addresslist_isready() == 1)
	{
	  gettimeofday_wrap(&tv);

	  diff = timeval_diff_usec(&tv, &lastprobe);
	  if(diff < -((int64_t)probe_window) || diff > (int64_t)probe_window)
	    {
	      timeval_cpy(&lastprobe, &tv);
	      timeval_add_usec(&lastprobe, wait_between * -1);
	    }

	  /*
	   * when probing at > HZ, scamper might find that select blocks it
	   * from achieving the specified packets per second rate if it sends
	   * one probe per select.  Based on the time spent in the last call
	   * to select, send the necessary number of packets to fill that
	   * window where we sent no packets.
	   */
	  for(;;)
	    {
	      timeval_cpy(&nextprobe, &lastprobe);
	      timeval_add_usec(&nextprobe, wait_between);

	      /* if the next probe is not due to be sent, don't send one */
	      if(timeval_cmp(&nextprobe, &tv) > 0)
		{
		  break;
		}

	      /*
	       * look for an address that we can send a probe to.  if
	       * scamper doesn't have a task on the probe queue waiting
	       * to be probed, then get a fresh task. if there's absolutely
	       * nothing that scamper can probe, then break.
	       */
	      if((task = scamper_queue_select()) == NULL)
		{
		  if((task = scamper_addresslist_get()) == NULL)
		    {
		      break;
		    }

		  if((target = scamper_target_alloc(task->dst)) == NULL)
		    {
		      scamper_task_free(task);
		    }
		  target->task = task;
		}

	      task->funcs->probe(task);
	      timeval_cpy(&lastprobe, &nextprobe);
	    }
	}
    }
  
  return 0; 
}

/*
 * cleanup:
 *
 * be nice to the system and clean up all our mallocs
 */
static void cleanup()
{
#ifndef WITHOUT_PRIVSEP
  scamper_privsep_cleanup();
#endif
  scamper_getsrc_cleanup();
  scamper_rtsock_cleanup();

  scamper_icmp4_close();
  scamper_icmp6_close();
  scamper_udp4_close();
  scamper_udp6_close();

  scamper_addr2mac_cleanup();

  scamper_do_trace_cleanup();
  scamper_do_ping_cleanup();

  scamper_addresslist_cleanup();

  scamper_dl_cleanup();

  if(options & OPT_DAEMON)
    {
      scamper_control_cleanup();
    }

  scamper_outfiles_cleanup();
  scamper_fds_cleanup();

  /* free the address cache, if one was used */
  if(addrcache != NULL)
    {
      scamper_addrcache_free(addrcache);
      addrcache = NULL;
    }

  if(list != NULL)
    {
      scamper_list_free(list);
      list = NULL;
    }

  scamper_queue_cleanup();
  scamper_targets_cleanup();
  scamper_probe_cleanup();

  return;
}

int main(int argc, char *argv[])
{
  pid_t            pid;
  struct sigaction si_sa;

  /*
   * if we are using dmalloc, then we want to get it to register its
   * logdump function to occur after we have used cleanup to free up
   * scamper's core data structures.  this is a dirty hack.
   */
#if defined(DMALLOC)
  free(malloc(1));
#endif

  atexit(cleanup);

  /* set the port to bind to first! */
  pid = getpid();
  sport = (pid & 0x7fff) + 0x8000;

  if(check_options(argc, argv) == -1)
    {
      return -1;
    }

  sigemptyset(&si_sa.sa_mask);
  si_sa.sa_flags   = 0;
  si_sa.sa_handler = scamper_hup;
  if(sigaction(SIGHUP, &si_sa, 0) == -1)
    {
      printerror(errno, strerror, __func__,
		 "could not set sigaction for SIGHUP");
      return -1;
    }

  sigemptyset(&si_sa.sa_mask);
  si_sa.sa_flags   = 0;
  si_sa.sa_handler = scamper_chld;
  if(sigaction(SIGCHLD, &si_sa, 0) == -1)
    {
      printerror(errno, strerror, __func__,
		 "could not set sigaction for SIGCHLD");
      return -1;
    }

  return scamper();
}
