/*
 * scamper_addresslist
 *
 * $Id: scamper_addresslist.c,v 1.69 2005/05/26 09:01:39 mjl Exp $
 *
 * this code deals with storing trace objects with their associated lists.
 *
 * Copyright (C) 2004-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <assert.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper.h"
#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_debug.h"
#include "scamper_file.h"
#include "scamper_outfiles.h"
#include "scamper_target.h"
#include "scamper_task.h"
#include "scamper_addresslist.h"
#include "scamper_fds.h"
#include "scamper_linepoll.h"
#include "scamper_privsep.h"
#include "scamper_addr.h"
#include "scamper_do_trace.h"
#include "scamper_do_ping.h"
#include "utils.h"
#include "mjl_list.h"
#include "mjl_splaytree.h"

/*
 * command
 *
 */
typedef struct command
{
  /* parameter to the command */
  void    *param;

  /* the type of the command */
  uint8_t  type;
} command_t;

#define COMMAND_TRACE 0x00
#define COMMAND_PING  0x01
#define COMMAND_CYCLE 0x02

/*
 * on_hold
 *
 * structure to keep details on the command on hold.
 */
typedef struct on_hold
{
  /* pointer to the task that has the lock on the address */
  scamper_task_t   *task;

  /* pointer to the source list that wants the address to be probed */
  scamper_source_t *source;

  /* pointer to the command that is waiting on the task to complete */
  command_t        *command;

  /* pointer to the node held in the task's on-hold list */
  void             *task_cookie;

  /* pointer to the node held in the source's on-hold list */
  void             *snode;
} on_hold_t;

/*
 * alf
 *
 * structure to keep state when reading an address list file.
 */
typedef struct alf
{
  /* the file descriptor that scamper is monitoring */
  scamper_fd_t       *fd;

  /* back pointer to the source struct that receives the target addresses */
  scamper_source_t   *source;

  /* the types of addresses that come out of this file */
  int                 af;

  scamper_linepoll_t *lp;
} alf_t;

/*
 * source_file
 *
 * structure to keep state on an addresslist 
 */
typedef struct source_file
{
  /* the name of the file */
  char            *filename;

  /* the targets that are held in the address list */
  slist_t         *commands;

  /* if set, the list should cycle when all targets have been probed */
  int              cycle;

  /* if set, the list should be reloaded at the next cycle */
  int              reload;

  /* if set, the list should be refreshed each cycle if it has changed */
  int              autoreload;

  /* the modification time of the file when it was last read */
  struct timespec  mtime;
} source_file_t;

/*
 * source
 *
 * struct that keeps a record of the various address sources that scamper
 * is dealing with at present.
 */
struct scamper_source
{
  /* the details of the list / cycle, included with each trace */
  scamper_list_t    *list;
  scamper_cycle_t   *cycle;

  /*
   * the name of the file that the source gets its target addresses from.
   * if this variable is not null, then the source is 'locked' - i.e. it
   * will refuse to take any other addresses
   */
  source_file_t     *file;

  /* where the data should be written, if the source has a preference */
  scamper_outfile_t *sof;

  /* list of sockaddrs to probe */
  slist_t           *commands;

  /* how many addresses per cycle of the sources clist to probe */
  int                priority;

  /* if set, this source is adhoc */
  int                adhoc;

  /* if not null, the currently open file being parsed for target addresses */
  alf_t             *alf;

  /* a list of files waiting to be parsed for addresses */
  slist_t           *alfq;

  /* list of on-hold structures off the target's list and put on hold */
  dlist_t           *on_hold;

  /* pointer to the list that holds this source, and the node item it has */
  void              *l;
  void              *ln;
};

/* the address cache that the scamper process shares */
extern scamper_addrcache_t *addrcache;

/*
 * keep state to ensure that stdin is only ever supplied as a target
 * address source once.
 */
static int stdin_used = 0;

/*
 * the active list is a circular list of active address sources.  blocked
 * is a doubly linked list of address sources that are blocked on another
 * event that prevents it from supplying an address to probe
 */
static clist_t  *active = NULL;
static dlist_t  *blocked = NULL;

/*
 * source_cur points to the current source that is supplying an address.
 * source_cnt records the number of addresses supplied from the list so
 * far this cycle.
 */
static scamper_source_t *source_cur = NULL;
static int               source_cnt = 0;

/*
 * source_def is the default list where target addresses end up if scamper
 * is not given a specific list to store the addresses in.
 */
static scamper_source_t *source_def = NULL;

/*
 * a splay tree of sources, indexed by name.  the tree does not include
 * source_def.
 */
static splaytree_t *source_tree = NULL;

/*
 * if scamper is restricted to probing a particular family of address family
 * (say IPv4 or IPv6) then this variable will record the family restriction.
 */
static int af = AF_UNSPEC;

/* forward declare alf_alloc, so alf_read can call it */
static alf_t *alf_alloc(int af, int fd, scamper_source_t *source);

static command_t *command_alloc(uint8_t type, void *param)
{
  command_t *command;

  if((command = malloc(sizeof(command_t))) != NULL)
    {
      command->type = type;
      command->param = param;
    }

  return command;
}

static void command_free(command_t *command)
{
  switch(command->type)
    {
    case COMMAND_TRACE:
    case COMMAND_PING:
      scamper_addr_free((scamper_addr_t *)command->param);
      break;

    case COMMAND_CYCLE:
      break;
    }

  free(command);
  return;
}

/*
 * command_probe
 *
 * given the commands list and a target address, create a probe command
 * and put it on the head / tail of the list depending on the value of
 * the pushhead parameter; 0 means tail, 1 means head.
 */
static int command_probe(slist_t *commands, uint8_t type, scamper_addr_t *addr)
{
  command_t *command;

  if((command = command_alloc(type, scamper_addr_use(addr))) == NULL)
    {
      scamper_addr_free(addr);
      return -1;
    }

  if(slist_tail_push(commands, command) == NULL)
    {
      command_free(command);
      return -1;
    }

  return 0;
}

/*
 * command_cycle
 *
 * given the commands list, append a cycle command to it.
 */
static int command_cycle(slist_t *commands)
{
  command_t *command;

  if((command = command_alloc(COMMAND_CYCLE, NULL)) == NULL)
    {
      return -1;
    }

  if(slist_tail_push(commands, command) == NULL)
    {
      command_free(command);
      return -1;
    }

  return 0;
}

/*
 * source_cycle
 *
 * change the cycle id that is passed with each trace
 */
static int source_cycle(scamper_source_t *source)
{
  scamper_cycle_t *cycle;

  /* allocate the new cycle object */
  if((cycle = scamper_cycle_alloc(source->list)) == NULL)
    {
      return -1;
    }

  /* we're onto the next cycle now */
  cycle->id = source->cycle->id + 1;

  /* swap the cycles around */
  scamper_cycle_free(source->cycle);
  source->cycle = cycle;

  return 0;
}

/*
 * source_file_free
 *
 * clean up all the memory associated with a source_file structure
 */
static void source_file_free(source_file_t *file)
{
  command_t *command;

  if(file->commands != NULL)
    {
      while((command = slist_head_pop(file->commands)) != NULL)
	{
	  command_free(command);
	}
      slist_free(file->commands);
    }

  if(file->filename != NULL) free(file->filename);
  free(file);
  return;
}

/*
 * source_file_alloc
 *
 * allocate the structure that keeps track of a address list and its
 * relationship to a source and the underlying file on the disk.
 */
static source_file_t *source_file_alloc(scamper_source_params_t *ssp, int fd)
{
  source_file_t *file;
  struct stat sb;

  /*
   * get the modification time of the file, so scamper can tell when it needs
   * to re-read the file
   */
  if(fstat(fd, &sb) == -1)
    {
      return NULL;
    }

  if((file = malloc_zero(sizeof(source_file_t))) == NULL)
    {
      return NULL;
    }

  if((file->filename = strdup(ssp->filename)) == NULL ||
     (file->commands = slist_alloc()) == NULL)
    {
      source_file_free(file);
      return NULL;
    }

  file->autoreload = ssp->autoreload;
  file->cycle = ssp->cycle;
  memcpy(&file->mtime, &sb.st_mtime, sizeof(file->mtime));

  return file;
}

/*
 * scamper_file_commands_foreach
 *
 * this is a callback function passed to slist_foreach.  its purpose is
 * to make a copy of the command pointed to by node in the slist pointed
 * to by param.
 */
static int source_file_commands_foreach(const void *param, const void *node)
{
  slist_t   *commands = (slist_t *)param;
  command_t *command  = (command_t*)node;

  if(command->type == COMMAND_TRACE || command->type == COMMAND_PING)
    {
      command_probe(commands, command->type, command->param);
    }
  else if(command->type == COMMAND_CYCLE)
    {
      command_cycle(commands);
    }
  else return -1;

  return 0;
}

static int source_file_targets_merge(scamper_source_t *source)
{
  slist_t *list = NULL;
  command_t *command;
  slist_t *commands;

  /* allocate a temporary list to work with */
  if((list = slist_alloc()) == NULL)
    {
      goto err;
    }

  /* construct a command list based on the list of targets */
  commands = source->file->commands;
  if(slist_foreach(commands, list, source_file_commands_foreach) != 0)
    {
      goto err;
    }

  /* append the list to the command list */
  slist_concat(source->commands, list);

  return 0;

 err:
  if(list != NULL)
    {
      while((command = slist_head_pop(list)) != NULL)
	{
	  command_free(command);
	}
      slist_free(list);
    }
  return -1;
}

/*
 * scamper_file_cycle
 *
 * the targets list has run empty and the source is cycling over a specific,
 * named list.  check to see how the source should be re-cycled
 *
 */
static int source_file_cycle(scamper_source_t *source)
{
  if(source_cycle(source) == -1)
    {
      return -1;
    }

  if(source_file_targets_merge(source) == -1)
    {
      return -1;
    }

  return 0;  
}

/*
 * source_cmp
 *
 * provide a sorting function for storing sources in a splay tree
 */
static int source_cmp(const void *a, const void *b)
{
  return strcasecmp(((scamper_source_t *)b)->list->name,
		    ((scamper_source_t *)a)->list->name);
}

/*
 * source_cur
 *
 * advance to the next source to read addresses from, and reset the
 * current count of how many addresses have been returned off the list
 * for this source-cycle
 */
static scamper_source_t *source_next()
{
  void *node;

  if((node = clist_node_next(source_cur->ln)) != source_cur->ln)
    {
      source_cur = clist_node_item(node);
    }

  source_cnt = 0;

  return source_cur;
}

/*
 * source_active_detach
 *
 * detach the source out of the active list.  move to the next source
 * if it is the current_src that is being read from.
 */
static void source_active_detach(scamper_source_t *source)
{
  void *node;

  assert(source->l == active);

  if((node = clist_node_next(source->ln)) != source->ln)
    {
      source_cur = clist_node_item(node);
    }
  else
    {
      source_cur = NULL;
    }

  source_cnt = 0;

  clist_node_pop(active, source->ln);
  source->l = NULL;
  source->ln = NULL;

  return;
}

/*
 * source_blocked_detach
 *
 * detach the source out of the blocked list.
 */
static void source_blocked_detach(scamper_source_t *source)
{
  assert(source->l == blocked);

  dlist_node_pop(blocked, source->ln);
  source->ln = NULL;
  source->l = NULL;
  return;
}

/*
 * source_active
 *
 * put the specified source onto the active list.  if there is no current
 * source being used to read addresses out then make it the current source.
 */
static int source_active_attach(scamper_source_t *source)
{
  if(source->l == active)
    {
      return 0;
    }

  if(source->l != NULL)
    {
      source_blocked_detach(source);
    }

  if((source->ln = clist_push(active, source)) == NULL)
    {
      return -1;
    }
  source->l = active;

  if(source_cur == NULL)
    {
      source_cur = source;
      source_cnt = 0;
    }

  return 0;
}

/*
 * source_blocked_attach
 *
 * put the specified source onto the blocked list.
 */
static int source_blocked_attach(scamper_source_t *source)
{
  if(source->l == blocked)
    {
      return 0;
    }

  if(source->l != NULL)
    {
      source_active_detach(source);
    }

  if((source->ln = dlist_tail_push(blocked, source)) == NULL)
    {
      return -1;
    }
  source->l = blocked;

  return 0;
}

/*
 * source_commands_flush
 *
 * go through the commands list and remove all the command entries
 */
static void source_commands_flush(scamper_source_t *source)
{
  command_t *command;

  if(source == NULL || source->commands == NULL)
    {
      return;
    }

  while((command = slist_head_pop(source->commands)) != NULL)
    {
      command_free(command);
    }

  return;
}

/*
 * source_onhold_flush
 *
 * go through the on-hold list on the source and remove all on-hold entries
 * leave the onhold list intact.
 */
static void source_onhold_flush(scamper_source_t *source)
{
  on_hold_t *onhold;

  if(source == NULL || source->on_hold == NULL)
    {
      return;
    }

  while((onhold = dlist_head_pop(source->on_hold)) != NULL)
    {
      /* detach the onhold structure from the blocking task */
      scamper_task_dehold(onhold->task, onhold->task_cookie);

      free(onhold);
    }

  return;
}

/*
 * source_free
 *
 * clean up the source
 */
static void source_free(scamper_source_t *source)
{
  int *fdp;

  /* detach the source from whatever list it is in */
  if(source->l == active)
    {
      source_active_detach(source);
    }
  else if(source->l == blocked)
    {
      source_blocked_detach(source);
    }

  /*
   * if there is a file currently feeding this source with addresses,
   * then detach this source from the file structure, and leave the file
   * to clean itself up
   */
  if(source->alf != NULL)
    {
      assert(source->alf->source == source);
      source->alf->source = NULL;
    }

  if(source->alfq != NULL)
    {
      while((fdp = slist_head_pop(source->alfq)) != NULL)
	{
	  close(*fdp);
	  free(fdp);
	}
    }

  /*
   * empty the source of any addresses that are currently on hold.
   * this occurs before we flush the commands list as it is possible that
   * an on-hold address might end up in the command list for this source.
   */
  source_onhold_flush(source);
  source_commands_flush(source);

  /* release this structure's hold on the scamper_outfile */
  if(source->sof != NULL) scamper_outfile_refdec(source->sof);

  splaytree_remove(source_tree, source);

  if(source->file != NULL) source_file_free(source->file);
  if(source->alfq != NULL) slist_free(source->alfq);
  if(source->list != NULL) scamper_list_free(source->list);
  if(source->cycle != NULL) scamper_cycle_free(source->cycle);
  if(source->commands != NULL) slist_free(source->commands);
  if(source->on_hold != NULL) dlist_free(source->on_hold);
  free(source);
  return;
}

static void addresslist_unhold(void *cookie)
{
  on_hold_t        *on_hold = (on_hold_t *)cookie;
  scamper_source_t *source  = on_hold->source;
  command_t        *command = on_hold->command;

  /* remove the source's copy of the on-hold structure */
  dlist_node_pop(source->on_hold, on_hold->snode);
  free(on_hold);

  /* push the command onto the front of the list */
  slist_head_push(source->commands, command);

  /* put the source on the active list */
  source_active_attach(source);

  return;
}

static int addresslist_onhold(scamper_source_t *source, scamper_task_t *task,
			      command_t *command)
{
  on_hold_t *on_hold;

  if((on_hold = malloc_zero(sizeof(on_hold_t))) == NULL)
    {
      return -1;
    }

  on_hold->task = task;
  on_hold->source = source;
  on_hold->command = command;

  if((on_hold->snode = dlist_tail_push(source->on_hold, on_hold)) == NULL)
    {
      goto cleanup;
    }

  if((on_hold->task_cookie =
      scamper_task_onhold(task, on_hold, addresslist_unhold)) == NULL)
    {
      goto cleanup;
    }

  return 0;

 cleanup:

  if(on_hold->snode != NULL) dlist_node_pop(source->on_hold, on_hold->snode);
  free(on_hold);

  return -1;
}

#if 0
static scamper_trace_t *trace_init(scamper_source_t *ss, scamper_addr_t *sa)
{
  scamper_trace_state_t *sts;
  scamper_trace_t *trace;

  if((trace = scamper_trace_alloc()) == NULL)
    {
      return NULL;
    }

  trace->dst = scamper_addr_use(sa);
  trace->list = scamper_list_use(ss->list);

  if(ss->cycle != NULL)
    {
      trace->cycle = scamper_cycle_use(ss->cycle);

      trace->flags    = trace->cycle->flags;
      trace->attempts = trace->cycle->attempts;
      trace->hoplimit = trace->cycle->hoplimit;
      trace->sport    = trace->cycle->sport;
      trace->dport    = trace->cycle->dport;
    }

  if(scamper_trace_state_alloc(trace) == -1)
    {
      scamper_trace_free(trace);
      return NULL;
    }

  sts = trace->param;
  if(ss->sof != NULL)
    {
      sts->outfile = scamper_outfile_refinc(ss->sof);
    }

  return trace;
}
#endif

/*
 * scamper_addresslist_get
 *
 * return the next address to trace.
 *
 * note that if this function returns null, it does not necessarily infer
 * that there are no more addresses to trace; if there are addresses
 * currently on-hold that are blocked on other traces in the active window
 * and no other traces to enqueue then this function will return null.
 */
scamper_task_t *scamper_addresslist_get()
{
  scamper_source_t *source;
  scamper_target_t *target;
  scamper_task_t *task;
  scamper_addr_t *addr;
  command_t *command;

  while((source = source_cur) != NULL)
    {
      assert(source_cnt < source->priority);

      while((command = slist_head_pop(source->commands)) != NULL)
	{
	  if(command->type == COMMAND_TRACE || command->type == COMMAND_PING)
	    {
	      addr = command->param;

	      /*
	       * if this command is blocked, then put it on hold and
	       * continue with the next available
	       */
	      if((target = scamper_target_find(addr)) != NULL)
		{
		  addresslist_onhold(source, target->task, command);
		  continue;
		}

	      if(command->type == COMMAND_TRACE)
		{
		  task = scamper_do_trace_alloctask(addr);
		}
	      else
		{
		  task = scamper_do_ping_alloctask(addr);
		}

	      /* don't need the command any longer */
	      command_free(command);

	      if(source->sof != NULL)
		{
		  task->outfile = scamper_outfile_refinc(source->sof);
		}

	      /* scamper is returning another address from this source */
	      if(++source_cnt == source->priority)
		{
		  /*
		   * move to the next address source if this address
		   * source has filled its quota
		   */
		  source = source_next();
		}

	      return task;
	    }
	  else if(command->type == COMMAND_CYCLE)
	    {
	      source_cycle(source);
	      command_free(command);
	    }
	}

      /*
       * if we get to here, the previous target list could not supply an
       * address.
       */
      assert(slist_count(source->commands) == 0);

      /*
       * if there is still the possibility of getting an address
       * out of the source then put the list in the blocked list.
       */
      if(source->alf != NULL || dlist_count(source->on_hold) > 0) 
	{
	  source_blocked_attach(source);
	}
      /*
       * if the source is the default list, then we merely detach it
       * from the active list and leave it 'dangling' in that it is not
       * in either of the active or blocked lists.
       */
      else if(source == source_def)
	{
	  source_active_detach(source);
	}
      /*
       * if the source is an 'adhoc' source where addresses are supplied
       * by some external process, then put the source onto the blocked
       * list, as there may be more addresses on the way
       */
      else if(source->adhoc == 1)
	{
	  source_blocked_attach(source);
	}
      /*
       * if we are working with a file and cycling it, then do the
       * cycle operation now.
       */
      else if(source->file != NULL && source->file->cycle == 1)
	{
	  if(source_file_cycle(source) == -1)
	    {
	      source_free(source);
	    }
	}
      /*
       * if none of the above tests hold true, then the source is no longer
       * required, so remove it.
       */
      else
	{
	  source_free(source);
	}
    }

  /* there shouldn't be any active address list sources available ... */
  assert(clist_count(active) == 0);

  return NULL;
}

/*
 * scamper_addresslist_isempty
 *
 * return to the caller if the address list is empty or not.
 */
int scamper_addresslist_isempty()
{
  /*
   * if there are either active or blocked address list sources, the list
   * can't be empty
   */
  if(clist_count(active) > 0 || dlist_count(blocked) > 0)
    {
      return 0;
    }

  /*
   * if the default source is waiting for more addresses then the list
   * can't be empty
   */
  if(source_def->alf != NULL)
    {
      return 0;
    }

  return 1;
}

/*
 * scamper_addresslist_isready
 *
 * return to the caller if the address list appears to be able to supply
 * an address for probing.
 */
int scamper_addresslist_isready()
{
  if(source_cur != NULL)
    {
      return 1;
    }

  return 0;
}

static struct sockaddr *resolve(const char *ipstr, const int af)
{
  struct sockaddr *addr;
  struct addrinfo  hints, *res, *res0;
  int error;

  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_flags    = AI_NUMERICHOST;
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_protocol = IPPROTO_UDP;
  hints.ai_family   = af;
 
  /* try and resolve the _numeric_ address */
  if((error = getaddrinfo(ipstr, NULL, &hints, &res0)) != 0 || res0 == NULL)
    {
      printerror(error, gai_strerror_wrap, __func__,
		 "could not resolve %s", ipstr);
      return NULL;
    }

  /*
   * scan for the first usable address because we used PF_UNSPEC.
   * INET or INET6 protocol families are usable
   */
  for(res = res0; res != NULL; res = res->ai_next)
    {
      if(res->ai_family == PF_INET || res->ai_family == PF_INET6)
	{
	  break;
	}
    }

  if(res == NULL)
    {
      freeaddrinfo(res0);
      return NULL;
    }

  /* make a copy of the address and return */
  if((addr = malloc(res->ai_addrlen)) == NULL)
    {
      printerror(errno, strerror, __func__, "could not malloc *addr");
      freeaddrinfo(res0);
      return NULL;
    }

  memcpy(addr, res->ai_addr, res->ai_addrlen);
  freeaddrinfo(res0);

  return addr;
}

/*
 * source_addaddr
 *
 * resolve the address and put it on the appropriate list.
 */
static int source_addaddr(scamper_source_t *source, char *str)
{
  struct sockaddr *sa = NULL;
  scamper_addr_t *addr = NULL;
  void *va;
  int type;

  assert(str != NULL);

  /*
   * turn the string representation of the address into a number, and then
   * extract the raw address from the sockaddr
   */
  if((sa = resolve(str, af)) == NULL)
    {
      goto err;
    }

  if(sa->sa_family == AF_INET)
    {
      type = SCAMPER_ADDR_TYPE_IPV4;
      va = &((struct sockaddr_in *)sa)->sin_addr;
    }
  else if(sa->sa_family == AF_INET6)
    {
      type = SCAMPER_ADDR_TYPE_IPV6;
      va = &((struct sockaddr_in6 *)sa)->sin6_addr;
    }
  else goto err;

  /*
   * get a pointer to the address
   */
  if((addr = scamper_addrcache_get(addrcache, type, va)) == NULL)
    {
      goto err;
    }

  /*
   * decide the list this address should end up on.  if there is a file
   * that is being managed for rotation and cycling, then it should go onto
   * that list first.  otherwise, just append the address to the targets
   */
  if(source->file == NULL)
    {
      if(command_probe(source->commands, COMMAND_TRACE, addr) == -1)
	{
	  goto err;
	}

      /*
       * we just loaded the address directly onto the source's targets.
       * make sure that the source is on the active list so the command can
       * be handled
       */
      source_active_attach(source);
    }
  else
    {
      /* push the address onto the list */
      if(slist_tail_push(source->file->commands, addr) == NULL)
	{
	  goto err;
	}
    }

  scamper_addr_free(addr);
  free(sa);
  return 0;

 err:
  if(addr != NULL) scamper_addr_free(addr);
  if(sa != NULL) free(sa);
  return -1;
}

/*
 * alf_free
 *
 * free up all resources related to an address-list-file
 */
static void alf_free(alf_t *alf, int feedlastline)
{
  int fd;

  if(alf->lp != NULL)
    {
      scamper_linepoll_free(alf->lp, feedlastline);
    }

  if(alf->fd != NULL)
    {
      fd = scamper_fds_fd(alf->fd);
      scamper_fds_del(alf->fd);
      close(fd);
    }

  free(alf);
  return;
}

/*
 * alf_read_line
 *
 * this callback receives a single line per call, which should contain an
 * address in string form.  it resolves any address in the line, and then
 * puts it in the appropriate list.  it will go onto the source's target
 * list if scamper is not managing list rotation.  otherwise it will go
 * onto the file's target list.
 */
static int alf_read_line(void *param, uint8_t *buf, int len)
{
  alf_t *alf;

  /* make sure the string contains only printable characters */
  if(string_isprint((char *)buf, len) == 0)
    {
      return 0;
    }

  alf = (alf_t *)param;

  /* null terminate at these characters */
  string_nullterm((char *)buf, " \r\t#");

  /* make sure the line isn't blank or a comment line */
  if(buf[0] == '\0' || buf[0] == '#')
    {
      return 0;
    }

  /* resolve the address */
  if(source_addaddr(alf->source, (char *)buf) == -1)
    {
      return -1;
    }

  return 0;
}

/*
 * alf_read
 *
 * this callback is called when the fd has something ready to read.
 */
static void alf_read(const int fd, void *param)
{
  scamper_source_t *source;
  alf_t *alf;
  uint8_t buf[1024];
  ssize_t rc;
  int *fdp;

  /* address list file */
  alf = (alf_t *)param;
  assert(scamper_fds_fd(alf->fd) == fd);

  /* if the source has been yanked from the address list, then tidy up */
  if((source = alf->source) == NULL)
    {
      alf_free(alf, 0);
      return;
    }

  /* try and read more from the file */
  if((rc = read(fd, buf, sizeof(buf))) > 0)
    {
      scamper_linepoll_handle(alf->lp, buf, rc);
    }
  else if(rc == 0)
    {
      scamper_debug(__func__, "eof on fd %d", fd);

      /*
       * if we've got to the end of the file, remove it from the list of
       * monitored file descriptors and close the file.
       */
      source->alf = NULL;
      alf_free(alf, 1);

      if(source->adhoc == 1)
	{
	  if((fdp = slist_head_pop(source->alfq)) != NULL)
	    {
	      source->alf = alf_alloc(af, *fdp, source);
	      free(fdp);
	    }
	}
      else
	{
	  assert(source->file != NULL);

	  if(source_file_targets_merge(source) == -1)
	    {
	      return;
	    }

	  source_active_attach(source);
	}
    }
  else
    {
      assert(rc == -1);
      if(errno != EAGAIN)
	{
	  printerror(errno, strerror, __func__, "read failed");
	}
    }

  return;
}

static alf_t *alf_alloc(int af, int fd, scamper_source_t *source)
{
  alf_t *alf;

  if((alf = malloc_zero(sizeof(alf_t))) == NULL)
    {
      return NULL;
    }

  if((alf->fd = scamper_fds_read_add(fd, alf_read, alf)) == NULL)
    {
      free(alf);
      return NULL;
    }

  if((alf->lp = scamper_linepoll_alloc(alf_read_line, alf)) == NULL)
    {
      scamper_fds_del(alf->fd);
      free(alf);
      return NULL;
    }

  alf->source = source;
  alf->af = af;

  return alf;
}

/*
 * source_openfile
 *
 * open the filename specified.  return the fd to the caller.
 */
static int source_openfile(char *filename)
{
  int fd;

  if(strcmp(filename, "-") == 0)
    {
      if(stdin_used == 0)
	{
	  if((fd = fileno(stdin)) == -1)
	    {
	      return -1;
	    }
	  stdin_used = 1;
	}
      else
	{
	  return -1;
	}
    }
  else
    {
#if defined(WITHOUT_PRIVSEP)
      fd = open(filename, O_RDONLY);
#else
      fd = scamper_privsep_openfile(filename, O_RDONLY, 0);
#endif
      /* make sure the fd is valid, otherwise bail */
      if(fd == -1)
	{
	  return -1;
	}
    }

  /* we don't want reads on this file descriptor to block */
  if(fcntl_set(fd, O_NONBLOCK) == -1)
    {
      close(fd);
      return -1;
    }

  return fd;
}

const char *scamper_source_getname(scamper_source_t *source)
{
  return source->list->name;
}

/*
 * scamper_addresslist_addarray
 *
 * take the ip array, resolve each item, and put them into the source
 * specified
 */
int scamper_addresslist_addarray(scamper_source_t *ss, char **iparray, int len)
{
  int i;

  assert(iparray != NULL);
  assert(len > 0);

  if(ss == NULL) ss = source_def;

  if(ss->adhoc != 1)
    {
      return -1;
    }

  for(i=0; i<len; i++)
    {
      assert(iparray[i] != NULL);
      if(source_addaddr(ss, iparray[i]) == -1)
	{
	  return -1;
	}
    }

  return 0;
}

/*
 * scamper_addresslist_addaddr
 *
 * take the address' string, resolve it, and put it on the source specified.
 */
int scamper_addresslist_addaddr(scamper_source_t *ss, char *str)
{
  return scamper_addresslist_addarray(ss, &str, 1);
}

/*
 * scamper_addresslist_addfile
 *
 * queue the contents of the specified file for reading on an ad-hoc
 * list.
 */
int scamper_addresslist_addfile(scamper_source_t *source, char *filename)
{
  int fd = -1, *fdp = NULL;

  if(source == NULL) source = source_def;

  if(source->adhoc != 1)
    {
      return -1;
    }

  if((fd = source_openfile(filename)) == -1)
    {
      return -1;
    }

  if(source->alf == NULL)
    {
      if((source->alf = alf_alloc(af, fd, source)) == NULL)
	{
	  goto err;
	}
    }
  else
    {
      if((fdp = malloc(sizeof fd)) == NULL)
	{
	  goto err;
	}
      *fdp = fd;

      if((slist_tail_push(source->alfq, fdp)) == NULL)
	{
	  goto err;
	}
    }

  return 0;

 err:
  if(fdp != NULL) free(fdp);
  if(fd != -1) close(fd);
  return -1;
}

/*
 * scamper_addresslist_addsource
 *
 * link an address list into rotation.
 */
scamper_source_t *scamper_addresslist_addsource(scamper_source_params_t *ssp)
{
  scamper_source_t *source;
  int fd = -1;
  int is_default = 0;

  if(strcasecmp(ssp->name, "default") == 0)
    {
      is_default = 1;
    }

  /*
   * don't let a source named default get into the tree if there already
   * is a default source
   */
  if(is_default && source_def != NULL)
    {
      return NULL;
    }

  /* if a source is not 'ad-hoc', then it has to specify a file */
  if(ssp->adhoc == 0 && ssp->filename == NULL)
    {
      return NULL;
    }

  if((source = malloc_zero(sizeof(scamper_source_t))) == NULL)
    {
      return NULL;
    }

  source->priority = ssp->priority;
  source->adhoc = ssp->adhoc;

  /* if we have been passed an output file to target, then use it */
  if(ssp->sof != NULL)
    {
      source->sof = scamper_outfile_refinc(ssp->sof);
    }

  /* the targets list is an ordered list of addresses to probe */
  if((source->commands = slist_alloc()) == NULL)
    {
      goto err;
    }

  /* the on-hold list is a collection of on_hold records for this source */
  if((source->on_hold = dlist_alloc()) == NULL)
    {
      goto err;
    }

  /*
   * this structure defines the source where a series of traces came
   * from.  it is shared by the source and by the traces that reference
   * it.
   */
  if((source->list = scamper_list_alloc(ssp->id,ssp->name,ssp->descr)) == NULL)
    {
      goto err;
    }

  /*
   * if the source being created isn't the default source, then the source
   * will understand and respect cycles.
   */
  if(is_default == 0)
    {
      if((source->cycle = scamper_cycle_alloc(source->list)) == NULL)
	{
	  goto err;
	}
      source->cycle->id = 1;
    }

  /* if there is a file to initialise the source with, then open it now */
  if(ssp->filename != NULL)
    {
      if((fd = source_openfile(ssp->filename)) == -1)
	{
	  goto err;
	}

      if((source->alf = alf_alloc(af, fd, source)) == NULL)
	{
	  goto err;
	}
    }

  /*
   * if the source is an adhoc list, then it can queue files of target
   * addresses.  otherwise, it can only operate on one file, which we
   * maintain the target addresses for.
   */
  if(source->adhoc == 1)
    {
      if((source->alfq = slist_alloc()) == NULL)
	{
	  goto err;
	}
    }
  else
    {
      if((source->file = source_file_alloc(ssp, fd)) == NULL)
	{
	  goto err;
	}
    }

  /*
   * put the source into the splaytree, where it can be found quickly
   * when needed.
   */
  if(splaytree_insert(source_tree, source) != 1)
    {
      goto err;
    }

  /*
   * put the source in the blocked list, as there currently is no
   * addresses ready to be probed at this time.
   */
  if(is_default == 0 && source_blocked_attach(source) == -1)
    {
      goto err;
    }

  return source;

 err:
  if(fd != -1) close(fd);
  source_free(source);
  return NULL;
}

/*
 * scamper_addresslist_getsource
 *
 * look for the address list specified by the name passed
 */
scamper_source_t *scamper_addresslist_getsource(char *name)
{
  scamper_source_t findme;
  scamper_list_t list;
  list.name = name;
  findme.list = &list;
  return (scamper_source_t *)splaytree_find(source_tree, &findme);
}

/*
 * scamper_addressslist_delsource
 *
 * remove the source from the list of available address sources if
 * possible
 */
int scamper_addresslist_delsource(scamper_source_t *source)
{
  /* the default source cannot be removed */
  if(strcasecmp(source->list->name, "default") == 0)
    {
      return -1;
    }

  /*
   * do not remove a source that has targets queued, or has the possibility
   * of having targets queued
   */
  if(slist_count(source->commands) > 0 || source->alf != NULL ||
     dlist_count(source->on_hold) > 0 || source->file != NULL)
    {
      return -1;
    }

  source_free(source);
  return 0;
}

/*
 * scamper_addresslist_cyclesource
 *
 * if the source is an adhoc source [and it isn't the default source]
 * then append a cycle command to the source.
 */
int scamper_addresslist_cyclesource(scamper_source_t *source)
{
  if(source->adhoc == 0 || source == source_def)
    {
      return -1;
    }

  if(command_cycle(source->commands) == -1)
    {
      return -1;
    }

  return 0;
}

/*
 * scamper_addresslist_empty
 *
 * remove all addresses and sources in the address list
 */
void scamper_addresslist_empty()
{
  scamper_source_t *source;

  /*
   * for each source, go through and empty the lists, close the files, and
   * leave the list of sources available to read from empty.
   */
  while((source = dlist_tail_pop(blocked)) != NULL)
    {
      source->l = NULL; source->ln = NULL;
      source_free(source);
    }

  while((source = clist_pop(active)) != NULL)
    {
      source->l = NULL; source->ln = NULL;
      source_free(source);
    }

  return;
}

void scamper_addresslist_foreach(void *param,
				 int (*func)(void *p, scamper_source_t *src))
{
  splaytree_inorder(source_tree, (splaytree_inorder_t)func, param);
  return;
}

int scamper_addresslist_init()
{
  scamper_source_params_t ssp;

  if(scamper_option_ipv4())      af = PF_INET;
  else if(scamper_option_ipv6()) af = PF_INET6;
  else                           af = PF_UNSPEC;

  if((active = clist_alloc()) == NULL)
    {
      return -1;
    }

  if((blocked = dlist_alloc()) == NULL)
    {
      return -1;
    }

  if((source_tree = splaytree_create(source_cmp)) == NULL)
    {
      return -1;
    }

  /* create the default address list source and put it into rotation */
  memset(&ssp, 0, sizeof(ssp));
  ssp.name     = "default";
  ssp.priority = 1;
  ssp.adhoc    = 1;

  if((source_def = scamper_addresslist_addsource(&ssp)) == NULL)
    {
      return -1;
    }

  return 0;  
}

void scamper_addresslist_cleanup()
{
  if(source_def != NULL)
    {
      source_free(source_def);
      source_def = NULL;
    }

  if(source_tree != NULL)
    {
      splaytree_destroy(source_tree, NULL);
      source_tree = NULL;
    }

  if(blocked != NULL)
    {
      dlist_free(blocked);
      blocked = NULL;
    }

  if(active != NULL)
    {
      clist_free(active);
      active = NULL;
    }

  return;
}
