/*
 * scamper_do_ping.c
 *
 * $Id: scamper_do_ping.c,v 1.22 2005/05/26 09:01:39 mjl Exp $
 *
 * Copyright (C) 2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "scamper.h"
#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_ping.h"
#include "scamper_getsrc.h"
#include "scamper_icmp_resp.h"
#include "scamper_dl.h"
#include "scamper_probe.h"
#include "scamper_task.h"
#include "scamper_queue.h"
#include "scamper_debug.h"
#include "utils.h"

/* the callback functions registered with the ping task */
static scamper_task_funcs_t ping_funcs;

/* ICMP ping probes are marked with the process' ID */
static pid_t pid;

/* packet buffer for generating the payload of an ICMP packet */
static char *pktbuf     = NULL;
static int   pktbuf_len = 0;

/* address cache used to avoid reallocating the same address multiple times */
extern scamper_addrcache_t *addrcache;

typedef struct ping_probe
{
  struct timeval tx;
  uint16_t       seq;
} ping_probe_t;

typedef struct ping_state
{
  ping_probe_t **probes;
  uint16_t       seq_min, seq_cur, seq_max;
} ping_state_t;

/*
 * ping_abort
 *
 * some internal consistency check failed
 */
static void ping_abort(scamper_task_t *task)
{
  scamper_task_free(task);
  return;
}

static void ping_stop(scamper_task_t *task, uint8_t reason, uint8_t data)
{
  scamper_ping_t *ping = task->data;

  ping->stop_reason = reason;
  ping->stop_data   = data;

  scamper_queue_done(task->queue, scamper_holdtime_get()*1000);

  return;  
}

static void ping_handleerror(scamper_task_t *task, int error)
{
  ping_stop(task, SCAMPER_PING_STOP_ERROR, error);
  return;
}

/*
 * do_ping_probe
 *
 * it is time to send a probe for this task.  figure out the form of the
 * probe to send, and then send it.
 */
static int do_ping_probe(scamper_task_t *task)
{
  scamper_ping_t  *ping  = task->data;
  ping_state_t    *state = task->state;
  ping_probe_t    *pp = NULL;
  scamper_probe_t  probe;
  uint8_t          proto, type;
  size_t           payload_len;
  int              i;

  if(ping->dst->type == SCAMPER_ADDR_TYPE_IPV4)
    {
      proto = IPPROTO_ICMP;
      type = ICMP_ECHO;
      payload_len = ping->size - sizeof(struct ip) - (1 + 1 + 2 + 2 + 2);
    }
  else if(ping->dst->type == SCAMPER_ADDR_TYPE_IPV6)
    {
      proto = IPPROTO_ICMPV6;
      type = ICMP6_ECHO_REQUEST;
      payload_len = ping->size - sizeof(struct ip6_hdr) - (1 + 1 + 2 + 2 + 2);
    }
  else
    {
      ping_abort(task);
      return -1;
    }

  /* make sure the global pktbuf is big enough for the probe we send */
  if(pktbuf_len < payload_len)
    {
      pktbuf = realloc_wrap(pktbuf, payload_len, __func__, "couldn't realloc");
      pktbuf_len = payload_len;
    }

  probe.pr_src       = ping->src;
  probe.pr_dst       = ping->dst;
  probe.pr_dl        = NULL;
  probe.pr_dl_hdr    = NULL;
  probe.pr_dl_size   = 0;
  probe.pr_ttl       = ping->ttl;
  probe.pr_proto     = proto;
  probe.pr_icmp_type = type;
  probe.pr_icmp_code = 0;
  probe.pr_icmp_id   = pid & 0xffff;
  probe.pr_icmp_seq  = state->seq_cur;
  probe.pr_data      = pktbuf;
  probe.pr_len       = payload_len;

  /* if the ping has to hold some pattern, then generate it now */  
  if(ping->pad_bytes == NULL)
    {
      memset(pktbuf, 0, payload_len);
    }
  else
    {
      for(i=0; i + ping->pad_len < payload_len; i += ping->pad_len)
	{
	  memcpy(pktbuf+i, ping->pad_bytes, ping->pad_len);
	}
      memcpy(pktbuf+i, ping->pad_bytes, payload_len - i);
    }

  /*
   * allocate a ping probe state record before we try and send the probe
   * as there is no point sending something into the wild that we can't
   * record
   */
  if((pp = malloc(sizeof(ping_probe_t))) == NULL)
    {
      ping_handleerror(task, errno);
      goto err;
    }

  if(scamper_probe(&probe) == -1)
    {
      ping_handleerror(task, probe.pr_errno);
      goto err;
    }

  /* fill out the details of the probe sent */
  pp->seq = state->seq_cur;
  timeval_cpy(&pp->tx, &probe.pr_tx);

  /* record the probe in the probes table */
  state->probes[state->seq_cur - state->seq_min] = pp;

  /* we've sent this sequence number now, so move to the next one */
  state->seq_cur++;

  /* re-queue the ping task */
  scamper_queue_wait(task->queue, ping->wait * 1000);

  return 0;

 err:
  if(pp != NULL) free(pp);
  return -1;
}

static int do_ping_handle_icmp(scamper_task_t *task, scamper_icmp_resp_t *ir)
{
  scamper_ping_t       *ping  = task->data;
  ping_state_t         *state = task->state;
  scamper_ping_reply_t *reply = NULL;
  scamper_ping_reply_t *temp;
  ping_probe_t         *probe;
  uint16_t              seq;
  void                 *addr;
  int                   type;

  if(ir->ir_af == AF_INET) type = SCAMPER_ADDR_TYPE_IPV4;
  else if(ir->ir_af == AF_INET6) type = SCAMPER_ADDR_TYPE_IPV6;
  else return -1;

  /* if this is an echo reply packet, then check the id and sequence */
  if(SCAMPER_ICMP_RESP_IS_ECHO_REPLY(ir))
    {
      /* if the response is not for us, then move on */
      if(ir->ir_icmp_id != (pid & 0xffff) ||
	 ir->ir_icmp_seq < state->seq_min ||
	 ir->ir_icmp_seq > state->seq_max)
	{
	  return 0;
	}

      seq = ir->ir_icmp_seq - state->seq_min;
    }
  /* if this is an ICMP response to an echo req, then check id and sequence */
  else if(SCAMPER_ICMP_RESP_INNER_IS_SET(ir) &&
	  SCAMPER_ICMP_RESP_INNER_IS_ICMP_ECHO_REQ(ir))
    {
      if(ir->ir_inner_icmp_id != (pid & 0xffff) ||
	 ir->ir_inner_icmp_seq < state->seq_min ||
	 ir->ir_inner_icmp_seq > state->seq_max)
	{
	  return 0;
	}

      seq = ir->ir_inner_icmp_seq - state->seq_min;
    }
  else return 0;

  /*
   * if the sequence number was in our range, but we have no record of the
   * probe, then just ignore the response
   */
  if((probe = state->probes[seq]) == NULL)
    {
      return 0;
    }

  /* allocate a reply structure for the response */
  if((reply = scamper_ping_reply_alloc()) == NULL)
    {
      goto err;
    }

  /* figure out where the response came from */
  addr = scamper_icmp_resp_src(ir);
  if((reply->addr = scamper_addrcache_get(addrcache, type, addr)) == NULL)
    {
      goto err;
    }

  /* put together details of the reply */
  reply->icmp_type  = ir->ir_icmp_type;
  reply->icmp_code  = ir->ir_icmp_code;
  reply->reply_ttl  = ir->ir_ip_ttl;
  reply->reply_size = ir->ir_ip_size;
  timeval_rtt(&reply->rtt, &probe->tx, &ir->ir_rx);

  /* put the reply into the ping table */
  if((temp = ping->ping_replies[seq]) == NULL)
    {
      ping->ping_replies[seq] = reply;
    }
  else
    {
      while(temp->next != NULL)
	{
	  temp = temp->next;
	}
      temp->next = reply;
    }

  return 0;

 err:
  if(reply != NULL) scamper_ping_reply_free(reply);
  return -1;
}

static int do_ping_handle_dl(scamper_task_t *task, scamper_dl_rec_t *dl_rec)
{
  return -1;
}

/*
 * do_ping_handle_timeout
 *
 * the ping object expired on the pending queue
 * that means it is either time to send the next probe, or write the
 * task out
 */
static int do_ping_handle_timeout(scamper_task_t *task)
{
  ping_state_t *state = task->state;

  if(state->seq_cur == state->seq_max)
    {
      ping_stop(task, SCAMPER_PING_STOP_COMPLETED, 0);
    }

  return 0;
}

static int do_ping_write(scamper_task_t *task)
{
  return 0;
}

static void do_ping_free(scamper_task_t *task)
{
  scamper_ping_t *ping;
  ping_state_t *state;
  int i;

  /* free any ping data collected */
  if((ping = task->data) != NULL)
    {
      scamper_ping_free(ping);
    }

  if((state = task->state) != NULL)
    {
      if(state->probes != NULL)
	{
	  for(i=0; i<state->seq_max - state->seq_min; i++)
	    {
	      if(state->probes[i] != NULL)
		{
		  free(state->probes[i]);
		}
	    }

	  free(state->probes);
	}

      free(state);
    }

  return;
}

scamper_task_t *scamper_do_ping_alloctask(scamper_addr_t *dst)
{
  scamper_task_t *task;
  scamper_ping_t *ping;
  ping_state_t   *state;
  size_t          size;

  /* firstly, allocate the task structure */
  if((task = scamper_task_alloc(dst, &ping_funcs)) == NULL)
    {
      goto err;
    }

  /* secondly, allocate and initialise the ping data structure */
  if((task->data = scamper_ping_alloc()) == NULL)
    {
      goto err;
    }
  ping = task->data;

  /* determine the source address used for sending probes */
  if((ping->src = scamper_getsrc(dst)) == NULL)
    {
      goto err;
    }

  /* initialise the rest of the ping data structure */
  ping->dst = scamper_addr_use(dst);
  ping->count = 10;
  ping->size  = 56;
  ping->wait  = 1;
  ping->ttl   = 64;

  /* allocate the necessary state to keep track of probes */
  if((task->state = malloc_zero(sizeof(ping_state_t))) == NULL)
    {
      goto err;
    }
  state = task->state;
  size = ping->count * sizeof(ping_probe_t *);
  if((state->probes = malloc_zero(size)) == NULL)
    {
      goto err;
    }
  state->seq_max = ping->count + state->seq_min;

  gettimeofday_wrap(&ping->start);

  return task;

 err:
  if(task != NULL) scamper_task_free(task);
  return NULL;
}

void scamper_do_ping_cleanup()
{
  if(pktbuf != NULL)
    {
      free(pktbuf);
      pktbuf = NULL;
    }

  return;
}

int scamper_do_ping_init()
{
  ping_funcs.probe                  = do_ping_probe;
  ping_funcs.handle_icmp            = do_ping_handle_icmp;
  ping_funcs.handle_dl              = do_ping_handle_dl;
  ping_funcs.handle_timeout         = do_ping_handle_timeout;
  ping_funcs.write                  = do_ping_write;
  ping_funcs.task_free              = do_ping_free;

  pid = getpid();

  return 0;
}
