/*
 * scamper_fds: manage events for file descriptors
 *
 * $Id: scamper_fds.c,v 1.14 2005/05/03 22:48:31 mjl Exp $
 *
 *          Matthew Luckie
 * 
 *          Supported by:
 *           The University of Waikato
 *           NLANR Measurement and Network Analysis
 *           CAIDA
 *           The WIDE Project
 *
 * Copyright (C) 2004 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/param.h>

#if 0
#if defined(__FreeBSD__) && __FreeBSD_version >= 410000
#define HAVE_KQUEUE
#elif defined(__OpenBSD__) && OpenBSD >= 200105
#define HAVE_KQUEUE
#else
#define HAVE_SELECT
#endif
#endif

#define HAVE_SELECT

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>

#if defined(HAVE_KQUEUE)
#include <sys/event.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper_debug.h"
#include "scamper_fds.h"
#include "utils.h"
#include "mjl_list.h"

struct scamper_fd
{
  int              fd;    /* the file descriptor being polled */
  scamper_fd_cb_t  cb;    /* callback registered for event */
  void            *param; /* parameter to pass to the callback */
  dlist_t         *list;  /* the list fd is in: read_fds or write_fds */
  dlist_node_t    *node;  /* node in the list */
};

static dlist_t *read_fds  = NULL;
static dlist_t *write_fds = NULL;

#if defined(HAVE_SELECT)
static int        nfds = -1;
#endif

#if defined(HAVE_KQUEUE)
static struct kevent *kevlist = NULL;
static int            kevlistlen = 0;
static int            kq = -1;
#endif

#if defined(HAVE_KQUEUE)
static int scamper_fds_add2(scamper_fd_t *fdn)
{
  struct kevent kev;
  int len;
  short filter;

  if(fdn->list == read_fds) filter = EVFILT_READ;
  else filter = EVFILT_WRITE;

  EV_SET(&kev, fdn->fd, filter, EV_ADD, 0, 0, fdn);

  if(kevent(kq, &kev, 1, NULL, 0, NULL) == -1)
    {
      printerror(errno, strerror, __func__, "could not kevent");
      return -1;
    }

  if((len = dlist_count(read_fds) + dlist_count(write_fds)) >= kevlistlen)
    {
      kevlistlen = len + 8;
      kevlist = realloc_wrap(kevlist, kevlistlen * sizeof(struct kevent),
			     __func__, "could not realloc list");
    }

  return 0;
}

static int scamper_fds_del2(scamper_fd_t *fdn)
{
  struct kevent kev;
  short filter;

  if(fdn->list == read_fds) filter = EVFILT_READ;
  else filter = EVFILT_WRITE;

  EV_SET(&kev, fdn->fd, filter, EV_DELETE, 0, 0, fdn);

  if(kevent(kq, &kev, 1, NULL, 0, NULL) == -1)
    {
      printerror(errno, strerror, __func__, "could not kevent");
      return -1;
    }

  return 0;
}

int scamper_fds_poll(struct timeval *timeout)
{
  struct timespec ts;
  struct kevent *kev;
  scamper_fd_t  *fdn;
  int i, c;

  ts.tv_sec  = timeout->tv_sec;
  ts.tv_nsec = timeout->tv_usec * 1000;

  if((c = kevent(kq, NULL, 0, kevlist, kevlistlen, &ts)) == -1)
    {
      printerror(errno, strerror, __func__, "kevent failed");
      return -1;
    }

  for(i=0; i<c; i++)
    {
      kev = &kevlist[i];
      fdn = kev->udata;

      assert(kev->ident == fdn->fd);

      fdn->cb(fdn->fd, fdn->param);
    }

  return 0;
}
#endif

#if defined(HAVE_SELECT)
static int scamper_fds_add2(scamper_fd_t *fdn)
{
  if(nfds < fdn->fd && nfds != -1) nfds = fdn->fd;
  return 0;
}

static int scamper_fds_del2(scamper_fd_t *fdn)
{
  nfds = -1;
  return 0;
}

int scamper_fds_poll(struct timeval *timeout)
{
  scamper_fd_t *fdn;
  fd_set        rfds, wfds;
  fd_set       *rfdsp, *wfdsp;
  dlist_node_t *node;
  int c;

  /* if the max file descriptor has not been set, then set it now */
  if(nfds == -1)
    {
      node = dlist_head_node(read_fds);
      while(node != NULL)
	{
	  fdn = (scamper_fd_t *)dlist_node_item(node);
	  if(nfds < fdn->fd) nfds = fdn->fd;
	  node = dlist_node_next(node);
	}

      node = dlist_head_node(write_fds);
      while(node != NULL)
	{
	  fdn = (scamper_fd_t *)dlist_node_item(node);
	  if(nfds < fdn->fd) nfds = fdn->fd;
	  node = dlist_node_next(node);
	}
    }

  /* compose the list of file descriptors to check for read events */
  if(dlist_count(read_fds) > 0)
    {
      rfdsp = &rfds;
      FD_ZERO(rfdsp);
      node = dlist_head_node(read_fds);
      while(node != NULL)
	{
	  fdn = (scamper_fd_t *)dlist_node_item(node);
	  FD_SET(fdn->fd, rfdsp);
	  node = dlist_node_next(node);
	}
    }
  else rfdsp = NULL;

  /* compose the list of file descriptors to check for write events */
  if(dlist_count(write_fds) > 0)
    {
      wfdsp = &wfds;
      FD_ZERO(wfdsp);
      node = dlist_head_node(write_fds);
      while(node != NULL)
	{
	  fdn = (scamper_fd_t *)dlist_node_item(node);
	  FD_SET(fdn->fd, wfdsp);
	  node = dlist_node_next(node);
	}
    }
  else wfdsp = NULL;

  /* find out which file descriptors have a read event */
  if((c = select(nfds+1, rfdsp, wfdsp, NULL, timeout)) == -1)
    {
      printerror(errno, strerror, __func__, "select failed");
      return -1;
    }

  /*
   * now check which ones in our set they correspond to and use the
   * relevant callback
   */
  node = dlist_head_node(read_fds);
  while(node != NULL && c > 0)
    {
      /*
       * get the next node that will be checked now, as fdn->cb may remove it
       * if it decides the fd should go.
       */
      fdn = (scamper_fd_t *)dlist_node_item(node);
      node = dlist_node_next(node);

      if(FD_ISSET(fdn->fd, rfdsp))
	{
	  fdn->cb(fdn->fd, fdn->param);
	  c--;
	}
    }

  node = dlist_head_node(write_fds);
  while(node != NULL && c > 0)
    {
      /*
       * get the next node that will be checked now, as fdn->cb may remove it
       * if it decides the fd should go.
       */
      fdn = (scamper_fd_t *)dlist_node_item(node);
      node = dlist_node_next(node);

      if(FD_ISSET(fdn->fd, wfdsp))
	{
	  fdn->cb(fdn->fd, fdn->param);
	  c--;
	}
    }

  return 0;
}
#endif /* HAVE_SELECT */

static scamper_fd_t *scamper_fds_add(int fd, scamper_fd_cb_t cb,void *param,
				     dlist_t *list)
{
  scamper_fd_t *fdn;

  if((fdn = malloc(sizeof(scamper_fd_t))) == NULL)
    {
      return NULL;
    }

  if((fdn->node = dlist_tail_push(list, fdn)) == NULL)
    {
      free(fdn);
      return NULL;
    }

  fdn->list  = list;
  fdn->fd    = fd;
  fdn->cb    = cb;
  fdn->param = param;

  if(scamper_fds_add2(fdn) == -1)
    {
      dlist_node_pop(fdn->list, fdn->node);
      free(fdn);
      return NULL;
    }

  return fdn;
}

scamper_fd_t *scamper_fds_read_add(int fd, scamper_fd_cb_t cb, void *param)
{
  return scamper_fds_add(fd, cb, param, read_fds);
}

scamper_fd_t *scamper_fds_write_add(int fd, scamper_fd_cb_t cb, void *param)
{
  return scamper_fds_add(fd, cb, param, write_fds);
}

void scamper_fds_del(scamper_fd_t *fdn)
{
  assert(fdn != NULL);

  scamper_fds_del2(fdn);

  if(fdn->node != NULL)
    {
      dlist_node_pop(fdn->list, fdn->node);
      fdn->node = NULL;
    }

  free(fdn);

  return;
}

int scamper_fds_fd(const scamper_fd_t *fdn)
{
  return fdn->fd;
}

int scamper_fds_init()
{
  if((read_fds = dlist_alloc()) == NULL)
    {
      printerror(errno, strerror, __func__, "malloc read fd list failed");
      return -1;
    }

  if((write_fds = dlist_alloc()) == NULL)
    {
      printerror(errno, strerror, __func__, "malloc write fd list failed");
      return -1;
    }

#if defined(HAVE_KQUEUE)
  if((kq = kqueue()) == -1)
    {
      printerror(errno, strerror, __func__, "kqueue failed");
      return -1;
    }
#endif

  return 0;
}

void scamper_fds_cleanup()
{
  scamper_fd_t *fdn;

  if(read_fds != NULL)
    {
      while((fdn = dlist_head_get(read_fds)) != NULL)
	{
	  scamper_fds_del(fdn);
	}
      dlist_free(read_fds);
      read_fds = NULL;
    }

  if(write_fds != NULL)
    {
      while((fdn = dlist_head_get(write_fds)) != NULL)
	{
	  scamper_fds_del(fdn);
	}
      dlist_free(write_fds);
      write_fds = NULL;
    }

#if defined(HAVE_KQUEUE)
  if(kevlist != NULL)
    {
      free(kevlist);
      kevlist = NULL;
    }

  if(kq != -1)
    {
      close(kq);
      kq = -1;
    }
#endif

  return;
}
