/*
 * scamper_file.c
 *
 * $Id: scamper_file.c,v 1.19 2005/04/22 00:04:51 mjl Exp $
 *
 * entry points to read/write scamper traceroute and warts files, and to
 * read skitter arts files.
 *
 * Copyright (C) 2004-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <netinet/in.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include <assert.h>

#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_file.h"
#include "scamper_file_warts.h"
#include "scamper_file_traceroute.h"
#include "scamper_file_arts.h"
#include "utils.h"

#define SCAMPER_FILE_NONE       (-1)
#define SCAMPER_FILE_TRACEROUTE  0
#define SCAMPER_FILE_ARTS        1
#define SCAMPER_FILE_WARTS       2

struct scamper_file
{
  char        *filename;
  int          fd;
  void        *state;
  int          type;
  char         error_str[256];
  uint32_t     capability;
  int          trace_count;
  int          eof;
};

struct handler
{
  char *type;
  int (*detect)(const scamper_file_t *sf);

  int (*init_read)(scamper_file_t *sf);
  int (*init_write)(scamper_file_t *sf);
  int (*init_append)(scamper_file_t *sf);

  scamper_trace_t *(*read_trace)(const scamper_file_t *sf);
  int (*write_trace)(const scamper_file_t *sf, const scamper_trace_t *trace);
  scamper_list_t *(*read_list)(const scamper_file_t *sf);
  void (*close)(scamper_file_t *sf);
};

static struct handler handlers[] = {
  {"traceroute",                        /* type */
   scamper_file_traceroute_is,          /* detect */
   scamper_file_traceroute_init_read,   /* init_read */
   NULL,                                /* init_write */
   NULL,                                /* init_append */
   scamper_file_traceroute_read_trace,  /* read_trace */
   scamper_file_traceroute_write_trace, /* write_trace */
   NULL,                                /* read_list */
   scamper_file_traceroute_close        /* close */
  },
  {"arts",
   scamper_file_arts_is,
   NULL,
   NULL,
   NULL,
   scamper_file_arts_read_trace,
   NULL,
   NULL,
   NULL,
  },
  {"warts",
   scamper_file_warts_is,
   scamper_file_warts_init_read,
   scamper_file_warts_init_write,
   scamper_file_warts_init_append,
   scamper_file_warts_read_trace,
   scamper_file_warts_write_trace,
   scamper_file_warts_read_list,
   scamper_file_warts_close
  }
};

static int handler_cnt = sizeof(handlers) / sizeof(struct handler);

int scamper_file_getfd(const scamper_file_t *sf)
{
  return sf->fd;
}

void *scamper_file_getstate(const scamper_file_t *sf)
{
  return sf->state;
}

char *scamper_file_getfilename(scamper_file_t *sf)
{
  return sf->filename;
}

void scamper_file_setstate(scamper_file_t *sf, void *state)
{
  sf->state = state;
  return;
}

int scamper_file_write_trace(scamper_file_t *sf, const scamper_trace_t *trace)
{
  if(sf->type != SCAMPER_FILE_NONE && handlers[sf->type].write_trace != NULL)
    {
      return handlers[sf->type].write_trace(sf, trace);
    }

  return -1;
}

scamper_trace_t *scamper_file_read_trace(scamper_file_t *sf)
{
  scamper_trace_t *trace = NULL;

  if(sf->type != SCAMPER_FILE_NONE && handlers[sf->type].read_trace != NULL)
    {
      if((trace = handlers[sf->type].read_trace(sf)) != NULL)
	{
	  sf->trace_count++;
	}
    }

  return trace;
}

scamper_list_t *scamper_file_read_list(scamper_file_t *sf)
{
  scamper_list_t *list = NULL;

  if(sf->type != SCAMPER_FILE_NONE && handlers[sf->type].read_list != NULL)
    {
      list = handlers[sf->type].read_list(sf);
    }

  return list;
}

/*
 * scamper_file_eof
 *
 */
int scamper_file_eof(scamper_file_t *sf)
{
  if(sf == NULL || sf->fd == -1) return -1;
  return sf->eof;
}

/*
 * scamper_file_close
 *
 */
void scamper_file_close(scamper_file_t *sf)
{
  if(sf->fd != -1)
    {
      if(sf->type != SCAMPER_FILE_NONE && handlers[sf->type].close != NULL)
	{
	  handlers[sf->type].close(sf);
	}
      else
	{
	  close(sf->fd);
	}
    }

  if(sf->filename)
    {
      free(sf->filename);
    }

  free(sf);

  return;
}

char *scamper_file_type_tostr(scamper_file_t *sf, char *buf, size_t len)
{
  if(sf->type != SCAMPER_FILE_NONE && handlers[sf->type].type != NULL)
    {
      strncpy(buf, handlers[sf->type].type, len);
      return buf;
    }

  return NULL;
}

static int file_type_get(char *type)
{
  int i;

  if(type != NULL)
    {
      for(i=0; i<handler_cnt; i++)
	{
	  if(strcasecmp(type, handlers[i].type) == 0)
	    {
	      return i;
	    }
	}
    }

  return SCAMPER_FILE_NONE;
}

static int file_type_detect(scamper_file_t *sf)
{
  int i;

  for(i=0; i<handler_cnt; i++)
    {
      if(handlers[i].detect(sf) == 1)
	{
	  return i;
	}
    }

  return SCAMPER_FILE_NONE;
}

static int file_open_read(scamper_file_t *sf)
{
  if((sf->type = file_type_detect(sf)) == SCAMPER_FILE_NONE)
    {
      return -1;
    }

  if(handlers[sf->type].init_read != NULL)
    {
      return handlers[sf->type].init_read(sf);
    }

  return 0;
}

static int file_open_write(scamper_file_t *sf)
{
  if(sf->type != SCAMPER_FILE_NONE && handlers[sf->type].init_write != NULL)
    {
      return handlers[sf->type].init_write(sf);
    }

  return 0;
}

static int file_open_append(scamper_file_t *sf)
{
  off_t off;

  /* seek to the end of the file so we can then measure its length */
  if((off = lseek(sf->fd, 0, SEEK_END)) == (off_t)-1)
    {
      return -1;
    }

  /* detect the type of the file that is being appended to */  
  if(off > (off_t)0)
    {
      sf->type = file_type_detect(sf);

      if(handlers[sf->type].init_append != NULL)
	{
	  return handlers[sf->type].init_append(sf);
	}
      else if(sf->type == SCAMPER_FILE_ARTS || SCAMPER_FILE_NONE)
	{
	  return -1;
	}
    }

  return 0;
}

static scamper_file_t *file_open(int fd, char *fn, char mode, int type)
{
  scamper_file_t *sf;
  int (*open_func)(scamper_file_t *sf);

  if(mode == 'r')      open_func = file_open_read;
  else if(mode == 'w') open_func = file_open_write;
  else if(mode == 'a') open_func = file_open_append;
  else return NULL;

  if((sf = (scamper_file_t *)malloc_zero(sizeof(scamper_file_t))) == NULL)
    {
      return NULL;
    }

  sf->type = type;
  sf->fd   = fd;

  if(fn != NULL && (sf->filename = strdup(fn)) == NULL)
    {
      return NULL;
    }

  if(open_func(sf) == -1)
    {
      scamper_file_close(sf);
      return NULL;
    }

  return sf;
}

scamper_file_t *scamper_file_openfd(int fd, char *fn, char mode, char *type)
{
  return file_open(fd, fn, mode, file_type_get(type));
}

/*
 * scamper_file_open
 *
 * open the file specified with the appropriate mode.
 * the modes that we know about are 'r' read-only, 'w' write-only on a
 * brand new file, and 'a' for appending.
 *
 * in 'w' mode [and conditionally for 'a'] an optional parameter may be
 * supplied that says what type of file should be written.
 *  'w' for warts
 *  't' for ascii traceroute
 *  'a' for arts [not implemented]
 *
 * when a file is opened for appending, this second parameter is only
 * used when the file is empty so that writes will be written in the
 * format expected.
 */
scamper_file_t *scamper_file_open(char *filename, char mode, char *type)
{
  scamper_file_t *sf;
  int ft = file_type_get(type);
  int flags = 0;
  int fd = -1;

  if(mode == 'r')
    {
      if(strcmp(filename, "-") == 0)
	{
	  fd = fileno(stdin);
	}
      else
	{
	  flags = O_RDONLY;
	}
    }
  else if(mode == 'w' || mode == 'a')
    {
      /* sanity check the type of file to be written */
      if(ft == SCAMPER_FILE_NONE || ft == SCAMPER_FILE_ARTS)
	{
	  return NULL;
	}

      if(strcmp(filename, "-") == 0)
	{
	  fd = fileno(stdout);
	}
      else
	{
	  if(mode == 'w') flags = O_WRONLY | O_TRUNC | O_CREAT;
	  else            flags = O_RDWR | O_APPEND | O_CREAT;
	}
    }
  else
    {
      return NULL;
    }

  if(fd == -1)
    {
      if(mode == 'r') fd = open(filename, flags);
      else            fd = open(filename, flags, S_IRUSR | S_IWUSR);

      if(fd == -1)
	{
	  return NULL;
	}
    }

  sf = file_open(fd, filename, mode, ft);

  return sf;
}
