/*
 * scamper_file_arts.c
 *
 * $Id: scamper_file_arts.c,v 1.22 2005/05/03 08:44:00 mjl Exp $
 *
 * code to read the legacy arts data file format into scamper_hop structures.
 *
 * Copyright (C) 2004-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <assert.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_file.h"
#include "scamper_file_arts.h"
#include "utils.h"

typedef struct arts_header
{
  uint16_t magic;
  uint32_t id;
  uint8_t  ver;
  uint32_t flags;
  uint16_t attr_num;
  uint32_t attr_length;
  uint32_t data_length;
} arts_header_t;

#define ARTS_MAGIC       0xdfb0
#define ARTS_IP_PATH     0x00003000
#define ARTS_IP_PATH_RTT 0x01

#define ARTS_STOP_NOHALT      0x00
#define ARTS_STOP_ICMPUNREACH 0x01
#define ARTS_STOP_LOOP        0x02
#define ARTS_STOP_GAPLIMIT    0x03

/*
 * arts_read_hdr:
 *
 * read the 20 byte header that is written out before each arts object
 * and parse it into ah
 */
static int arts_read_hdr(const scamper_file_t *sf, arts_header_t *ah)
{
  uint8_t  buf[20];
  uint32_t junk32;
  uint16_t junk16;
  int       ret;
  size_t    rc;
  int       fd;
  
  fd = scamper_file_getfd(sf);

  if((ret = read_wrap(fd, buf, &rc, 20)) != 0)
    {
      /* have we hit the eof? */
      if(ret == -2 && rc == 0)
	{
	  return 0;
	}

      fprintf(stderr, "arts_read_hdr: read %d of 20 bytes\n", (int)rc);
      return -1;
    }

  /* read the magic section of the header */
  memcpy(&junk16, buf, 2);
  ah->magic = ntohs(junk16);
  if(ah->magic != ARTS_MAGIC)
    {
      fprintf(stderr, "arts_read_hdr: expected magic 0x%02x got 0x%02x\n",
	      ARTS_MAGIC, ah->magic);
      return -1;
    }

  /* the id field is stored in the upper 28 bits of the 32 bit field */
  memcpy(&junk32, buf+2, 4);
  junk32  = ntohl(junk32);
  ah->id  = junk32 >> 4;
  ah->ver = junk32 & 0x0f;

  memcpy(&junk32, buf+6, 4);
  ah->flags = ntohl(junk32);

  memcpy(&junk16, buf+10, 2);
  ah->attr_num = ntohs(junk16);

  memcpy(&junk32, buf+12, 4);
  ah->attr_length = ntohl(junk32);

  memcpy(&junk32, buf+16, 4);
  ah->data_length = ntohl(junk32);

  return 1;
}

static int arts_hop_read(scamper_trace_hop_t *hop, const uint8_t *buf,
			 const arts_header_t *ah)
{
  uint32_t junk32;
  int       i = 0;

  /* set defaults for data items stored with this hop */
  hop->icmp_type   = 11;
  hop->icmp_code   = 0;
  hop->rtt.tv_sec  = 0;
  hop->rtt.tv_usec = 0;
  hop->probe_id    = 0;
  hop->reply_ttl   = 0;

  /* read the 1 byte hop number this path entry refers to */
  hop->probe_ttl = buf[i++];
  assert(hop->probe_ttl > 0);

  /* the IPv4 address of the hop that responded */
  if((hop->addr = scamper_addr_alloc(SCAMPER_ADDR_TYPE_IPV4, buf+i)) != NULL)
    {
      return -1;
    }
  i += 4;

  /* arts 1 always stores RTT per hop; arts > 1 conditionally stores it */
  if(ah->ver == 1 || (ah->flags & ARTS_IP_PATH_RTT && ah->ver > 1))
    {
      /* RTT, stored in microseconds */
      memcpy(&junk32, buf+i, 4); i += 4;
      junk32 = ntohl(junk32);
      hop->rtt.tv_sec  = junk32 / 1000000;
      hop->rtt.tv_usec = junk32 % 1000000;
      
      /* num tries */
      hop->probe_id = buf[i++]-1;
    }

  return i;
}

static scamper_trace_hop_t *arts_hops_read(const arts_header_t *ah,
					   const uint8_t *buf,
					   int count, int *off)
{
  scamper_trace_hop_t *head = NULL, *hop = NULL;
  int i = 0;

  if(count == 0)
    {
      return NULL;
    }

  while(count-- > 0)
    {
      if(hop != NULL)
	{
	  hop->next = scamper_trace_hop_alloc();
	  hop = hop->next;
	}
      else
	{
	  head = hop = scamper_trace_hop_alloc();
	}

      if(hop == NULL)
	{
	  while(head != NULL)
	    {
	      hop = head;
	      head = head->next;
	      scamper_trace_hop_free(hop);
	    }
	  return NULL;
	}

      i += arts_hop_read(hop, buf+i, ah);
    }

  *off += i;

  return head;
}

scamper_trace_t *arts_read_trace(const scamper_file_t *sf,
				 const arts_header_t *ah)
{
  scamper_trace_t     *trace = NULL;
  uint8_t             *buf = NULL;
  int                  i;
  uint32_t             junk32;
  uint8_t              junk8;
  uint8_t              hop_distance;
  uint8_t              halt_reason;
  uint8_t              halt_reason_data;
  uint8_t              reply_ttl;
  uint32_t             rtt;
  scamper_trace_hop_t *hops;
  scamper_trace_hop_t *hop;
  uint8_t              num_hop_recs;
  uint8_t              max_hop;
  int                  fd;
  size_t               rc;

  fd = scamper_file_getfd(sf);

  if(lseek(fd, ah->attr_length, SEEK_CUR) == -1)
    {
      fprintf(stderr, "scamper_file_read_arts: 1b\n");
      goto err;
    }

  if((buf = malloc(ah->data_length)) == NULL)
    {
      fprintf(stderr, "arts_read_trace: malloc %d for trace object failed\n",
	      ah->data_length);
      goto err;
    }

  if(read_wrap(fd, buf, &rc, ah->data_length) != 0)
    {
      fprintf(stderr, "arts_read_trace: read %d expected %d\n", (int)rc,
	      ah->data_length);
      goto err;
    }

  if((trace = scamper_trace_alloc()) == NULL)
    {
      fprintf(stderr, "arts_read_trace: scamper_trace_alloc failed\n");
      goto err;
    }

  i = 0;

  if((trace->src = scamper_addr_alloc(SCAMPER_ADDR_TYPE_IPV4, buf+i)) == NULL)
    {
      goto err;
    }
  i += 4;

  if((trace->dst = scamper_addr_alloc(SCAMPER_ADDR_TYPE_IPV4, buf+i)) == NULL)
    {
      goto err;
    }
  i += 4;

  /*
   * read the RTT of the last hop
   * arts prior to version 2 stores a timeval struct in the file for
   * recording RTT, which is wasteful
   */
  memcpy(&junk32, buf+i, 4); i += 4;
  rtt = ntohl(junk32);
  if(ah->ver < 2)
    {
      rtt *= 1000000;
      memcpy(&junk32, buf+i, 4); i += 4;
      rtt += ntohl(junk32);
    }

  /*
   * the hop distance field tells us how many hops a packet takes to a
   * destination
   */
  hop_distance = buf[i]; i++;

  /*
   * read the next 8 bit field.  the first bit says if the trace was 
   * successful in probing to the end host, and the other 7 bits say
   * how many hops actually responded to a probe.
   */
  junk8 = buf[i]; i++;
  if(junk8 & 0x40)
    {
      trace->stop_reason = SCAMPER_TRACE_STOP_COMPLETED;
    }
  num_hop_recs = junk8 & 0x7f;

  /*
   * arts versions after 1 (and arts version 1 conditionally) store
   * data that tells us why the trace stopped
   */
  if(ah->ver > 1 || ((junk8 & 0x40) && ah->ver == 1))
    {
      halt_reason      = buf[i]; i++;
      halt_reason_data = buf[i]; i++;

      switch(halt_reason)
	{
	case ARTS_STOP_NOHALT:
	  trace->stop_reason = SCAMPER_TRACE_STOP_NONE;
	  break;

	case ARTS_STOP_ICMPUNREACH:
	  trace->stop_reason = SCAMPER_TRACE_STOP_UNREACH;
	  trace->stop_data   = halt_reason_data;
	  break;

	case ARTS_STOP_LOOP:
	  trace->stop_reason = SCAMPER_TRACE_STOP_LOOP;
	  trace->stop_data   = halt_reason_data;
	  break;

	case ARTS_STOP_GAPLIMIT:
	  trace->stop_reason = SCAMPER_TRACE_STOP_DEAD;
	  trace->stop_reason = halt_reason_data;
	  break;
	}
    }

  if(num_hop_recs == 0)
    {
      free(buf);
      return trace;
    }

  /*
   * arts >= 2 stores the TTL of reply packet from a destination so we
   * can estimate the number of hops on the reverse path
   */
  if(ah->ver >= 2)
    {
      reply_ttl = buf[i]; i++;
    }

  if((hops = arts_hops_read(ah, buf+i, num_hop_recs, &i)) == NULL)
    {
      fprintf(stderr, "arts_read_trace: arts_hops_read %d failed\n",
	      num_hop_recs);
      goto err;
    }

  /* the num_hop_recs variable tells us how many hops were responsive */
  max_hop = 0;
  for(hop = hops; hop != NULL; hop = hop->next)
    {
      /* is this TTL > other TTLs we've seen so far? */
      if(max_hop < hop->probe_ttl)
	{
	  max_hop = hop->probe_ttl;
	}
    }

  assert(i == ah->data_length);
  free(buf); buf = NULL;

  if(max_hop == 0)
    {
      return trace;
    }

  if(scamper_trace_hops_alloc(trace, max_hop) == -1)
    {
      goto err;
    }
  trace->hop_count = max_hop;

  /*
   * now loop through the hops array stored in this procedure
   * and assemble the responses into trace->hops. order them based
   * on the probe's ttl then by attempt
   */
  trace->hops[hops->probe_ttl-1] = hop = hops;
  while(hop->next != NULL) 
    {
      if(hop->probe_ttl != hop->next->probe_ttl)
	{
	  i = hop->next->probe_ttl-1;
	  trace->hops[i] = hop->next;
	  hop->next = NULL;
	  hop = trace->hops[i];
	}
      else hop = hop->next;
    }

  return trace;

 err:
  if(trace != NULL) scamper_trace_free(trace);
  if(buf != NULL) free(buf);
  return NULL;
}

/*
 * scamper_file_read_arts
 * legacy arts only recognises IPv4 traces
 */
scamper_trace_t *scamper_file_arts_read_trace(const scamper_file_t *sf)
{
  scamper_trace_t *trace;
  arts_header_t    ah;
  int              i, fd;

  fd = scamper_file_getfd(sf);

  /*
   * look for an IP Path arts object
   * we must skip over any other data objects
   */
  for(;;)
    {
      i = arts_read_hdr(sf, &ah);
      if(i == -1)
	{
	  return NULL;
	}
      else if(i == 0)
	{
	  return NULL;
	}

      if(ah.id == ARTS_IP_PATH)
	{
	  break;
	}

      if(lseek(fd, ah.attr_length + ah.data_length, SEEK_CUR) == -1)
	{
	  fprintf(stderr, "scamper_file_read_arts: 2\n");
	  return NULL;
	}
    }

  trace = arts_read_trace(sf, &ah);
  return trace;
}

int scamper_file_arts_is(const scamper_file_t *sf)
{
  uint16_t magic16;
  int fd;

  fd = scamper_file_getfd(sf);

  if(lseek(fd, 0, SEEK_SET) == -1)
    {
      return 0;
    }

  if(read_wrap(fd, &magic16, NULL, sizeof(magic16)) != 0)
    {
      return 0;
    }

  magic16 = ntohs(magic16);

  if(magic16 == ARTS_MAGIC)
    {
      if(lseek(fd, 0, SEEK_SET) == -1)
	{
	  return 0;
	}
      return 1;
    }

  return 0;
}
