/*
 * scamper_file_warts.c
 *
 * the Waikato ARTS file format replacement
 *
 * $Id: scamper_file_warts.c,v 1.60 2005/05/03 08:44:00 mjl Exp $
 *
 * code to read scamper's traceroute-like file format into scamper_hop
 * structures.
 *
 * Copyright (C) 2004-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <assert.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_file.h"
#include "scamper_file_warts.h"

#include "mjl_splaytree.h"
#include "utils.h"

#define WARTS_MAGIC 0x1205

/* various types of WARTS headers */
#define WARTS_CYCLE_NEW      0x01
#define WARTS_CYCLE_FIN      0x02
#define WARTS_ADDR           0x03
#define WARTS_TRACE          0x04
#define WARTS_LIST           0x05
#define WARTS_REVERSE        0x06
#define WARTS_ALIAS          0x07

/*
 * trace attributes: 2 bytes each.
 * the first 4 bits are the type, the second 12 bits are the length
 */
#define WARTS_TRACE_ATTR_HDR(type, len) ((type << 12) | len)
#define WARTS_TRACE_ATTR_HDR_TYPE(hdr)  ((hdr >> 12) & 0xf)
#define WARTS_TRACE_ATTR_HDR_LEN(hdr)    (hdr & 0x0fff)
#define WARTS_TRACE_ATTR_EOF     0x0000
#define WARTS_TRACE_ATTR_PMTUD   0x1

/* how many entries to grow the table by each time */
#define WARTS_ADDR_TABLEGROW 1000

/*
 * warts_addr
 *
 * this structure associates an address with an id number.
 */
typedef struct warts_addr
{
  scamper_addr_t *addr;
  uint32_t id;
} warts_addr_t;

#define warts_addr_af(wa)   ((wa)->addr->af)
#define warts_addr_addr(wa) ((wa)->addr->addr)

/*
 * warts_hdr
 *
 * this object is written at the start of every other object.
 * the type field says what type of record follows.
 * the length field (16 bits) is excessive.  perhaps this should
 * be re-factored to 4 bits type [16 types], 12 bits length [4096 byte
 * records].
 */
typedef struct warts_hdr
{
  uint16_t magic;
  uint16_t len;
  uint8_t  type;
} warts_hdr_t;

/*
 * warts_state
 *
 */
typedef struct warts_state
{
  /* an array of the lists that have been declared */
  uint32_t         list_count;
  scamper_list_t **list_table;

  /*
   * either a tree or an array of the addresses declared.
   * if the file is being written to, the tree is active so that address ids
   * can be found quickly.  otherwise if the file is being read, the addresses
   * are stored in a table for instantaneous location.
   */
  uint32_t         addr_count;
  splaytree_t     *addr_tree;
  warts_addr_t   **addr_table;
} warts_state_t;

typedef struct warts_var
{
  int     id;
  ssize_t size;
  int     tlv_id;
} warts_var_t;

static const warts_var_t trace_vars[] =
{
  { 1, 4, -1},                                /* list id assigned by warts */
  { 2, 4, -1},                                /* cycle id assigned by warts */
  { 3, 4, -1},                                /* source address key */
  { 4, 4, -1},                                /* destination address key */
  { 5, 8, -1},                                /* start timestamp */
  { 6, 1, -1},                                /* stop reason */
  { 7, 1, -1},                                /* stop data */
  { 8, 1, -1},                                /* flags */
  { 9, 1, -1},                                /* attempts */
  {10, 1, -1},                                /* hoplimit */
  {11, 1, -1},                                /* type */
  {12, 2, -1},                                /* probe_size */
  {13, 2, -1},                                /* sport */
  {14, 2, -1},                                /* dport */
};
static const int trace_vars_cnt = sizeof(trace_vars)/sizeof(warts_var_t);
static const int trace_vars_mfb = 2;

static const warts_var_t pmtud_vars[] =
{
  { 1, 2, -1},                                /* interface mtu */
  { 2, 2, -1},                                /* path mtu */
  { 3, 2, SCAMPER_TRACE_PMTUD_TLV_OUTMTU},    /* mtu to gateway */
};
static const int pmtud_vars_cnt = sizeof(pmtud_vars)/sizeof(warts_var_t);
static const int pmtud_vars_mfb = 1;

static const warts_var_t hop_vars[] =
{
  { 1, 4, -1},                                /* address id */
  { 2, 1, -1},                                /* probe ttl */
  { 3, 1, -1},                                /* reply ttl */
  { 4, 1, -1},                                /* flags */
  { 5, 1, -1},                                /* probe id */
  { 6, 4, -1},                                /* round trip time */
  { 7, 2, -1},                                /* icmp type / code */
  { 8, 2, -1},                                /* probe size */
  { 9, 2, -1},                                /* reply size */
  {10, 2, SCAMPER_TRACE_HOP_TLV_REPLY_IPID},  /* ipid of reply packet */
  {11, 1, SCAMPER_TRACE_HOP_TLV_REPLY_IPTOS}, /* tos bits of reply packet */
  {12, 2, SCAMPER_TRACE_HOP_TLV_NHMTU},       /* next hop mtu in ptb message */
  {13, 2, SCAMPER_TRACE_HOP_TLV_INNER_IPLEN}, /* ip->len from inside icmp */
  {14, 1, SCAMPER_TRACE_HOP_TLV_INNER_IPTTL}, /* ip->ttl from inside icmp */
};
static const int hop_vars_cnt = sizeof(hop_vars)/sizeof(warts_var_t);
static const int hop_vars_mfb = 2;

/*
 * set_flag
 *
 * small routine to set a flag bit.
 */
static void set_flag(uint8_t *flags, int id)
{
  int i, j;

  assert(id > 0);

  if(id % 7 == 0)
    {
      i = (id / 7) - 1;
      j = 7;
    }
  else
    {
      i = id / 7;
      j = id % 7;
    }

  flags[i] |= (0x1 << (j-1));

  return;
}

/*
 * warts_hdr_read
 *
 */
static int warts_hdr_read(const scamper_file_t *sf, warts_hdr_t *hdr)
{
  int      fd;
  uint8_t  buf[5];
  size_t   rc;
  int      ret;

  fd = scamper_file_getfd(sf);

  if((ret = read_wrap(fd, buf, &rc, 5)) != 0)
    {
      /* have we hit the eof? */
      if(rc == 0 && ret == -2)
	{
	  return 0;
	}

      fprintf(stderr, "warts_read_hdr: read %d of 5 bytes\n", (int)rc);
      return -1;
    }

  memcpy(&hdr->magic, buf, 2);
  hdr->magic = ntohs(hdr->magic);

  memcpy(&hdr->len, buf+2, 2);
  hdr->len = ntohs(hdr->len);

  hdr->type = buf[4];

  return 1;
}

/*
 * warts_hdr_write
 *
 */
static int warts_hdr_write(const scamper_file_t *sf,
			   const uint8_t type, const uint16_t len)
{
  int      fd;
  off_t    off = 0;
  uint8_t  buf[5];
  uint16_t junk16;
  size_t   wc;
  int      ret;

  fd = scamper_file_getfd(sf);

  junk16 = htons(WARTS_MAGIC);
  memcpy(buf,   &junk16, 2);
  junk16 = htons(len);
  memcpy(buf+2, &junk16, 2);
  buf[4] = type;

  if(fd != 1 && (off = lseek(fd, 0, SEEK_CUR)) == (off_t)-1)
    {
      return -1;
    }

  if((ret = write_wrap(fd, buf, &wc, 5)) != 0)
    {
      if(wc != 0 && fd != 1) ftruncate(fd, off);
      return -1;
    }

  return 0;
}

static uint8_t *warts_read(const scamper_file_t *sf, const warts_hdr_t *hdr)
{
  uint8_t *buf;
  int      fd;

  if((fd = scamper_file_getfd(sf)) == -1)
    {
      return NULL;
    }

  if((buf = (void *)malloc(hdr->len)) == NULL)
    {
      return NULL;
    }

  if(read_wrap(fd, buf, NULL, hdr->len) != 0)
    {
      free(buf);
      return NULL;
    }

  return buf;
}

/*
 * warts_write
 *
 * this function will write a record to disk, appending a warts_header
 * on the way out to the disk.  if the write fails for whatever reason
 * (as in the disk is full and only a partial recrd can be written), then
 * the write will be retracted in its entirety.
 */
static int warts_write(const scamper_file_t *sf, const uint8_t type,
		       const void *buf, const size_t len)
{
  off_t  off = 0;
  int    fd;

  fd = scamper_file_getfd(sf);
  if(fd != 1 && (off = lseek(fd, 0, SEEK_CUR)) == (off_t)-1)
    {
      return -1;
    }

  if(warts_hdr_write(sf, type, len) == -1)
    {
      return -1;
    }

  if(write_wrap(fd, buf, NULL, len) != 0)
    {
      /*
       * if we could not write the buf out, then truncate the warts file
       * at the hdr we just wrote out above.
       */
      if(fd != 1) ftruncate(fd, off);
      return -1;
    }

  return 0;
}

static int warts_addr_cmp(const void *a, const void *b)
{
  warts_addr_t *wa = (warts_addr_t *)a;
  warts_addr_t *wb = (warts_addr_t *)b;

  return scamper_addr_cmp(wa->addr, wb->addr);
}

static void warts_addr_free(warts_addr_t *addr)
{
  if(addr->addr != NULL)
    {
      scamper_addr_free(addr->addr);
    }
  free(addr);
  return;
}

/*
 * warts_read_addr
 *
 * read an address structure out of the file and record it in the splay
 * tree of addresses.
 *
 * each address record consists of
 *   - an id assigned to the address, modulo 255
 *   - the address family the address belongs to
 *   - the address [length determined by record length]
 */
static int warts_addr_read(const scamper_file_t *sf, const warts_hdr_t *hdr)
{
  warts_state_t *state;
  warts_addr_t  *addr = NULL;
  uint8_t       *buf = NULL;
  void          *tmp;
  size_t         size;
  int            type;

  /* the data has to be at least 3 bytes long to be valid */
  assert(hdr->len > 2);

  state = scamper_file_getstate(sf);

  if(state->addr_count % WARTS_ADDR_TABLEGROW == 0)
    {
      size = sizeof(warts_addr_t*)*(state->addr_count + WARTS_ADDR_TABLEGROW);
      if((tmp = (void *)realloc(state->addr_table, size)) == NULL)
	{
	  return -1;
	}

      state->addr_table = tmp;
    }

  if((buf = warts_read(sf, hdr)) == NULL)
    {
      return -1;
    }

  type = buf[1];

  /* allocate a warts-addr structure, and initialise it */
  if((addr = malloc_zero(sizeof(struct warts_addr))) == NULL)
    {
      goto err;
    }
  if((addr->addr = scamper_addr_alloc(type, buf+2)) == NULL)
    {
      goto err;
    }
  addr->id = state->addr_count;
  free(buf); buf = NULL;

  state->addr_table[state->addr_count++] = addr;

  return 0;

 err:
  if(addr != NULL) warts_addr_free(addr);
  if(buf != NULL) free(buf);
  return -1;
}

/*
 * warts_addr_write
 *
 *
 */
static int warts_addr_write(const scamper_file_t *sf,
			    scamper_addr_t *sa, uint32_t *id)
{
  uint8_t        buf[18];
  size_t         size;
  warts_state_t *state;
  warts_addr_t  *wa;

  size = scamper_addr_size(sa);
  state = scamper_file_getstate(sf);

  /* create the warts addr struct to associate an id with an address */
  if((wa = malloc_zero(sizeof(warts_addr_t))) == NULL)
    {
      return -1;
    }
  wa->addr = scamper_addr_use(sa);
  wa->id = state->addr_count;

  /* create the structure to write to disk */
  buf[0] = wa->id % 255;
  buf[1] = sa->type;
  memcpy(buf+2, sa->addr, size);

  /*
   * insert the warts_addr into a tree so we can find the assigned id
   * quickly.  if we can't put it in the tree, then free the struct and
   * return
   */
  if(splaytree_insert(state->addr_tree, wa) != 1)
    {
      warts_addr_free(wa);
      return -1;
    }

  /*
   * write the warts address record to disk.  if we can't write it, we
   * don't discard it from the tree as it is there now, and doesn't do
   * any real harm being there.
   */
  if(warts_write(sf, WARTS_ADDR, buf, size+2) == -1)
    {
      return -1;
    }

  state->addr_count++;

  *id = wa->id;
  return 0;
}

static int warts_addr_find(const scamper_file_t *sf,
			   scamper_addr_t *sa, uint32_t *id)
{
  warts_state_t *state;
  warts_addr_t   findme;
  warts_addr_t  *wa;

  state = scamper_file_getstate(sf);

  findme.addr = sa;
  findme.id = 0;
  if((wa = splaytree_find(state->addr_tree, &findme)) == NULL)
    {
      return -1;
    }

  *id = wa->id;
  return 0;
}

/*
 * warts_addr_getid
 *
 */
static int warts_addr_getid(const scamper_file_t *sf,
			    scamper_addr_t *sa, uint32_t *id)
{
  if(warts_addr_find(sf, sa, id) == 0)
    {
      return 0;
    }

  if(warts_addr_write(sf, sa, id) == 0)
    {
      return 0;
    }

  return -1;
}

static scamper_cycle_t *warts_cyclenew_read(const scamper_file_t *sf,
					    const warts_hdr_t *hdr)
{
  scamper_cycle_t *cycle;
  uint8_t *buf;
  uint32_t junk32;
  uint32_t list_id;
  int      i;

  if((buf = warts_read(sf, hdr)) == NULL)
    {
      return NULL;
    }

  if((cycle = malloc(sizeof(scamper_cycle_t))) == NULL)
    {
      free(buf);
      return NULL;
    }  

  /*
   * the first 4 bytes is the list id for the cycle.
   * the second 4 bytes is the actual cycle id.
   */
  i = 0;
  memcpy(&junk32, buf+i, 4); i += 4;
  list_id = ntohl(junk32);
  memcpy(&junk32, buf+i, 4); i += 4;
  cycle->id = ntohl(junk32);

  return cycle;
}

static int warts_cyclenew_write(const scamper_file_t *sf,
				const scamper_cycle_t *cycle)
{
  uint8_t *buf;
  size_t   size;
  int      i;
  uint32_t junk32;

  size = 4 + 4; /* room for 32 bit list and cycle ids */

  if((buf = malloc(size)) == NULL)
    {
      return -1;
    }

  i = 0;

  /* list and cycle ids */
  junk32 = htonl(cycle->list->id);
  memcpy(buf+i, &junk32, 4); i += 4;
  junk32 = htonl(cycle->id);
  memcpy(buf+i, &junk32, 4); i += 4;

  if(warts_write(sf, WARTS_CYCLE_NEW, buf, size) == -1)
    {
      free(buf);
      return -1;
    }

  free(buf);  

  return 0;
}

static scamper_cycle_t *warts_cyclefin_read(const scamper_file_t *sf,
					    const warts_hdr_t *hdr)
{
  scamper_cycle_t *cycle;
  uint32_t         list_id, cycle_id;
  uint8_t         *buf;

  if((buf = warts_read(sf, hdr)) == NULL)
    {
      return NULL;
    }

  memcpy(&list_id,  buf,   4); list_id  = ntohl(list_id);
  memcpy(&cycle_id, buf+4, 4); cycle_id = ntohl(cycle_id);

  cycle = NULL;

  return cycle;
}

static int warts_cyclefin_write(const scamper_file_t *sf,
				const scamper_cycle_t *cycle)
{
  uint8_t *buf;
  size_t   size;
  uint32_t junk32;

  size = 8;
  if((buf = malloc(size)) == NULL)
    {
      return -1;
    }

  junk32 = htonl(cycle->list->id);
  memcpy(buf,   &junk32, 4);
  junk32 = htonl(cycle->id);
  memcpy(buf+4, &junk32, 4);

  if(warts_write(sf, WARTS_CYCLE_FIN, buf, size) == -1)
    {
      free(buf);
      return -1;
    }

  free(buf);
  return 0;
}

static int warts_list_store(const scamper_file_t *sf, scamper_list_t *list,
			    const uint32_t id)
{
  warts_state_t *state = scamper_file_getstate(sf);
  scamper_list_t **table;
  uint32_t *int_ptr;
  size_t size;

  if(id != state->list_count)
    {
      return -1;
    }

  /* malloc an integer to store with the list internally */
  if((int_ptr = malloc(sizeof(uint32_t))) == NULL)
    {
      return -1;
    }
  *int_ptr = id;

  /* make the table one entry bigger */
  size = (state->list_count+1) * sizeof(scamper_list_t *);
  if((table = realloc(state->list_table, size)) == NULL)
    {
      free(int_ptr);
      return -1;
    }
  state->list_table = table;

  /* storing the list in the table increases the reference count */
  table[state->list_count++] = scamper_list_use(list);
  list->internal = int_ptr;

  return 0;
}

static scamper_list_t *warts_list_read(const scamper_file_t *sf,
				       const warts_hdr_t *hdr)
{
  scamper_list_t *list = NULL;
  uint8_t  *buf = NULL;
  uint32_t  junk32, id;
  char     *name;
  int       len;
  int       i;

  if((buf = warts_read(sf, hdr)) == NULL)
    {
      goto err;
    }

  if((list = malloc_zero(sizeof(scamper_list_t))) == NULL)
    {
      goto err;
    }

  i = 0;

  /* this is the internal list id used by warts */
  memcpy(&junk32, buf+i, 4); i += 4;
  id = ntohl(junk32);

  /* this is the list id assigned by a human */
  memcpy(&junk32, buf+i, 4); i += 4;
  list->id = ntohl(junk32);

  /* get the name of the list */
  name = (char *)(buf+i);
  if((len = strlen(name)) != 0)
    {
      if((list->name = malloc(len+1)) == NULL)
	{
	  goto err;
	}
      memcpy(list->name, name, len+1);
    }

  i += len + 1;

  /* check to see if we have a list description to include */
  if(i < hdr->len)
    {
      if((list->descr = malloc(hdr->len - i + 1)) == NULL)
	{
	  goto err;
	}

      memcpy(list->descr, buf+i, hdr->len-i);
      list->descr[hdr->len-i] = '\0';
    }

  if(warts_list_store(sf, list, id) == -1)
    {
      goto err;
    }

  return list;

 err:
  if(list != NULL)
    {
      if(list->name != NULL)  free(list->name);
      if(list->descr != NULL) free(list->descr);
      free(list);
    }
  if(buf != NULL)
    {
      free(buf);
    }
  return NULL;
}

/*
 * warts_list_write
 *
 * take a list structure and write it to disk.  update the state held, too
 *
 * note that we do not write a null termination byte at the end, as we know
 * how long the string will be from the hdr->len.
 *
 */
static int warts_list_write(const scamper_file_t *sf, scamper_list_t *list,
			    uint32_t *id)
{
  warts_state_t *state = scamper_file_getstate(sf);
  uint32_t  junk32;
  uint8_t  *buf = NULL;
  size_t    size;
  int       i;
  int       namelen;
  int       descrlen;

  /* allocate the list an id and store it with it */
  *id = state->list_count;

  if(warts_list_store(sf, list, *id) == -1)
    {
      goto err;
    }

  namelen  = list->name != NULL ? strlen(list->name) : 0;
  descrlen = list->descr != NULL ? strlen(list->descr) : 0;
  size = 4 + 4 + namelen + 1 + descrlen;

  if((buf = malloc(size)) == NULL)
    {
      goto err;
    }

  i = 0;

  /* id assigned by warts */
  junk32 = htonl(*id);
  memcpy(buf+i, &junk32, 4); i += 4;

  /* list id assigned by a person */
  junk32 = htonl(list->id);
  memcpy(buf+i, &junk32, 4); i += 4;

  /* list name */
  if(list->name != NULL)
    memcpy(buf+i, list->name, namelen+1);
  else
    buf[i] = '\0';

  i += namelen + 1;

  /* list descr */
  if(list->descr != NULL)
    memcpy(buf+i, list->descr, descrlen);

  /* write the list record to disk */
  if(warts_write(sf, WARTS_LIST, buf, size) == -1)
    {
      goto err;
    }

  free(buf);
  return 0;

 err:
  if(list->internal != NULL)
    {
      scamper_list_free(state->list_table[--state->list_count]);
      free(list->internal);
      list->internal = NULL;
    }
  if(buf != NULL)
    {
      free(buf);
    }
  return -1;
}

/*
 * warts_list_getid
 *
 * given a scamper_list structure, return the id to use internally to
 * uniquely identify it.  allocate the id if necessary.
 */
static int warts_list_getid(const scamper_file_t *sf, scamper_list_t *list,
			    uint32_t *id)
{
  /* already have an id assigned, return it. */
  if(list->internal != NULL)
    {
      *id = *((uint32_t *)list->internal);
      return 0;
    }

  if(warts_list_write(sf, list, id) == -1)
    {
      return -1;
    }

  return 0;
}

static int warts_trace_params_flags(const scamper_trace_t *trace,
				    uint8_t *flags)
{
  int i, j, max_id = 0;

  /* unset all the flags possible */
  memset(flags, 0, trace_vars_mfb);

  /* for now, we include the base data items */
  for(i=1; i<=14; i++) set_flag(flags, i);
  max_id = 14;

  /* figure out how many bytes have been used */
  j = max_id / 7;
  if((max_id % 7) != 0) j++;

  /* skip through and set the 'more flags' bit for all flag bytes necessary */
  for(i=0; i<j-1; i++)
    {
      flags[i] |= 0x80;
    }

  return j;  
}

static size_t warts_trace_params_size(const scamper_trace_t *trace)
{
  uint8_t flags[trace_vars_mfb];
  size_t size;
  int    i, j;

  size = warts_trace_params_flags(trace, flags);

  /* first two bytes define the size of the pmtud data stored */
  size += 2;

  /* for now, set all the trace parameter flags */
  for(i=0; i<trace_vars_mfb; i++)
    {
      for(j=0; j<7; j++)
	{
	  if((flags[i] & (0x1 << j)) != 0)
	    {
	      size += trace_vars[(i*7)+j].size;
	    }
	}

      if((flags[i] & 0x80) == 0) break;
    }

  return size;
}

/*
 * warts_trace_params_read
 *
 */
static int warts_trace_params_read(scamper_trace_t *trace,
				   const warts_state_t *state,
				   const uint8_t *buf)
{
  uint16_t      size;
  uint16_t      off;
  int           i, j, id;
  uint8_t       flags;
  uint16_t      junk16;
  uint32_t      junk32;
  warts_addr_t *wa;

  /* the length field that says how much data is in the parameter block */
  memcpy(&junk16, buf, 2);
  size = ntohs(junk16);
  off = 2;

  /* determine how many flag bytes to read to get to the actual data */
  for(i=0;;i++)
    {
      flags = buf[2+i];
      if((flags & 0x80) == 0) break;
    }
  off += (i+1);

  for(i=0;;i++)
    {
      flags = buf[2+i];

      for(j=0; j<7; j++)
	{
	  id = (i*7)+j;

	  switch(id)
	    {
	    case 0: /* list id */
	      memcpy(&junk32, buf+off, 4); junk32 = ntohl(junk32);
	      if(junk32 >= state->list_count) goto err;
	      trace->list = scamper_list_use(state->list_table[junk32]);
	      break;

	    case 1: /* cycle id */
	      break;

	    case 2: /* source address key */
	      memcpy(&junk32, buf+off, 4); junk32 = ntohl(junk32);
	      if(junk32 >= state->addr_count) goto err;
	      wa = state->addr_table[junk32];
	      trace->src = scamper_addr_use(wa->addr);
	      break;

	    case 3: /* destination address key */
	      memcpy(&junk32, buf+off, 4); junk32 = ntohl(junk32);
	      if(junk32 >= state->addr_count) goto err;
	      wa = state->addr_table[junk32];
	      trace->dst = scamper_addr_use(wa->addr);
	      break;

	    case 4: /* start timestamp */
	      memcpy(&junk32, buf+off, 4);
	      trace->start.tv_sec  = ntohl(junk32);
	      memcpy(&junk32, buf+off+4, 4);
	      trace->start.tv_usec = ntohl(junk32);
	      break;

	    case 5: /* stop reason */
	      trace->stop_reason = buf[off];
	      break;

	    case 6: /* stop data */
	      trace->stop_data = buf[off];
	      break;

	    case 7: /* flags */
	      trace->flags = buf[off];
	      break;

	    case 8: /* attempts */
	      trace->attempts = buf[off];
	      break;

	    case 9: /* hoplimit */
	      trace->hoplimit = buf[off];
	      break;

	    case 10: /* trace type */
	      trace->type = buf[off];
	      break;

	    case 11: /* probe size */
	      memcpy(&junk16, buf+off, 2);
	      trace->probe_size = ntohs(junk16);
	      break;

	    case 12: /* sport */
	      memcpy(&junk16, buf+off, 2);
	      trace->sport = ntohs(junk16);
	      break;

	    case 13: /* dport */
	      memcpy(&junk16, buf+off, 2);
	      trace->dport = ntohs(junk16);
	      break;

	    default:
	      goto done;
	    }

	  if(trace_vars[id].size >= 0)
	    {
	      off += trace_vars[id].size;
	    }
	}

      if((flags & 0x80) == 0) break;
    }

 done:
  return size;

 err:
  return -1;
}

static int warts_trace_params_write(const scamper_file_t *sf,
				    const scamper_trace_t *trace,
				    uint8_t *buf)
{
  int       i, j, id;
  uint16_t  off;
  uint16_t  junk16;
  uint32_t  junk32;
  uint8_t   flags;
  uint32_t  src_id, dst_id, list_id;

  if(warts_addr_getid(sf, trace->src, &src_id) == -1)
    {
      goto err;
    }

  if(warts_addr_getid(sf, trace->dst, &dst_id) == -1)
    {
      goto err;
    }

  if(warts_list_getid(sf, trace->list, &list_id) == -1)
    {
      goto err;
    }

  off = 2;
  off += warts_trace_params_flags(trace, buf+2);

  for(i=0;;i++)
    {
      flags = buf[2+i];
      for(j=0; j<7; j++)
	{
	  if((flags & (0x1 << j)) != 0)
	    {
	      id = (i*7)+j;

	      switch(id)
		{
		case 0: /* list id */
		  junk32 = htonl(list_id);
		  memcpy(buf+off, &junk32, 4);
		  break;

		case 1: /* cycle id */
		  break;

		case 2: /* source address key */
		  junk32 = htonl(src_id);
		  memcpy(buf+off, &junk32, 4);
		  break;

		case 3: /* destination address key */
		  junk32 = htonl(dst_id);
		  memcpy(buf+off, &junk32, 4);
		  break;

		case 4: /* start timestamp */
		  junk32 = htonl(trace->start.tv_sec);
		  memcpy(buf+off, &junk32, 4);
		  junk32 = htonl(trace->start.tv_usec);
		  memcpy(buf+off+4, &junk32, 4);
		  break;

		case 5: /* stop reason */
		  buf[off] = trace->stop_reason;
		  break;

		case 6: /* stop data */
		  buf[off] = trace->stop_data;
		  break;

		case 7: /* flags */
		  buf[off] = trace->flags;
		  break;

		case 8: /* attempts */
		  buf[off] = trace->attempts;
		  break;

		case 9: /* hoplimit */
		  buf[off] = trace->hoplimit;
		  break;

		case 10: /* trace type */
		  buf[off] = trace->type;
		  break;

		case 11: /* probe size */
		  junk16 = htons(trace->probe_size);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 12: /* sport */
		  junk16 = htons(trace->sport);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 13: /* dport */
		  junk16 = htons(trace->dport);
		  memcpy(buf+off, &junk16, 2);
		  break;

		default:
		  goto err;
		}

	      if(trace_vars[id].size >= 0)
		{
		  off += trace_vars[id].size;
		}
	    }
	}

      if((flags & 0x80) == 0) break;
    }

  junk16 = htons(off);
  memcpy(buf, &junk16, 2);

  return off;

 err:
  return -1;
}

static int warts_pmtud_flags(const scamper_trace_t *trace, uint8_t *flags)
{
  static const int tlv_idx[] = {
    3, /* SCAMPER_TRACE_PMTUD_TLV_OUTMTU */
  };
  scamper_tlv_t *tlv;
  int i, j, id, max_id;

  /* unset all the flags possible */
  memset(flags, 0, pmtud_vars_mfb);

  /* for now, we include the base data items */
  set_flag(flags, 1);
  set_flag(flags, 2);
  max_id = 2;

  /* go through the TLVs and decide which flags to set */
  for(tlv = trace->pmtud->tlvs; tlv != NULL; tlv = tlv->tlv_next)
    {
      id = tlv_idx[tlv->tlv_type-1];
      set_flag(flags, id);
      if(id > max_id) max_id = id;
    }

  /* figure out how many bytes have been used */
  j = max_id / 7;
  if((max_id % 7) != 0) j++;

  /* skip through and set the 'more flags' bit for all flag bytes necessary */
  for(i=0; i<j-1; i++)
    {
      flags[i] |= 0x80;
    }

  return j;
}

static int warts_pmtud_size(const scamper_trace_t *trace)
{
  uint8_t flags[pmtud_vars_mfb];
  int size;
  int i, j;

  /* figure out which flags to set */
  size = warts_pmtud_flags(trace, flags);

  /* first two bytes define the size of the pmtud data stored */
  size += 2;

  /* figure out actually how much data to store */
  for(i=0; i<pmtud_vars_mfb; i++)
    {
      for(j=0; j<7; j++)
	{
	  if((flags[i] & (0x1 << j)) != 0)
	    {
	      size += pmtud_vars[(i*7)+j].size;
	    }
	}

      if((flags[i] & 0x80) == 0) break;
    }

  return size;
}

static int warts_pmtud_read(scamper_trace_t *trace, const uint8_t *buf)
{
  uint8_t  flags;
  uint16_t size;
  uint16_t off;
  uint16_t junk16;
  int      i, j, id;

  /* figure out how much data is in the pmtud header data */
  memcpy(&junk16, buf, 2);
  size = ntohs(junk16);
  off = 2;

  /* determine how many flag bytes to read to get to the actual data */
  for(i=0;;i++)
    {
      flags = buf[2+i];
      if((flags & 0x80) == 0) break;
    }
  off += (i + 1);

  /* read the data items depending on which flags are set */
  for(i=0;;i++)
    {
      flags = buf[2+i];

      for(j=0; j<7; j++)
	{
	  if((flags & (0x1 << i)) != 0)
	    {
	      id = (i*7)+j;

	      switch(id)
		{
		case 0: /* interface mtu */
		  memcpy(&junk16, buf+off, 2);
		  trace->pmtud->ifmtu = ntohs(junk16);
		  break;

		case 1: /* path mtu */
		  memcpy(&junk16, buf+off, 2);
		  trace->pmtud->pmtu = ntohs(junk16);
		  break;

		case 2: /* mtu to gateway */
		  memcpy(&junk16, buf+off, 2); junk16 = ntohs(junk16);
		  scamper_tlv_set(&trace->pmtud->tlvs, pmtud_vars[i].tlv_id,
				  2, &junk16);
		  break;

		default:
		  goto done;
		}

	      if(pmtud_vars[id].size != -1) 
		{
		  off += pmtud_vars[id].size;
		}
	    }
	}

      if((flags & 0x80) == 0) break;
    }

 done:
  return size;
}

static int warts_pmtud_write(const scamper_trace_t *trace, uint8_t *buf)
{
  uint8_t   flags;
  uint16_t  off;
  uint16_t  junk16;
  int       i, j, id;

  /* write the flags into the buffer */
  i = warts_pmtud_flags(trace, buf+2);

  /* figure out where the data items get written */
  off = 2 + i;

  for(i=0;;i++)
    {
      flags = buf[2+i];

      for(j=0; j<7; j++)
	{
	  if((flags & (0x1 << j)) != 0)
	    {
	      id = (i*7)+j;
	      switch(id)
		{
		case 0: /* interface mtu */
		  junk16 = htons(trace->pmtud->ifmtu);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 1: /* path mtu */
		  junk16 = htons(trace->pmtud->pmtu);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 2: /* mtu to gateway */
		  SCAMPER_TRACE_PMTUD_GET_OUTMTU(trace->pmtud, junk16);
		  junk16 = htons(junk16);
		  memcpy(buf+off, &junk16, 2);
		  break;
		}

	      if(pmtud_vars[id].size >= 0)
		{
		  off += pmtud_vars[id].size;
		}
	    }
	}

      if((flags & 0x80) == 0) break;
    }

  /* copy size in */
  junk16 = htons(off);
  memcpy(buf, &junk16, 2);

  return off;
}

static int warts_hop_flags(const scamper_trace_hop_t *hop, uint8_t *flags)
{
  static const int tlv_idx[] = {
    10, /* SCAMPER_TRACE_HOP_TLV_REPLY_IPID */
    11, /* SCAMPER_TRACE_HOP_TLV_REPLY_IPTOS */
    12, /* SCAMPER_TRACE_HOP_TLV_NHMTU */
    13, /* SCAMPER_TRACE_HOP_TLV_INNER_IPLEN */
    14, /* SCAMPER_TRACE_HOP_TLV_INNER_IPTTL */
  };
  scamper_tlv_t *tlv;
  int i, j, id, max_id = 9;

  /* unset all the flags possible */
  memset(flags, 0, hop_vars_mfb);

  /* for now, we include all the base data items */
  for(i=1; i<=9; i++) set_flag(flags, i);

  /* go through the TLVs and decide which flags to set */
  for(tlv = hop->tlvs; tlv != NULL; tlv = tlv->tlv_next)
    {
      assert(tlv->tlv_type-1 < sizeof(tlv_idx)/sizeof(int));
      id = tlv_idx[tlv->tlv_type-1];
      set_flag(flags, id);
      if(id > max_id) max_id = id;
    }

  /* figure out how many bytes have been used */
  j = max_id / 7;
  if((max_id % 7) != 0) j++;

  /* skip through and set the 'more flags' bit for all flag bytes necessary */
  for(i=0; i<j-1; i++)
    {
      flags[i] |= 0x80;
    }

  return j;
}

static int warts_hop_size(const scamper_trace_hop_t *hop)
{
  uint8_t  flags[hop_vars_mfb];
  uint16_t size;
  int i, j;

  /* figure out which flags to set */
  size = warts_hop_flags(hop, flags);

  /* first two bytes define the size of the hop data stored */
  size += 2;

  /* figure out actually how much data to store */
  for(i=0; i<hop_vars_mfb; i++)
    {
      for(j=0; j<7; j++)
	{
	  if((flags[i] & (0x1 << j)) != 0)
	    {
	      size += hop_vars[(i*7)+j].size;
	    }
	}

      if((flags[i] & 0x80) == 0) break;
    }

  return size;
}

static int warts_hop_read(const scamper_trace_t *trace,
			  scamper_trace_hop_t *hop,
			  const uint8_t *buf, const warts_state_t *state)
{
  warts_addr_t *wa;
  uint8_t  flags;
  uint16_t size;
  uint16_t off;
  uint16_t junk16;
  uint32_t junk32;
  int      i, j, id;

  /* figure out how much data is actually in the hop data */
  memcpy(&junk16, buf, 2);
  size = ntohs(junk16);
  off = 2;

  /* determine how many flag bytes to read to get to the actual data */
  for(i=0;;i++)
    {
      flags = buf[2+i];
      if((flags & 0x80) == 0) break;
    }
  off += (i + 1);

  /* read the data items depending on which flags are set */
  for(i=0;;i++)
    {
      flags = buf[2+i];

      for(j=0; j<7; j++)
	{
	  if((flags & (0x1 << j)) != 0)
	    {
	      id = (i*7)+j;

	      switch(id)
		{
		case 0:  /* address id */
		  memcpy(&junk32, buf+off, 4); junk32 = ntohl(junk32);
		  if(junk32 >= state->addr_count) goto err;
		  wa = state->addr_table[junk32];
		  hop->addr = scamper_addr_use(wa->addr);
		  break;

		case 1:  /* probe ttl */
		  hop->probe_ttl = buf[off];
		  break;

		case 2:  /* reply ttl */
		  hop->reply_ttl = buf[off];
		  break;

		case 3:  /* flags */
		  hop->flags = buf[off];
		  break;

		case 4:  /* probe id */
		  hop->probe_id = buf[off];
		  break;

		case 5:  /* round trip time */
		  memcpy(&junk32, buf+off, 4); junk32 = ntohl(junk32);
		  hop->rtt.tv_sec  = junk32 / 1000000;
		  hop->rtt.tv_usec = junk32 % 1000000;
		  break;

		case 6:  /* icmp type / code */
		  hop->icmp_type = buf[off];
		  hop->icmp_code = buf[off+1];
		  break;

		case 7:  /* probe size */
		  memcpy(&junk16, buf+off, 2);
		  hop->probe_size = ntohs(junk16);
		  break;

		case 8:  /* reply size */
		  memcpy(&junk16, buf+off, 2);
		  hop->reply_size = ntohs(junk16);

		case 9:  /* ipid of reply packet */
		case 11: /* next hop mtu in ptb message */
		case 12: /* ip->len from inside icmp */
		  memcpy(&junk16, buf+off, 2); junk16 = ntohs(junk16);
		  if(scamper_tlv_set(&hop->tlvs, hop_vars[id].tlv_id,
				     2, &junk16) == NULL)
		    {
		      goto err;
		    }
		  break;

		case 10: /* tos bits of reply packet */
		case 13: /* ip->ttl from inside icmp */
		  if(scamper_tlv_set(&hop->tlvs, hop_vars[id].tlv_id,
				     1, buf+off) == NULL)
		    {
		      goto err;
		    }
		  break;

		default:
		  goto done;
		}

	      if(hop_vars[id].size >= 0)
		{
		  off += hop_vars[id].size;
		}
	    }
	}

      if((flags & 0x80) == 0) break;
    }

 done:
  return size;

 err:
  return -1;
}

static int warts_hop_write(const scamper_trace_t *trace,
			   const scamper_trace_hop_t *hop,
			   const uint32_t addr_id, uint8_t *buf)
{
  scamper_tlv_t *tlv;
  int      i, j, id;
  uint16_t off;
  uint16_t junk16;
  uint32_t junk32;
  uint8_t  flags;

  /* write the flags into the buffer */
  i = warts_hop_flags(hop, buf+2);

  /* figure out where the data items get written */
  off = 2 + i;

  for(i=0;;i++)
    {
      flags = buf[2+i];

      for(j=0; j<7; j++)
	{
	  if((flags & (0x1 << j)) != 0)
	    {
	      id = (i*7)+j;
	      switch(id)
		{
		case 0: /* address id */
		  junk32 = htonl(addr_id);
		  memcpy(buf+off, &junk32, 4);
		  break;

		case 1: /* probe ttl */
		  buf[off] = hop->probe_ttl;
		  break;

		case 2: /* reply ttl */
		  buf[off] = hop->reply_ttl;
		  break;

		case 3: /* flags */
		  buf[off] = hop->flags;
		  break;

		case 4: /* probe id */
		  buf[off] = hop->probe_id;
		  break;

		case 5: /* round trip time */
		  junk32  = hop->rtt.tv_sec * 1000000;
		  junk32 += hop->rtt.tv_usec;
		  junk32  = htonl(junk32);
		  memcpy(buf+off, &junk32, 4);
		  break;

		case 6: /* icmp type / code */
		  buf[off]   = hop->icmp_type;
		  buf[off+1] = hop->icmp_code;
		  break;

		case 7: /* probe size */
		  junk16 = htons(hop->probe_size);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 8: /* reply size */
		  junk16 = htons(hop->reply_size);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 9: /* ipid of reply packet */
		case 11: /* next hop mtu in ptb message */
		case 12: /* ip->len from inside icmp */
		  tlv = scamper_tlv_get(hop->tlvs, hop_vars[id].tlv_id);
		  if(tlv == NULL) goto err;
		  junk16 = htons(tlv->tlv_val_16);
		  memcpy(buf+off, &junk16, 2);
		  break;

		case 10: /* tos bits of reply packet */
		case 13: /* ip->ttl from inside icmp */
		  tlv = scamper_tlv_get(hop->tlvs, hop_vars[id].tlv_id);
		  if(tlv == NULL) goto err;
		  buf[off] = tlv->tlv_val_8;
		  break;
		}

	      if(hop_vars[id].size >= 0)
		{
		  off += hop_vars[id].size;
		}
	    }
	}

      if((flags & 0x80) == 0) break;
    }

  junk16 = htons(off);
  memcpy(buf, &junk16, 2);

  return off;

 err:
  return -1;
}

static scamper_trace_hop_t *warts_hops_read(const scamper_trace_t *trace,
					    warts_state_t *state,
					    const uint8_t *buf,
					    uint16_t count, int *off)
{
  scamper_trace_hop_t *head = NULL, *hop = NULL;
  int i = 0;

  if(count == 0)
    {
      return NULL;
    }

  while(count-- > 0)
    {
      if(hop != NULL)
	{
	  hop->next = scamper_trace_hop_alloc();
	  hop = hop->next;
	}
      else
	{
	  head = hop = scamper_trace_hop_alloc();
	}

      if(hop == NULL)
	{
	  while(head != NULL)
	    {
	      hop = head;
	      head = head->next;
	      scamper_trace_hop_free(hop);
	    }
	  return NULL;
	}

      assert(state->addr_count > 0);
      i += warts_hop_read(trace, hop, buf+i, state);
    }

  *off += i;

  return head;
}

/*
 * warts_trace_read
 *
 */
static scamper_trace_t *warts_trace_read(const scamper_file_t *sf,
					 const warts_hdr_t *hdr)
{
  scamper_trace_t     *trace = NULL;
  warts_state_t       *state;
  uint8_t             *buf = NULL;
  uint16_t             junk16;
  int                  i, off;
  scamper_trace_hop_t *hops;
  scamper_trace_hop_t *hop;
  uint16_t             count;
  int                  max_ttl;
  uint8_t              type;
  uint16_t             len;

  if((buf = warts_read(sf, hdr)) == NULL)
    {
      goto err;
    }

  if((trace = scamper_trace_alloc()) == NULL)
    {
      goto err;
    }

  state = scamper_file_getstate(sf);
  off = 0;

  /* read the trace's parameters */
  if((i = warts_trace_params_read(trace, state, buf)) == -1)
    {
      goto err;
    }
  off += i;

  /*
   * the next two bytes tell us how many scamper_hops to read out of trace
   * if we did not get any responses, we are done.
   */
  memcpy(&junk16, buf+off, 2); off += 2;
  if((count = ntohs(junk16)) == 0)
    {
      goto done;
    }

  /* read all the hop records */
  if((hops = warts_hops_read(trace, state, buf+off, count, &off)) == NULL)
    {
      goto err;
    }

  /* work out the maximum ttl probed with that got a response */
  max_ttl = 0;
  for(i=0, hop = hops; i < count; i++)
    {
      if(hop->probe_ttl > max_ttl)
	{
	  max_ttl = hop->probe_ttl;
	}
      hop = hop->next;
    }

  /* with that max_ttl, allocate enough hops to string the trace together */
  if(scamper_trace_hops_alloc(trace, max_ttl) == -1)
    {
      free(hops);
      goto err;
    }
  trace->hop_count = max_ttl;

  /*
   * now loop through the hops array stored in this procedure
   * and assemble the responses into trace->hops.
   */
  trace->hops[hops->probe_ttl-1] = hop = hops;
  while(hop->next != NULL) 
    {
      if(hop->probe_ttl != hop->next->probe_ttl)
	{
	  i = hop->next->probe_ttl-1;
	  trace->hops[i] = hop->next;
	  hop->next = NULL;
	  hop = trace->hops[i];
	}
      else hop = hop->next;
    }

  for(;;)
    {
      memcpy(&junk16, buf+off, 2); off += 2;
      if(junk16 == WARTS_TRACE_ATTR_EOF)
	{
	  break;
	}

      junk16 = ntohs(junk16);
      type = WARTS_TRACE_ATTR_HDR_TYPE(junk16);
      len  = WARTS_TRACE_ATTR_HDR_LEN(junk16);

      if(type == WARTS_TRACE_ATTR_PMTUD)
	{
	  if((trace->pmtud=malloc_zero(sizeof(scamper_trace_pmtud_t))) == NULL)
	    {
	      goto err;
	    }

	  if((i = warts_pmtud_read(trace, buf+off)) == -1)
	    {
	      goto err;
	    }

	  i += off;

	  memcpy(&junk16, buf+i, 2); i += 2;
	  if((count = ntohs(junk16)) != 0)
	    {
	      if((hops = warts_hops_read(trace,state,buf+i,count,&i)) == NULL)
		{
		  goto err;
		}

	      trace->pmtud->hops = hops;
	    }
	}

      off += len;
    }

  assert(off == hdr->len);

 done:
  free(buf);
  return trace;

 err:
  if(buf != NULL) free(buf);
  if(trace != NULL) scamper_trace_free(trace);
  return NULL;
}

static int warts_trace_write(const scamper_file_t *sf,
			     const scamper_trace_t *trace)
{
  scamper_trace_hop_t *hop;
  uint8_t             *buf = NULL;
  uint32_t            *hop_idxs = NULL;
  uint32_t            *pmtu_idxs = NULL;
  uint16_t             junk16;
  size_t               size;
  int                  i, j, k;
  int                  hop_recs;
  int                  pmtu_len = 0;
  int                  pmtu_recs = 0;
  uint32_t             addr_id;

  assert(trace != NULL);
  assert(trace->list != NULL);

  /*
   * size of trace parameters, plus a 2 byte field that records how many
   * hop records to read
   */
  size = warts_trace_params_size(trace) + 2;

  /* count the number of records that we have to write to disk */
  hop_recs = 0;
  for(i=0; i<trace->hop_count; i++)
    {
      for(hop = trace->hops[i]; hop != NULL; hop = hop->next)
	{
	  hop_recs++;
	}
    }

  /* for each hop, get a warts_addr_t->id and record it */
  if(hop_recs > 0)
    {
      if((hop_idxs = (uint32_t *)malloc(hop_recs*sizeof(uint32_t))) == NULL)
	{
	  goto err;
	}
      for(i=0, j=0; i<trace->hop_count; i++)
	{
	  for(hop = trace->hops[i]; hop != NULL; hop = hop->next)
	    {
	      if(warts_addr_getid(sf, hop->addr, &addr_id) == -1)
		{
		  goto err;
		}

	      /* for each hop, remember the address id */
	      hop_idxs[j++] = addr_id;

	      /* for each hop, figure out how much space it will take up */
	      size += warts_hop_size(hop);
	    }
	}
    }

  /* figure out how much space we need for PMTU data, if we have it */
  if(trace->pmtud != NULL)
    {
      /* count the number of hop records */
      pmtu_recs = 0;
      for(hop = trace->pmtud->hops; hop != NULL; hop = hop->next)
	{
	  pmtu_recs++;
	}

      /* allocate an array of address indexes for the pmtu hop addresses */
      if((pmtu_idxs = (uint32_t *)malloc(pmtu_recs*sizeof(uint32_t))) == NULL)
	{
	  goto err;
	}

      pmtu_len = warts_pmtud_size(trace) +
                 2;  /* number of hop records */

      for(hop = trace->pmtud->hops, j=0; hop != NULL; hop = hop->next)
	{
	  if(warts_addr_getid(sf, hop->addr, &addr_id) == -1)
	    {
	      goto err;
	    }

	  /* for each hop, remember the address id */
	  pmtu_idxs[j++] = addr_id;

	  pmtu_len += warts_hop_size(hop);
	}

      size += (pmtu_len + 2); /* size of attribute header */
    }

  size += 2; /* EOF */

  if((buf = malloc(size)) == NULL)
    {
      goto err;
    }
  i = 0;

  /* trace parameters */
  if((j = warts_trace_params_write(sf, trace, buf+i)) == -1)
    {
      goto err;
    }
  i += j;

  /* hop record count */
  junk16 = htons(hop_recs);
  memcpy(buf+i, &junk16, 2); i += 2;

  /* write each traceroute hop record */
  if(hop_recs > 0)
    {
      for(j=0, k=0; j<trace->hop_count; j++)
	{
	  for(hop = trace->hops[j]; hop != NULL; hop = hop->next)
	    {
	      i += warts_hop_write(trace, hop, hop_idxs[k++], buf+i);
	    }
	}

      /* don't need the hop address index any longer */
      free(hop_idxs);
      hop_idxs = NULL;
    }

  /* write the PMTU data */
  if(trace->pmtud != NULL)
    {
      junk16 = htons(WARTS_TRACE_ATTR_HDR(WARTS_TRACE_ATTR_PMTUD, pmtu_len));
      memcpy(buf+i, &junk16, 2); i += 2;

      /* write details of the pmtud measurement */
      if((j = warts_pmtud_write(trace, buf+i)) == -1)
	{
	  goto err;
	}
      i += j;

      /* write the number of hop records */
      junk16 = htons(pmtu_recs);
      memcpy(buf+i, &junk16, 2); i += 2;

      for(hop = trace->pmtud->hops, k = 0; hop != NULL; hop = hop->next)
	{
	  if((j = warts_hop_write(trace, hop, pmtu_idxs[k++], buf+i)) == -1)
	    {
	      goto err;
	    }
	  i += j;
	}

      if(pmtu_idxs != NULL)
	{
	  free(pmtu_idxs);
	  pmtu_idxs = NULL;
	}
    }

  /* write the end of trace attributes header */
  junk16 = WARTS_TRACE_ATTR_EOF;
  memcpy(buf+i, &junk16, 2); i += 2;

  assert(i == size);

  if(warts_write(sf, WARTS_TRACE, buf, size) == -1)
    {
      goto err;
    }

  free(buf);
  return 0;

 err:
  if(buf != NULL) free(buf);
  if(hop_idxs != NULL) free(hop_idxs);
  if(pmtu_idxs != NULL) free(hop_idxs);
  return -1;
}

static int warts_read_type(const scamper_file_t *sf, warts_hdr_t *hdr_type)
{
  scamper_cycle_t *cycle;
  warts_hdr_t  hdr;
  int fd;

  fd = scamper_file_getfd(sf);

  for(;;)
    {
      if(warts_hdr_read(sf, &hdr) != 1)
	{
	  return -1;
	}

      if(hdr.magic != WARTS_MAGIC)
	{
	  return -1;
	}

      if(hdr.type == hdr_type->type)
	{
	  break;
	}

      if(hdr.type == WARTS_ADDR)
	{
	  if(warts_addr_read(sf, &hdr) == -1)
	    {
	      return -1;
	    }
	}
      else if(hdr.type == WARTS_LIST)
	{
	  if(warts_list_read(sf, &hdr) == NULL)
	    {
	      return -1;
	    }
	}
      else if(hdr.type == WARTS_CYCLE_NEW)
	{
	  if(warts_cyclenew_read(sf, &hdr) == NULL)
	    {
	      return -1;
	    }
	}
      else if(hdr.type == WARTS_CYCLE_FIN)
	{
	  if((cycle = warts_cyclefin_read(sf, &hdr)) == NULL)
	    {
	      return -1;
	    }

	  scamper_cycle_free(cycle);
	}
      else
	{
	  if(lseek(fd, hdr.len, SEEK_CUR) == -1)
	    {
	      return -1;
	    }
	}
    }

  memcpy(hdr_type, &hdr, sizeof(hdr));
  return 0;
}

scamper_trace_t *scamper_file_warts_read_trace(const scamper_file_t *sf)
{
  warts_hdr_t hdr;

  hdr.type = WARTS_TRACE;
  if(warts_read_type(sf, &hdr) == -1)
    {
      return NULL;
    }

  return warts_trace_read(sf, &hdr);
}

/*
 * scamper_file_warts_write
 *
 * take a trace and write all the bits out to disk.
 */
int scamper_file_warts_write_trace(const scamper_file_t *sf,
				   const scamper_trace_t *trace)
{
  return warts_trace_write(sf, trace);
}

scamper_list_t *scamper_file_warts_read_list(const scamper_file_t *sf)
{
  warts_hdr_t hdr;

  hdr.type = WARTS_LIST;
  if(warts_read_type(sf, &hdr) == -1)
    {
      return NULL;
    }

  return warts_list_read(sf, &hdr);
}

int scamper_file_warts_write_fincycle(const scamper_file_t *sf,
				      const scamper_cycle_t *c)
{
  return warts_cyclefin_write(sf, c);
}

int scamper_file_warts_write_newcycle(const scamper_file_t *sf,
				      const scamper_cycle_t *c)
{
  return warts_cyclenew_write(sf, c);
}

int scamper_file_warts_init_read(scamper_file_t *sf)
{
  warts_state_t *state;

  if((state = (warts_state_t *)malloc_zero(sizeof(warts_state_t))) == NULL)
    {
      return -1;
    }

  scamper_file_setstate(sf, state);
  return 0;
}

/*
 * scamper_file_warts_init_write
 *
 * get the scamper_file_t object ready to write warts objects and keep state
 */
int scamper_file_warts_init_write(scamper_file_t *sf)
{
  warts_state_t *state;

  if((state = (warts_state_t *)malloc_zero(sizeof(warts_state_t))) == NULL)
    {
      return -1;
    }

  if((state->addr_tree = splaytree_create(warts_addr_cmp)) == NULL)
    {
      free(state);
      return -1;
    }

  scamper_file_setstate(sf, state);

  return 0;
}

/*
 * scamper_file_warts_init_append
 *
 * go through the file and form the address dictionary.
 */
int scamper_file_warts_init_append(scamper_file_t *sf)
{
  warts_state_t *state;
  warts_hdr_t    hdr;
  int            i, j, fd;

  /* init the warts structures as if we were reading the file */
  if(scamper_file_warts_init_read(sf) == -1)
    {
      return -1;
    }

  fd = scamper_file_getfd(sf);

  for(;;)
    {
      if((i = warts_hdr_read(sf, &hdr)) == -1)
	{
	  return -1;
	}
      else if(i == 0) break;

      if(hdr.magic != WARTS_MAGIC)
	{
	  return -1;
	}

      if(hdr.type == WARTS_ADDR)
	{
	  if(warts_addr_read(sf, &hdr) == -1)
	    {
	      return -1;
	    }
	}
      else
	{
	  if(lseek(fd, hdr.len, SEEK_CUR) == -1)
	    {
	      return -1;
	    }
	}
    }

  /*
   * all the addresses are in a table.  put them into a splay tree so we can
   * find them quickly, and then trash the address table
   */
  state = scamper_file_getstate(sf);
  if((state->addr_tree = splaytree_create(warts_addr_cmp)) == NULL)
    {
      return -1;
    }

  /* put them in the splay tree */
  for(i=0; i<state->addr_count; i++)
    {
      if((j = splaytree_insert(state->addr_tree, state->addr_table[i])) == -1)
	{
	  return -1;
	}
      else if(j == 0)
	{
	  return -1;
	}
    }

  /* free up the address table */
  free(state->addr_table);
  state->addr_table = NULL;

  return 0;
}

int scamper_file_warts_is(const scamper_file_t *sf)
{
  uint16_t magic16;
  int fd;

  fd = scamper_file_getfd(sf);

  if(lseek(fd, 0, SEEK_SET) == -1)
    {
      return 0;
    }

  if(read_wrap(fd, &magic16, NULL, sizeof(magic16)) != 0)
    {
      return 0;
    }

  magic16 = ntohs(magic16);

  if(magic16 == WARTS_MAGIC)
    {
      if(lseek(fd, 0, SEEK_SET) == -1)
	{
	  return 0;
	}
      return 1;
    }

  return 0;
}

void scamper_file_warts_close(scamper_file_t *sf)
{
  scamper_list_t *list;
  warts_state_t *state;
  int i;

  if((state = scamper_file_getstate(sf)) == NULL)
    {
      return;
    }

  if(state->list_table != NULL)
    {
      for(i=0; i<state->list_count; i++)
	{
	  list = state->list_table[i];
	  free(list->internal);
	  list->internal = NULL;
	  scamper_list_free(list);
	}
      free(state->list_table);
    }

  if(state->addr_table != NULL)
    {
      for(i=0; i<state->addr_count; i++)
	{
	  warts_addr_free(state->addr_table[i]);
	}
      free(state->addr_table);
    }

  if(state->addr_tree != NULL)
    {
      splaytree_destroy(state->addr_tree, (splaytree_free_t)warts_addr_free);
    }

  free(state);

  if((i = scamper_file_getfd(sf)) != -1)
    {
      close(i);
    }

  return;
}
