/*
 * scamper_icmp6.c
 *
 * $Id: scamper_icmp6.c,v 1.41 2005/05/26 22:32:32 mjl Exp $
 *
 * Copyright (C) 2003-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if defined(__FreeBSD__) && __FreeBSD_version < 430000
#include <sys/param.h>
#endif

#if defined(__sun__)
#define _XPG4_2
#define __EXTENSIONS__
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <netinet/in.h>

#if defined(__linux__)
#define __FAVOR_BSD
#include <sys/ioctl.h>
#endif

#include <netinet/in_systm.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <assert.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

/*
 * if the system has the RFC 2292 IPV6_HOPLIMIT sockopt, then make it
 * behave like the RFC 3542 IPV6_RECVHOPLIMIT sockopt
 */
#if !defined(IPV6_RECVHOPLIMIT) && defined(IPV6_HOPLIMIT)
#define IPV6_RECVHOPLIMIT IPV6_HOPLIMIT
#endif

#include "scamper.h"
#include "scamper_addr.h"
#include "scamper_dl.h"
#include "scamper_probe.h"
#include "scamper_icmp_resp.h"

#include "scamper_debug.h"
#include "utils.h"

/*
 * if the [linux] system has SO_TIMESTAMP, then do not use SIOCGSTAMP, as
 * that requires an extra system call
 */
#if defined(SO_TIMESTAMP)
#undef SIOCGSTAMP
#endif

static int      icmp6 = -1;
static uint8_t *pktbuf = NULL;
static size_t   pktbuf_len = 0;

void scamper_icmp6_build(scamper_probe_t *probe, uint8_t *pktbuf)
{
  struct ip6_hdr   *ip6;
  struct icmp6_hdr *icmp;

  ip6 = (struct ip6_hdr *)pktbuf;
  ip6->ip6_flow = 0;
  ip6->ip6_vfc  = 0x60;
  ip6->ip6_plen = htons(8 + probe->pr_len);
  ip6->ip6_nxt  = IPPROTO_ICMPV6;
  ip6->ip6_hlim = probe->pr_ttl;
  memcpy(&ip6->ip6_src, probe->pr_src->addr, 16);
  memcpy(&ip6->ip6_dst, probe->pr_dst->addr, 16);

  icmp = (struct icmp6_hdr *)pktbuf;
  icmp->icmp6_type  = probe->pr_icmp_type;
  icmp->icmp6_code  = probe->pr_icmp_code;
  icmp->icmp6_cksum = 0;
  icmp->icmp6_id    = htons(probe->pr_icmp_id);
  icmp->icmp6_seq   = htons(probe->pr_icmp_seq);

  /* if there is data to include in the payload, copy it in now */
  if(probe->pr_len > 0)
    {
      memcpy(pktbuf + 8, probe->pr_data, probe->pr_len);
    }

  icmp->icmp6_cksum = in_cksum(icmp, 8 + probe->pr_len);

  return;
}

int scamper_icmp6_probe(scamper_probe_t *probe)
{
  struct sockaddr_in6  sin6;
  struct icmp6_hdr    *icmp;
  char                 addr[128];
  void                *dst;
  size_t               len;
  int                  i, icmphdrlen;

  assert(probe != NULL);
  assert(probe->pr_proto == IPPROTO_ICMPV6);
  assert(probe->pr_dst != NULL);
  assert(probe->pr_src != NULL);
  assert(probe->pr_len > 0 || probe->pr_data == NULL);

  switch(probe->pr_icmp_type)
    {
    case ICMP6_ECHO_REQUEST:
      icmphdrlen = (1 + 1 + 2 + 2 + 2);
      break;

    default:
      probe->pr_errno = EINVAL;
      return -1;
    }

  icmphdrlen = (1 + 1 + 2 + 2 + 2);
  len = probe->pr_len + icmphdrlen;
  dst = probe->pr_dst->addr;

  i = probe->pr_ttl;
  if(setsockopt(icmp6, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &i, sizeof(i)) == -1)
    {
      printerror(errno, strerror, __func__, "could not set hlim to %d", i);
      return -1;
    }

  i = len;
  if(setsockopt(icmp6, SOL_SOCKET, SO_SNDBUF, (char *)&i, sizeof(i)) == -1)
    {
      printerror(errno, strerror, __func__,
		 "could not set buffer to %d bytes", i);
      return -1;
    }

  if(pktbuf_len < len)
    {
      pktbuf = realloc_wrap(pktbuf, len, __func__, "could not realloc");
      pktbuf_len = len;
    }

  icmp = (struct icmp6_hdr *)pktbuf;
  icmp->icmp6_type  = probe->pr_icmp_type;
  icmp->icmp6_code  = probe->pr_icmp_code;
  icmp->icmp6_cksum = 0;
  icmp->icmp6_id    = htons(probe->pr_icmp_id);
  icmp->icmp6_seq   = htons(probe->pr_icmp_seq);

  /* if there is data to include in the payload, copy it in now */
  if(probe->pr_len > 0)
    {
      memcpy(pktbuf + icmphdrlen, probe->pr_data, probe->pr_len);
    }

  icmp->icmp6_cksum = in_cksum(icmp, icmphdrlen + probe->pr_len);

  sockaddr_compose((struct sockaddr *)&sin6, AF_INET6, dst, 0);

  /* get the transmit time immediately before we send the packet */
  gettimeofday_wrap(&probe->pr_tx);

  i = sendto(icmp6, pktbuf, len, 0, (struct sockaddr *)&sin6,
	     sizeof(struct sockaddr_in6));

  /* if we sent the probe successfully, there is nothing more to do here */
  if(i == len)
    {
      return 0;
    }

  probe->pr_errno = errno;

  /* error condition, could not send the packet at all */
  if(i == -1)
    {
      printerror(probe->pr_errno, strerror, __func__,
		 "could not send to %s (%d ttl, %d seq, %d len)",
		 inet_ntop(AF_INET6, dst, addr, sizeof(addr)),
		 probe->pr_ttl, probe->pr_icmp_seq, len);
    }
  /* error condition, sent a portion of the probe */
  else
    {
      fprintf(stderr,
	      "scamper_icmp6_probe: sent %d bytes of %d byte packet to %s",
	      i, (int)len,
	      inet_ntop(AF_INET6, dst, addr, sizeof(addr)));
    }

  return -1;
}

/*
 * icmp6_recv_ip_outer
 *
 * copy the outer-details of the ICMP6 message into the response structure.
 * get details of when the packet was received.
 */
static void icmp6_recv_ip_outer(scamper_icmp_resp_t *resp,
				struct msghdr *msg,
				struct icmp6_hdr *icmp,
				struct sockaddr_in6 *from,
				size_t size)
{
  int16_t hlim = -1;

#if defined(IPV6_RECVHOPLIMIT) || defined(SO_TIMESTAMP)
  /* get the HLIM field of the ICMP6 packet returned */
  struct cmsghdr *cmsg = (struct cmsghdr *)CMSG_FIRSTHDR(msg);
  while(cmsg != NULL)
    {
#if defined(IPV6_RECVHOPLIMIT)
      if(cmsg->cmsg_level==IPPROTO_IPV6 && cmsg->cmsg_type==IPV6_RECVHOPLIMIT)
	{
	  memcpy(&hlim, CMSG_DATA(cmsg), sizeof(hlim));
	}
#endif

#if defined(SO_TIMESTAMP)
      if(cmsg->cmsg_level==SOL_SOCKET && cmsg->cmsg_type==SCM_TIMESTAMP)
	{
	  timeval_cpy(&resp->ir_rx, (struct timeval *)CMSG_DATA(cmsg));
	  resp->ir_flags |= SCAMPER_ICMP_RESP_FLAG_KERNRX;
	}
#endif
      cmsg = (struct cmsghdr *)CMSG_NXTHDR(msg, cmsg);
    }
#endif

#if defined(__linux__) && !defined(SO_TIMESTAMP) && defined(SIOCGSTAMP)
  if(ioctl(icmp6, SIOCGSTAMP, &resp->ir_rx) != -1)
    {
      resp->ir_flags |= SCAMPER_ICMP_RESP_FLAG_KERNRX;
    }
#else
  gettimeofday_wrap(&resp->ir_rx);
#endif

  memcpy(&resp->ir_ip_src.v6, &from->sin6_addr, sizeof(struct in6_addr));

  resp->ir_af        = AF_INET6;
  resp->ir_icmp_type = icmp->icmp6_type;
  resp->ir_icmp_code = icmp->icmp6_code;
  resp->ir_ip_hlim   = hlim;
  resp->ir_ip_size   = size;

  return;
}

/*
 * scamper_icmp6_recv
 *
 * handle receiving an ICMPv6 packet.
 *
 * if the packet is an ICMP response that we should concern ourselves with
 * (i.e. it is in response to one of our UDP probes) then we fill out
 * the attached icmp_response structure and return zero.
 *
 * if we should ignore this packet, or an error condition occurs, then
 * we return -1.
 */
int scamper_icmp6_recv(scamper_icmp_resp_t *resp)
{
  struct sockaddr_in6  from;
  u_char               pktbuf[65535];
  size_t               pktbuflen;
  u_char               ctrlbuf[256];
  int                  pktoffset;
  struct icmp6_hdr    *icmp;
  struct ip6_hdr      *ip;
  struct udphdr       *udp;
  uint8_t              type, code;
  uint8_t              nh;
  struct msghdr        msg;
  struct iovec         iov;

  memset(&iov, 0, sizeof(iov));
  iov.iov_base = (caddr_t)pktbuf;
  iov.iov_len  = sizeof(pktbuf);

  msg.msg_name       = (caddr_t)&from;
  msg.msg_namelen    = sizeof(from);
  msg.msg_iov        = &iov;
  msg.msg_iovlen     = 1;
  msg.msg_control    = (caddr_t)ctrlbuf;
  msg.msg_controllen = sizeof(ctrlbuf);

  if((pktbuflen = recvmsg(icmp6, &msg, 0)) == -1)
    {
      printerror(errno, strerror, __func__, "could not recvmsg");
      return -1;
    }
  
  icmp = (struct icmp6_hdr *)pktbuf;
  if(pktbuflen < sizeof(struct icmp6_hdr))
    {
      return -1; 
    }

  type = icmp->icmp6_type;
  code = icmp->icmp6_code;

  /* check to see if the ICMP type / code is what we want */ 
  if((type != ICMP6_TIME_EXCEEDED || code != ICMP6_TIME_EXCEED_TRANSIT) && 
      type != ICMP6_DST_UNREACH && type != ICMP6_PACKET_TOO_BIG &&
      type != ICMP6_ECHO_REPLY)
    {
      scamper_debug(__func__,"ICMP6 type %d / code %d not wanted", type, code);
      return -1;
    }

  pktoffset  = sizeof(struct icmp6_hdr);
  ip         = (struct ip6_hdr *)(pktbuf + pktoffset);

  if(type == ICMP6_ECHO_REPLY)
    {
      resp->ir_icmp_id  = ntohs(icmp->icmp6_id);
      resp->ir_icmp_seq = ntohs(icmp->icmp6_seq);
      memcpy(&resp->ir_inner_ip_dst.v6, &from.sin6_addr,
	     sizeof(struct in6_addr));

      icmp6_recv_ip_outer(resp, &msg, icmp, &from, 
			  pktbuflen + sizeof(struct ip6_hdr));

      scamper_icmp_resp_print(resp);
      return 0;
    }

  nh         = ip->ip6_nxt;
  pktoffset += sizeof(struct ip6_hdr);

  /* search for a UDP header in this packet */
  while(pktoffset + sizeof(struct udphdr) <= pktbuflen)
    {
      /* we cannot read the contents of this packet, or there is no next hdr */
      if(nh == IPPROTO_ESP || nh == IPPROTO_TCP)
        {
	  break;
        }
      else if(nh == IPPROTO_FRAGMENT)
        {
          nh         = ((struct ip6_frag *)(pktbuf+pktoffset))->ip6f_nxt;
          pktoffset += sizeof(struct ip6_frag);
	  continue;
        }
      else if(nh == IPPROTO_AH)
        {
	  /*
	   * the len field in the IPSec Authentication Header specifies the
	   * length of the payload, not including the header, in 32 bit words.
	   * the minimum size of an IPSec authentication header is 2 32 bit
	   * words.
	   */
	  nh         = pktbuf[pktoffset];
	  pktoffset += (pktbuf[pktoffset+1] + 2) << 2;
	  continue;
	}

      if(nh != IPPROTO_UDP && nh != IPPROTO_ICMPV6)
        {
	  /*
	   * IPv6 headers have the length of the particular header encoded
	   * in the header.  The general case of the len field is that it
	   * specifies the length of the header in 8 byte units, not including
	   * the first 8 bytes.
	   * we add 1 to the length field supplied and then shift the bits 3
	   * over to effect a multiplication by 8, as was shown in traceroute6
	   */
          nh         =  pktbuf[pktoffset];
          pktoffset += (pktbuf[pktoffset+1] + 1)<<3;
	  continue;
	}

      resp->ir_flags |= SCAMPER_ICMP_RESP_FLAG_INNER_IP;

      /* record details of the IP header and the ICMP headers */
      icmp6_recv_ip_outer(resp, &msg, icmp, &from,
			  pktbuflen + sizeof(struct ip6_hdr));

      memcpy(&resp->ir_inner_ip_dst.v6, &ip->ip6_dst, sizeof(struct in6_addr));
      resp->ir_inner_ip_proto = nh;
      resp->ir_inner_ip_hlim  = ip->ip6_hlim;
      resp->ir_inner_ip_size  = ntohs(ip->ip6_plen) + sizeof(struct ip6_hdr);

      if(type == ICMP6_PACKET_TOO_BIG)
	{
	  resp->ir_icmp_nhmtu = (ntohl(icmp->icmp6_mtu) % 0xffff);
	}

      if(nh == IPPROTO_UDP)
	{
          udp = (struct udphdr *)(pktbuf+pktoffset);
	  resp->ir_inner_udp_sport = ntohs(udp->uh_sport);
	  resp->ir_inner_udp_dport = ntohs(udp->uh_dport);
	}
      else if(nh == IPPROTO_ICMPV6)
	{
	  icmp = (struct icmp6_hdr *)(pktbuf+pktoffset);
	  resp->ir_inner_icmp_type = icmp->icmp6_type;
	  resp->ir_inner_icmp_code = icmp->icmp6_code;
	  resp->ir_inner_icmp_id   = ntohs(icmp->icmp6_id);
	  resp->ir_inner_icmp_seq  = ntohs(icmp->icmp6_seq);
	}

      scamper_icmp_resp_print(resp);

      return 0;
    }

  return -1;
}

int scamper_icmp6_close()
{
  if(icmp6 != -1)
    {
      if(close(icmp6) == -1)
	{
	  printerror(errno, strerror, __func__, "close failed");
	  return -1;
	}

      icmp6 = -1;
    }

  if(pktbuf != NULL)
    {
      free(pktbuf);
      pktbuf = NULL;
    }

  return 0;
}

int scamper_icmp6_open()
{
#if defined(ICMP6_FILTER)
  struct icmp6_filter filter;
#endif

  int opt;

  if((icmp6 = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) == -1)
    {
      printerror(errno, strerror, __func__, "could not open ICMP socket");
      return -1;
    }

  opt = 65535 + 128;
  if(setsockopt(icmp6, SOL_SOCKET, SO_RCVBUF, &opt, sizeof(opt)) == -1)
    {
      printerror(errno, strerror, __func__, "could not SO_RCVBUF");
      return -1;
    }

#if defined(SO_TIMESTAMP)
  opt = 1;
  if(setsockopt(icmp6, SOL_SOCKET, SO_TIMESTAMP, &opt, sizeof(opt)) == -1)
    {
      printerror(errno, strerror, __func__, "could not set SO_TIMESTAMP");
      return -1;
    }
#endif

#if defined(ICMP6_FILTER)
  /*
   * if the operating system has filtering capabilities for the ICMP6
   * raw socket, then install a filter that passes the three ICMP message
   * types that scamper cares about / processes.
   */
  ICMP6_FILTER_SETBLOCKALL(&filter);
  ICMP6_FILTER_SETPASS(ICMP6_DST_UNREACH, &filter);
  ICMP6_FILTER_SETPASS(ICMP6_PACKET_TOO_BIG, &filter);
  ICMP6_FILTER_SETPASS(ICMP6_TIME_EXCEEDED, &filter);
  ICMP6_FILTER_SETPASS(ICMP6_ECHO_REPLY, &filter);
  if(setsockopt(icmp6,IPPROTO_ICMPV6,ICMP6_FILTER,&filter,sizeof(filter))== -1)
    {
      printerror(errno, strerror, __func__, "could not IPV6_FILTER");
      return -1;
    }
#endif

#if defined(IPV6_DONTFRAG)
  opt = 1;
  if(setsockopt(icmp6, IPPROTO_IPV6, IPV6_DONTFRAG, &opt, sizeof(opt)) == -1)
    {
      printerror(errno, strerror, __func__, "could not set IPV6_DONTFRAG");
      return -1;
    }
#endif

  /*
   * ask the icmp6 socket to supply the TTL of any packet it receives
   * so that scamper might be able to infer the length of the reverse path
   */
#if defined(IPV6_RECVHOPLIMIT)
  opt = 1;
  if(setsockopt(icmp6, IPPROTO_IPV6,IPV6_RECVHOPLIMIT, &opt,sizeof(opt)) == -1)
    {
      printerror(errno, strerror, __func__, "could not set IPV6_RECVHOPLIMIT");
    }
#endif

  return icmp6;
}
