/*
 * scamper_ping.h
 *
 * $Id: scamper_ping.h,v 1.6 2005/05/03 08:44:00 mjl Exp $
 *
 * Copyright (C) 2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __SCAMPER_PING_H
#define __SCAMPER_PING_H

#define SCAMPER_PING_REPLY_IS_ICMP_ECHO_REPLY(reply) ( \
 (reply->addr->type == SCAMPER_ADDR_TYPE_IPV4 && reply->icmp_type == 0) || \
 (reply->addr->type == SCAMPER_ADDR_TYPE_IPV6 && reply->icmp_type == 129))

#define SCAMPER_PING_REPLY_IS_ICMP_UNREACH(reply) ( \
 (reply->addr->type == SCAMPER_ADDR_TYPE_IPV4 && reply->icmp_type == 3) || \
 (reply->addr->type == SCAMPER_ADDR_TYPE_IPV6 && reply->icmp_type == 1))

#define SCAMPER_PING_REPLY_IS_ICMP_TTL_EXP(reply) ( \
 (reply->addr->af == SCAMPER_ADDR_TYPE_IPV4 && reply->icmp_type == 11) || \
 (reply->addr->af == SCAMPER_ADDR_TYPE_IPV6 && reply->icmp_type == 3))

#define SCAMPER_PING_STOP_NONE      0x00 /* null reason */
#define SCAMPER_PING_STOP_COMPLETED 0x01 /* sent all probes */
#define SCAMPER_PING_STOP_ERROR     0x02 /* error occured during ping */

/*
 * scamper_ping_reply
 *
 * a ping reply structure keeps track of how a ping packet was responded to.
 * the default structure has enough fields for interesting pieces out of an
 * echo reply packet.
 *
 * if the icmp type/code is not an ICMP echo reply packet, then the TLVs
 * defined above may be present in the response.
 */
typedef struct scamper_ping_reply
{
  /* where the response came from */
  scamper_addr_t            *addr;

  /* the icmp type / code returned */
  uint8_t                    icmp_type;
  uint8_t                    icmp_code;

  /* the TTL / size of the packet that is returned */
  int16_t                    reply_ttl;
  uint16_t                   reply_size;

  /* the time elapsed between sending the probe and getting this response */
  struct timeval             rtt;

  /* if a single probe gets more than one response, they get chained */
  struct scamper_ping_reply *next;

} scamper_ping_reply_t;

/*
 * scamper_ping
 *
 * this structure contains details of a ping between a source and a
 * destination.  is specifies the parameters to the ping and the
 * replies themselves.
 */
typedef struct scamper_ping
{
  /* source and destination addresses of the ping */
  scamper_addr_t        *src;
  scamper_addr_t        *dst;

  /* the list / cycle that this ping is in relation to */
  scamper_list_t        *list;
  scamper_cycle_t       *cycle;

  /* why the ping finished */
  uint8_t                stop_reason;
  uint8_t                stop_data;

  /* when the ping started */
  struct timeval         start;

  /* the pattern to use inside of a probe.  if null then all zeros */
  uint8_t               *pad_bytes;
  uint16_t               pad_len;

  /* ping options */
  uint16_t               count; /* -c option to ping */
  uint16_t               size;  /* -s option to ping */
  uint8_t                wait;  /* -i option to ping */
  uint8_t                ttl;   /* -m option to ping */

  /* actual data collected with the ping */
  scamper_ping_reply_t **ping_replies;
  uint16_t               ping_sent;
} scamper_ping_t;

/* basic routines to allocate and free scamper_ping structures */
scamper_ping_t       *scamper_ping_alloc(void);
void                  scamper_ping_free(scamper_ping_t *ping);

/* basic routines to allocate and free scamper_ping_reply structures */
scamper_ping_reply_t *scamper_ping_reply_alloc(void);
void                  scamper_ping_reply_free(scamper_ping_reply_t *reply);

#endif /* __SCAMPER_PING_H */
