/*
 * scamper_tcp4.c
 *
 * $Id: scamper_tcp4.c,v 1.11 2005/05/26 09:01:40 mjl Exp $
 *
 * Copyright (C) 2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#if defined(__linux__)
#define __FAVOR_BSD
#endif

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include <arpa/inet.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <assert.h>

#include "scamper_addr.h"
#include "scamper_dl.h"
#include "scamper_probe.h"
#include "scamper_tcp4.h"
#include "scamper_debug.h"
#include "utils.h"

static int      tcp4 = -1;
static uint8_t *pktbuf = NULL;
static size_t   pktbuf_len = 0;

static void tcp_cksum(struct ip *ip, struct tcphdr *tcp, size_t len)
{
  uint16_t *w, ulen;
  uint8_t psuedo[12];
  int sum = 0;
  int i;

  /*
   * the TCP checksum includes a checksum calculated over a psuedo header
   * that includes the src and dst IP addresses, the protocol type, and
   * the TCP length.
   */
  ulen = htons(len);
  memcpy(&psuedo[0],  &ip->ip_src, 4);
  memcpy(&psuedo[4],  &ip->ip_dst, 4);
  memcpy(&psuedo[10], &ulen, 2);
  psuedo[8] = 0;
  psuedo[9] = IPPROTO_TCP;

  /* compute the checksum over the psuedo header */
  w = (uint16_t *)psuedo;
  for(i=6; i>0; i--)
    {
      sum += *w++;
    }

  /* compute the checksum over the body of the TCP message */
  w = (uint16_t *)tcp;
  while(len > 1)
    {
      sum += *w++;
      len -= 2;
    }

  if(len != 0)
    {
      sum += ((uint8_t *)w)[0];
    }

  /* fold the checksum */
  sum  = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  if((tcp->th_sum = ~sum) == 0)
    {
      tcp->th_sum = 0xffff;
    }

  return;
}

int scamper_tcp4_probe(scamper_probe_t *probe)
{
  struct sockaddr_in  sin4;
  struct tcphdr      *tcp;
  struct ip          *ip;
  uint16_t            dport;
  size_t              len;
  int                 i, hdrlen;
  void               *dst;
  char                addr[128];

  assert(probe != NULL);
  assert(probe->pr_proto == IPPROTO_TCP);
  assert(probe->pr_dst != NULL);
  assert(probe->pr_src != NULL);
  assert(probe->pr_len > 0 || probe->pr_data == NULL);

  /* take a copies of these variables, it makes this function easier */
  dport  = probe->pr_tcp_dport;
  dst    = probe->pr_dst->addr;

  /* compute various lengths, for sake of readability */
  hdrlen = sizeof(struct ip) + sizeof(struct tcphdr);
  len    = hdrlen;

  i = hdrlen;
  if(setsockopt(tcp4, SOL_SOCKET, SO_SNDBUF, (char *)&i, sizeof(i)) == -1)
    {
      printerror(errno, strerror, __func__,
		 "could not set buffer to %d bytes", i);
      return -1;
    }

  if(pktbuf_len < len)
    {
      pktbuf = realloc_wrap(pktbuf, len, __func__, "could not realloc");
      pktbuf_len = len;
    }

  ip = (struct ip *)pktbuf;
  ip->ip_v = 4;
  ip->ip_hl = 5;
  ip->ip_tos = 0;
  ip->ip_len = len;
  ip->ip_id = 0;
  ip->ip_off = IP_DF;
  ip->ip_ttl = probe->pr_ttl;
  ip->ip_p = IPPROTO_TCP;
  ip->ip_sum = 0;
  memcpy(&ip->ip_src, probe->pr_src->addr, sizeof(ip->ip_src));
  memcpy(&ip->ip_dst, dst, sizeof(ip->ip_dst));

#if defined(__linux__) || (defined(__OpenBSD__) && OpenBSD >= 199706)
  ip->ip_len = htons(ip->ip_len);
  ip->ip_off = htons(ip->ip_off);
#endif

  tcp = (struct tcphdr *)(pktbuf + sizeof(struct ip));
  tcp->th_sport = htons(probe->pr_tcp_sport);
  tcp->th_dport = htons(dport);
  tcp->th_seq   = htons(probe->pr_tcp_seq);
  tcp->th_ack   = 0;
  tcp->th_x2    = 0;
  tcp->th_off   = (sizeof(struct tcphdr) >> 2);
  tcp->th_flags = TH_SYN;
  tcp->th_win   = 0;
  tcp->th_sum   = 0;
  tcp->th_urp   = 0;

  /* if there is data to include in the payload, copy it in now */
#if 0
  if(probe->pr_len > 0)
    {
      memcpy(pktbuf + hdrlen, probe->pr_data, probe->pr_len);
    }
#endif

  /* compute the checksum over the tcp portion of the probe */
  tcp_cksum(ip, tcp, sizeof(struct tcphdr)); /* + probe->pr_len); */

  sockaddr_compose((struct sockaddr *)&sin4, AF_INET, dst, dport);

  /* get the transmit time immediately before we send the packet */
  gettimeofday_wrap(&probe->pr_tx);

  i = sendto(tcp4, pktbuf, len, 0, (struct sockaddr *)&sin4,
	     sizeof(struct sockaddr_in));

  /* if we sent the probe successfully, there is nothing more to do here */
  if(i == len)
    {
      return 0;
    }

  probe->pr_errno = errno;

  /* error condition, could not send the packet at all */
  if(i == -1)
    {
      printerror(probe->pr_errno, strerror, __func__,
		 "could not send to %s (%d ttl, %d dport, %d len)",
		 inet_ntop(AF_INET, dst, addr, sizeof(addr)),
		 probe->pr_ttl, dport, len);
    }
  /* error condition, sent a portion of the probe */
  else
    {
      fprintf(stderr,
	      "scamper_tcp4_probe: sent %d bytes of %d byte packet to %s",
	      i, (int)len,
	      inet_ntop(AF_INET, dst, addr, sizeof(addr)));
    }

  return -1;
}

int scamper_tcp4_close()
{
  if(tcp4 != -1)
    {
      if(close(tcp4) == -1)
	{
	  printerror(errno, strerror, __func__, "close failed");
	  return -1;
	}

      tcp4 = -1;
    }

  if(pktbuf != NULL)
    {
      free(pktbuf);
      pktbuf = NULL;
    }

  return 0;
}

int scamper_tcp4_open(int tcp_sport)
{
  struct sockaddr_in sin4;
  int on;

  if((tcp4 = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
    {
      printerror(errno, strerror, __func__, "could not open socket");
      return -1;
    }

  on = 1;
  if(setsockopt(tcp4, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on)) == -1)
    {
      printerror(errno, strerror, __func__, "could not IP_HDRINCL");
      return -1;
    }

  sockaddr_compose((struct sockaddr *)&sin4, AF_INET, NULL, tcp_sport);
  if(bind(tcp4, (struct sockaddr *)&sin4, sizeof(sin4)) == -1)
    {
      printerror(errno, strerror, __func__, "could not bind port %d",
		 tcp_sport);
      return -1;
    }

  return 0;
}
