/*
 * scamper_trace.c
 *
 * $Id: scamper_trace.c,v 1.52 2005/05/03 08:44:01 mjl Exp $
 *
 * Copyright (C) 2003-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/udp.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#if defined(__APPLE__)
#include <stdint.h>
#endif

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "utils.h"

int scamper_trace_hops_alloc(scamper_trace_t *trace, const int hops)
{
  size_t size;

  size = sizeof(scamper_trace_hop_t *) * hops;
  if((trace->hops = (scamper_trace_hop_t **)malloc_zero(size)) != NULL)
    {
      return 0;
    }

  return -1;
}

void scamper_trace_hop_free(scamper_trace_hop_t *hop)
{
  if(hop != NULL)
    {
      scamper_tlv_free(hop->tlvs);
      scamper_addr_free(hop->addr);
      free(hop);
    }
  return;
}

scamper_trace_hop_t *scamper_trace_hop_alloc()
{
  scamper_trace_hop_t *hop;

  if((hop = malloc_zero(sizeof(struct scamper_trace_hop))) == NULL)
    {
      return NULL;
    }

  return hop;
}

/*
 * scamper_trace_probe_headerlen
 *
 * return the length of headers sent on probe packets with this trace
 */
int scamper_trace_probe_headerlen(const scamper_trace_t *trace)
{
  int len;

  switch(trace->dst->type)
    {
    case SCAMPER_ADDR_TYPE_IPV4:
      len = sizeof(struct ip);
      break;

    case SCAMPER_ADDR_TYPE_IPV6:
      len = sizeof(struct ip6_hdr);
      break;

    default:
      return -1;
    }

  switch(trace->type)
    {
    case SCAMPER_TRACE_TYPE_UDP:
      len += sizeof(struct udphdr);
      break;

    case SCAMPER_TRACE_TYPE_ICMP_ECHO:
      len += (1 + 1 + 2 + 2 + 2);
      break;

    default:
      return -1;
    }

  return len;
}

/*
 * scamper_trace_free
 *
 */
void scamper_trace_free(scamper_trace_t *trace)
{
  scamper_trace_hop_t *hop, *hop_next;
  int i;

  if(trace == NULL) return;

  if(trace->hops != NULL)
    {
      for(i=0; i<trace->hop_count; i++)
	{
	  hop = trace->hops[i];
	  while(hop != NULL)
	    {
	      hop_next = hop->next;
	      scamper_trace_hop_free(hop);
	      hop = hop_next;
	    }
	}
      free(trace->hops);
    }

  if(trace->pmtud != NULL)
    {
      hop = trace->pmtud->hops;
      while(hop != NULL)
	{
	  hop_next = hop->next;
	  scamper_trace_hop_free(hop);
	  hop = hop_next;
	}

      free(trace->pmtud);
    }

  if(trace->tlvs != NULL) scamper_tlv_free(trace->tlvs);

  if(trace->dst != NULL) scamper_addr_free(trace->dst);
  if(trace->src != NULL) scamper_addr_free(trace->src);

  if(trace->cycle != NULL) scamper_cycle_free(trace->cycle);
  if(trace->list != NULL) scamper_list_free(trace->list);

  free(trace);
  return;
}

/*
 * scamper_trace_alloc
 *
 * allocate the trace and all the possibly necessary data fields
 */
scamper_trace_t *scamper_trace_alloc()
{
  return (struct scamper_trace *)malloc_zero(sizeof(struct scamper_trace));
}
