/*
 * scamper_udp4.c
 *
 * $Id: scamper_udp4.c,v 1.34 2005/05/28 00:33:17 mjl Exp $
 *
 * Copyright (C) 2003-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if defined(__APPLE__)
#include <stdint.h>
#endif

#if defined(__sun__)
#define IP_HDR_HTONS
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#if defined(__linux__)
#define __FAVOR_BSD
#define IP_HDR_HTONS
#endif

#if defined(__OpenBSD__) && OpenBSD >= 199706
#define IP_HDR_HTONS
#endif

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdio.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#include "scamper_addr.h"
#include "scamper_dl.h"
#include "scamper_probe.h"
#include "scamper_udp4.h"

#include "scamper_debug.h"
#include "utils.h"

static int udp4 = -1;

/*
 * these variables are used to store a packet buffer that is allocated
 * in the scamper_udp4_probe function large enough for the largest probe
 * the routine sends
 */
static uint8_t *pktbuf = NULL;
static size_t   pktbuf_len = 0;

static void udp_cksum(struct ip *ip, struct udphdr *udp, size_t len)
{
  uint16_t *w;
  uint8_t psuedo[12];
  int sum = 0;
  int i;

  /*
   * the UDP checksum includes a checksum calculated over a psuedo header
   * that includes the src and dst IP addresses, the protocol type, and
   * the UDP length.
   */
  memcpy(&psuedo[0],  &ip->ip_src, 4);
  memcpy(&psuedo[4],  &ip->ip_dst, 4);
  memcpy(&psuedo[10], &udp->uh_ulen, 2);
  psuedo[8] = 0;
  psuedo[9] = IPPROTO_UDP;

  /* compute the checksum over the psuedo header */
  w = (uint16_t *)psuedo;
  for(i=6; i>0; i--)
    {
      sum += *w++;
    }

  /* compute the checksum over the body of the UDP message */
  w = (uint16_t *)udp;
  while(len > 1)
    {
      sum += *w++;
      len -= 2;
    }

  if(len != 0)
    {
      sum += ((uint8_t *)w)[0];
    }

  /* fold the checksum */
  sum  = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  if((udp->uh_sum = ~sum) == 0)
    {
      udp->uh_sum = 0xffff;
    }

  return;
}

void scamper_udp4_build(scamper_probe_t *probe, uint8_t *pktbuf)
{
  struct ip     *ip;
  struct udphdr *udp;

  ip = (struct ip *)pktbuf;
  ip->ip_v   = 4;
  ip->ip_hl  = 5;
  ip->ip_tos = 0;
  ip->ip_len = htons(20 + 8 + probe->pr_len);
  ip->ip_id  = 0;
  ip->ip_off = htons(IP_DF);
  ip->ip_ttl = probe->pr_ttl;
  ip->ip_p   = IPPROTO_UDP;
  ip->ip_sum = 0;
  memcpy(&ip->ip_src, probe->pr_src->addr, sizeof(ip->ip_src));
  memcpy(&ip->ip_dst, probe->pr_dst->addr, sizeof(ip->ip_dst));
  ip->ip_sum = in_cksum(ip, sizeof(struct ip));

  udp = (struct udphdr *)(pktbuf + sizeof(struct ip));
  udp->uh_sport = htons(probe->pr_udp_sport);
  udp->uh_dport = htons(probe->pr_udp_dport);
  udp->uh_ulen  = htons(sizeof(struct udphdr) + probe->pr_len);
  udp->uh_sum   = 0;

  /* if there is data to include in the payload, copy it in now */
  if(probe->pr_len > 0)
    {
      memcpy(pktbuf + 20 + 8, probe->pr_data, probe->pr_len);
    }

  /* compute the checksum over the udp portion of the probe */
  udp_cksum(ip, udp, sizeof(struct udphdr) + probe->pr_len);

  return;
}

int scamper_udp4_probe(scamper_probe_t *probe)
{
  struct sockaddr_in  sin4;
  int                 i;
  char                addr[128];
  struct ip          *ip;
  struct udphdr      *udp;
  size_t              len;
  int                 hdrlen;
  void               *dst;
  uint16_t            dport;

  assert(probe != NULL);
  assert(probe->pr_proto == IPPROTO_UDP);
  assert(probe->pr_dst != NULL);
  assert(probe->pr_src != NULL);
  assert(probe->pr_len > 0 || probe->pr_data == NULL);

  /* take a copies of these variables, it makes this function easier */
  dport  = probe->pr_udp_dport;
  dst    = probe->pr_dst->addr;

  /* compute various lengths, for sake of readability */
  hdrlen = sizeof(struct ip) + sizeof(struct udphdr);
  len    = hdrlen + probe->pr_len;

  i = len;
  if(setsockopt(udp4, SOL_SOCKET, SO_SNDBUF, (char *)&i, sizeof(i)) == -1)
    {
      printerror(errno, strerror, __func__,
                 "could not set buffer to %d bytes", i);
      return -1;
    }

  if(pktbuf_len < len)
    {
      pktbuf = realloc_wrap(pktbuf, len, __func__, "could not realloc");
      pktbuf_len = len;
    }

  ip = (struct ip *)pktbuf;
  ip->ip_v   = 4;
  ip->ip_hl  = 5; /* (sizeof(struct ip) >> 2); */
  ip->ip_tos = 0;
  ip->ip_len = len;
  ip->ip_id  = 0;
  ip->ip_off = IP_DF;
  ip->ip_ttl = probe->pr_ttl;
  ip->ip_p   = IPPROTO_UDP;
  ip->ip_sum = 0;
  memcpy(&ip->ip_src, probe->pr_src->addr, sizeof(ip->ip_src));
  memcpy(&ip->ip_dst, dst, sizeof(ip->ip_dst));

#if defined(IP_HDR_HTONS)
  ip->ip_len = htons(ip->ip_len);
  ip->ip_off = htons(ip->ip_off);
#endif

  udp = (struct udphdr *)(pktbuf + sizeof(struct ip));
  udp->uh_sport = htons(probe->pr_udp_sport);
  udp->uh_dport = htons(dport);
  udp->uh_ulen  = htons(sizeof(struct udphdr) + probe->pr_len);
  udp->uh_sum   = 0;

  /* if there is data to include in the payload, copy it in now */
  if(probe->pr_len > 0)
    {
      memcpy(pktbuf + hdrlen, probe->pr_data, probe->pr_len);
    }

  /* compute the checksum over the udp portion of the probe */
  udp_cksum(ip, udp, sizeof(struct udphdr) + probe->pr_len);

  sockaddr_compose((struct sockaddr *)&sin4, AF_INET, dst, dport);

  /* get the transmit time immediately before we send the packet */
  gettimeofday_wrap(&probe->pr_tx);

  i = sendto(udp4, pktbuf, len, 0, (struct sockaddr *)&sin4,
	     sizeof(struct sockaddr_in));

  /* if we sent the probe successfully, there is nothing more to do here */
  if(i == len)
    {
      return 0;
    }

  probe->pr_errno = errno;

  /* error condition, could not send the packet at all */
  if(i == -1)
    {
      printerror(probe->pr_errno, strerror, __func__,
		 "could not send to %s (%d ttl, %d dport, %d len)",
		 inet_ntop(AF_INET, dst, addr, sizeof(addr)),
		 probe->pr_ttl, dport, len);
    }
  /* error condition, sent a portion of the probe */
  else
    {
      fprintf(stderr,
	      "scamper_udp4_probe: sent %d bytes of %d byte packet to %s",
	      i, (int)len,
	      inet_ntop(AF_INET, dst, addr, sizeof(addr)));
    }

  return -1;
}

int scamper_udp4_close()
{
  if(udp4 != -1)
    {
      if(close(udp4) == -1)
	{
	  printerror(errno, strerror, __func__, "close failed");
	  return -1;
	}

      udp4 = -1;
    }

  if(pktbuf != NULL)
    {
      free(pktbuf);
      pktbuf = NULL;
    }

  return 0;
}

int scamper_udp4_open(int udp_sport)
{
  struct sockaddr_in sin4;
  int                on;

  if((udp4 = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) == -1)
    {
      printerror(errno, strerror, __func__, "could not open socket");
      return -1;
    }

  on = 1;
  if(setsockopt(udp4, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on)) == -1)
    {
      printerror(errno, strerror, __func__, "could not IP_HDRINCL");
      return -1;
    }

  sockaddr_compose((struct sockaddr *)&sin4, AF_INET, NULL, udp_sport);

  if(bind(udp4, (struct sockaddr *)&sin4, sizeof(sin4)) == -1)
    {
      printerror(errno, strerror, __func__, "could not bind to port %d",
                 udp_sport);
      return -1;
    }

  return 0;
}
