/*
 * scamper_udp6.c
 *
 * $Id: scamper_udp6.c,v 1.31 2005/05/26 22:32:32 mjl Exp $
 *
 * Copyright (C) 2003-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#if defined(__linux__)
#define __FAVOR_BSD
#endif

#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdio.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#if defined(__linux__)
#if !defined(IPV6_MTU_DISCOVER)
#define IPV6_MTU_DISCOVER 23
#endif
#endif

#include "scamper_addr.h"
#include "scamper_dl.h"
#include "scamper_probe.h"
#include "scamper_udp6.h"

#include "scamper_debug.h"
#include "utils.h"

static int udp6 = -1;

static void udp_cksum(struct ip6_hdr *ip6, struct udphdr *udp, size_t len)
{
  uint16_t *w;
  uint8_t psuedo[40];
  int sum = 0;
  int i;

  /* build the psuedo header */
  memcpy(psuedo, &ip6->ip6_src, 16);
  memcpy(psuedo+16, &ip6->ip6_dst, 16);
  psuedo[32] = 0; psuedo[33] = 0;
  memcpy(psuedo+34, &udp->uh_ulen, 2);
  psuedo[36] = 0; psuedo[37] = 0; psuedo[38] = 0;
  psuedo[39] = IPPROTO_UDP;

  /* compute the checksum over the psuedo header */
  w = (uint16_t *)psuedo;
  for(i=20; i>0; i--)
    {
      sum += *w++;
    }

  /* compute the checksum over the body of the UDP message */
  w = (uint16_t *)udp;
  while(len > 1)
    {
      sum += *w++;
      len -= 2;
    }

  if(len != 0)
    {
      sum += ((uint8_t *)w)[0];
    }

  /* fold the checksum */
  sum  = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  if((udp->uh_sum = ~sum) == 0)
    {
      udp->uh_sum = 0xffff;
    }

  return;
}

void scamper_udp6_build(scamper_probe_t *probe, uint8_t *pktbuf)
{
  struct ip6_hdr *ip6;
  struct udphdr  *udp;

  ip6 = (struct ip6_hdr *)pktbuf;
  ip6->ip6_flow = 0;
  ip6->ip6_vfc  = 0x60;
  ip6->ip6_plen = htons(sizeof(struct udphdr) + probe->pr_len);
  ip6->ip6_nxt  = IPPROTO_UDP;
  ip6->ip6_hlim = probe->pr_ttl;
  memcpy(&ip6->ip6_src, probe->pr_src->addr, 16);
  memcpy(&ip6->ip6_dst, probe->pr_dst->addr, 16);

  udp = (struct udphdr *)(pktbuf + sizeof(struct ip6_hdr));
  udp->uh_sport = htons(probe->pr_udp_sport);
  udp->uh_dport = htons(probe->pr_udp_dport);
  udp->uh_ulen  = htons(sizeof(struct udphdr) + probe->pr_len);
  udp->uh_sum   = 0;

  /* if there is data to include in the payload, copy it in now */
  if(probe->pr_len > 0)
    {
      memcpy(pktbuf + 40 + 8, probe->pr_data, probe->pr_len);
    }

  /* compute the checksum over the udp portion of the probe */
  udp_cksum(ip6, udp, sizeof(struct udphdr) + probe->pr_len);

  return;
}

/*
 * scamper_udp6_probe:
 *
 * given the address, hop limit, destination UDP port number, and size, send
 * a UDP probe packet encapsulated in an IPv6 header.
 *
 * the size parameter is useful when doing path MTU discovery, and represents
 * how large the packet should be including IPv6 and UDP headers
 *
 * this function returns 0 on success, -1 otherwise
 */
int scamper_udp6_probe(scamper_probe_t *probe)
{
  struct sockaddr_in6  sin6;
  int                  i;
  char                 addr[128];
  size_t               len;
  int                  hdrlen;
  void                *dst;
  uint16_t             dport;

  assert(probe != NULL);
  assert(probe->pr_proto == IPPROTO_UDP);
  assert(probe->pr_dst != NULL);
  assert(probe->pr_src != NULL);
  assert(probe->pr_len > 0 || probe->pr_data == NULL);

  /* take copies of these variables, it makes this function easier */
  dport  = probe->pr_udp_dport;
  dst    = probe->pr_dst->addr;

  /* compute various lengths, for sake of readability */
  hdrlen = sizeof(struct ip6_hdr) + sizeof(struct udphdr);
  len    = hdrlen + probe->pr_len;

  i = probe->pr_ttl;
  if(setsockopt(udp6, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &i, sizeof(i)) == -1)
    {
      printerror(errno, strerror, __func__, "could not set hlim to %d", i);
      return -1;
    }

  i = probe->pr_len;
  if(setsockopt(udp6, SOL_SOCKET, SO_SNDBUF, (char *)&i, sizeof(i)) == -1)
    {
      printerror(errno, strerror, __func__,
                 "could not set buffer to %d bytes", i);
      return -1;
    }

  sockaddr_compose((struct sockaddr *)&sin6, AF_INET6, dst, dport);
  
  /* get the transmit time immediately before we send the packet */
  gettimeofday_wrap(&probe->pr_tx);

  i = sendto(udp6, probe->pr_data, probe->pr_len, 0, (struct sockaddr *)&sin6,
	     sizeof(struct sockaddr_in6));

  /* if we sent the probe successfully, there is nothing more to do here */
  if(i == probe->pr_len)
    {
      return 0;
    }

  /* get a copy of the errno variable as it is immediately after the sendto */
  probe->pr_errno = errno;

  inet_ntop(AF_INET6, dst, addr, sizeof(addr));

  /* error condition, could not send the packet at all */
  if(i == -1)
    {
      printerror(probe->pr_errno, strerror, __func__,
		 "could not send to %s (%d hlim, %d dport, %d len)",
		 addr, probe->pr_ttl, dport, probe->pr_len);
    }
  /* error condition, sent a portion of the probe */
  else
    {
      fprintf(stderr,
	      "scamper_udp6_probe: sent %d bytes of %d byte packet to %s",
	      i, (int)probe->pr_len, addr);
    }

  return -1;
}

int scamper_udp6_close()
{
  if(udp6 != -1)
    {
      if(close(udp6) == -1)
	{
	  printerror(errno, strerror, __func__, "close failed");
	  return -1;
	}

      udp6 = -1;
    }

  return 0;
}

int scamper_udp6_open(int udp_sport)
{
  struct sockaddr_in6 sin6;

#if defined(IPV6_DONTFRAG)
  int opt;
#endif

  if((udp6 = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP)) == -1)
    {
      printerror(errno, strerror, __func__, "could not open socket");
      return -1;
    }

  sockaddr_compose((struct sockaddr *)&sin6, AF_INET6, NULL, udp_sport);

  if(bind(udp6, (struct sockaddr *)&sin6, sizeof(sin6)) == -1)
    {
      printerror(errno, strerror, __func__, "could not bind to udp port %d",
                 udp_sport);
      return -1;
    }

#if defined(IPV6_DONTFRAG)
  opt = 1;
  if(setsockopt(udp6, IPPROTO_IPV6, IPV6_DONTFRAG, &opt, sizeof(opt)) == -1)
    {
      printerror(errno, strerror, __func__, "could not set IPV6_DONTFRAG");
      return -1;
    }
#endif

  return 0;
}
