/*
 * utils.c
 *
 * $Id: utils.c,v 1.45 2005/05/28 10:20:03 mjl Exp $
 *
 * Copyright (C) 2003-2005 The University of Waikato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#if defined(__APPLE__)
#define _BSD_SOCKLEN_T_
#define HAVE_SOCKADDR_SA_LEN
#include <stdint.h>
#endif

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define HAVE_SOCKADDR_SA_LEN
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/param.h>

#if !defined(__sun__)
#include <sys/sysctl.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>

#if defined(AF_LINK)
#include <net/if_dl.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>

#include <assert.h>

#if defined(DMALLOC)
#include <dmalloc.h>
#endif

#if defined(__sun__)
# define s6_addr32 _S6_un._S6_u32
#elif !defined(s6_addr32)
# define s6_addr32 __u6_addr.__u6_addr32
#endif

#include "utils.h"

#if defined(HAVE_SOCKADDR_SA_LEN)
int sockaddr_len(const struct sockaddr *sa)
{
  return sa->sa_len;
}
#else
int sockaddr_len(const struct sockaddr *sa)
{
  if(sa->sa_family == AF_INET)  return sizeof(struct sockaddr_in);
  if(sa->sa_family == AF_INET6) return sizeof(struct sockaddr_in6);

#if defined(AF_LINK)
  if(sa->sa_family == AF_LINK)  return sizeof(struct sockaddr_dl);
#endif

  return -1;
}
#endif

void *addr_dup(const int af, const void *addr_in)
{
  void   *addr;
  size_t  size;

  if(af == AF_INET)
    {
      size = sizeof(struct in_addr);
    }
  else if(af == AF_INET6)
    {
      size = sizeof(struct in6_addr);
    }
  else return NULL;

  if((addr = malloc(size)) != NULL)
    {
      memcpy(addr, addr_in, size);
    }

  return addr;
}

struct sockaddr *sockaddr_dup(const struct sockaddr *sa)
{
  struct sockaddr *addr;
  int len;

  if((len = sockaddr_len(sa)) == -1)
    {
      return NULL;
    }

  if((addr = malloc(len)) != NULL)
    {
      memcpy(addr, sa, len);
    }

  return addr;
}

void *sockaddr_addr(const struct sockaddr *sa)
{
  void *addr = NULL;

  assert(sa->sa_family == AF_INET || sa->sa_family == AF_INET6);

  if(sa->sa_family == AF_INET)
    {
      addr = &((struct sockaddr_in *)sa)->sin_addr;
    }
  else if(sa->sa_family == AF_INET6)
    {
      addr = &((struct sockaddr_in6 *)sa)->sin6_addr;
    }

  return addr;
}

void *sockaddr_addr_dup(const struct sockaddr *sa)
{
  return addr_dup(sa->sa_family, sockaddr_addr(sa));
}

int sockaddr_compose(struct sockaddr *sa, 
		     const int af, const void *addr, const int port)
{
  socklen_t len;
  struct sockaddr_in  *sin4;
  struct sockaddr_in6 *sin6;

  if(af == AF_INET)
    {
      len = sizeof(struct sockaddr_in);
      memset(sa, 0, len);
      sin4 = (struct sockaddr_in *)sa;
      if(addr != NULL) memcpy(&sin4->sin_addr, addr, sizeof(struct in_addr));
      sin4->sin_port = htons(port);
    }
  else if(af == AF_INET6)
    {
      len = sizeof(struct sockaddr_in6);
      memset(sa, 0, len);
      sin6 = (struct sockaddr_in6 *)sa;
      if(addr != NULL) memcpy(&sin6->sin6_addr, addr, sizeof(struct in6_addr));
      sin6->sin6_port = htons(port);
    }
  else return -1;

#if defined(HAVE_SOCKADDR_SA_LEN)
  sa->sa_len    = len;
#endif

  sa->sa_family = af;

  return 0;
}

char *sockaddr_tostr(const struct sockaddr *sa, char *buf, const size_t len)
{
  if(sa->sa_family == AF_INET)
    {
      inet_ntop(AF_INET, &((struct sockaddr_in *)sa)->sin_addr, buf, len);
    }
  else if(sa->sa_family == AF_INET6)
    {
      inet_ntop(AF_INET6, &((struct sockaddr_in6 *)sa)->sin6_addr, buf, len);
    }
  else
    {
      return NULL;
    }

  return buf;
}

int addr4_cmp(const void *va, const void *vb)
{
  struct in_addr *a, *b;

  a = (struct in_addr *)va;
  b = (struct in_addr *)vb;

  if(a->s_addr < b->s_addr) return -1;
  if(a->s_addr > b->s_addr) return  1;

  return 0;
}

int addr6_cmp(const void *va, const void *vb)
{
  struct in6_addr *a, *b;
  int i;

  a = (struct in6_addr *)va;
  b = (struct in6_addr *)vb;

  for(i=0; i<4; i++)
    {
      if(a->s6_addr32[i] < b->s6_addr32[i]) return -1;
      if(a->s6_addr32[i] > b->s6_addr32[i]) return  1;
    }

  return 0;
}

/*
 * addr_cmp:
 * this function is used to provide a sorting order, not for advising the
 * numerical order of the addresses passed in.
 */
int addr_cmp(int af, const void *a, const void *b)
{
  if(af == AF_INET)  return addr4_cmp(a, b);
  if(af == AF_INET6) return addr6_cmp(a, b);
  return 0;
}

/*
 * malloc_wrap
 *
 * allocate some memory, zero it, and return a pointer to it.
 * if we fail, the program exits with an appropriate error message.
 * i don't like exit()
 */
void *malloc_wrap(const size_t size, const char *format, ...)
{
  char     message[512];
  char    *error_str = NULL;
  va_list  ap;
  void    *ptr;

  ptr = malloc(size);
  if(ptr == NULL)
    {
      va_start(ap, format);
      vsnprintf(message, sizeof(message), format, ap);
      va_end(ap);

      error_str = strerror(errno);

      if(error_str != NULL) fprintf(stderr, "%s: %s\n", message, error_str);
      else                  fprintf(stderr, "%s\n", message);

      exit(-1);
    }

  memset(ptr, 0, size);

  return ptr;
}

/*
 * malloc_zero
 *
 * allocate some memory, zero it, and return a pointer to it.
 */
#if !defined(DMALLOC)
void *malloc_zero(const size_t size)
{
  void *ptr;

  if((ptr = malloc(size)) != NULL)
    {
      memset(ptr, 0, size);
    }

  return ptr;
}
#endif

void *realloc_wrap(void *data, const size_t size, const char *func,
		   const char *format, ...)
{
  char     message[512];
  char     err[128];
  char     fs[64];
  char    *str = NULL;
  va_list  ap;
  void    *ptr;

  if((ptr = realloc(data, size)) == NULL)
    {
      va_start(ap, format);
      vsnprintf(message, sizeof(message), format, ap);
      va_end(ap);

      if(func != NULL)
	{
	  snprintf(fs, sizeof(fs), "%s: ", func);
	}
      else
	{
	  fs[0] = '\0';
	}

      if((str = strerror(errno)) == NULL)
	{
	  err[0] = '\0';
	}
      else
	{
	  snprintf(err, sizeof(err), ": %s", str);
	}

      fprintf(stderr, "%s%s%s\n", fs, err, message);
      exit(-1);
    }

  return ptr;
}

int gettimeofday_wrap(struct timeval *tv)
{
  struct timezone tz;
  return gettimeofday(tv, &tz);
}

int timeval_cmp(const struct timeval *a, const struct timeval *b)
{
  if(a->tv_sec == b->tv_sec)
    {
      if(a->tv_usec == b->tv_usec)     return 0;
      else if(a->tv_usec < b->tv_usec) return -1;
      else                             return 1;
    }
  else if(a->tv_sec < b->tv_sec)
    {
      return -1;
    }
  else             
    {
      return 1;
    }
}

void timeval_add_msec(struct timeval *tv, const int msec)
{
  tv->tv_sec  += (msec / 1000);
  tv->tv_usec += ((msec % 1000) * 1000);

  /* check for overflow of usec's */
  if(tv->tv_usec >= 1000000)
    {
      tv->tv_sec++;
      tv->tv_usec -= 1000000;
    }
  else if(tv->tv_usec < 0)
    {
      tv->tv_sec--;
      tv->tv_usec += 1000000;
    }

  return;
}

/*
 * timeval_diff_msec
 * return the millisecond difference between the two timevals.
 * a - b
 */
int64_t timeval_diff_msec(const struct timeval *a, const struct timeval *b)
{
  int64_t temp, a_sec,  a_usec, b_sec, b_usec;

  a_sec  = (int64_t)a->tv_sec  * (int64_t)1000;
  a_usec = (int64_t)a->tv_usec / (int64_t)1000;
  b_sec  = (int64_t)b->tv_sec  * (int64_t)1000;
  b_usec = (int64_t)b->tv_usec / (int64_t)1000;

  temp = a_sec - b_sec + a_usec - b_usec;
  return temp;
}

void timeval_add_usec(struct timeval *tv, const int64_t usec)
{
  tv->tv_sec  += (usec / 1000000);
  tv->tv_usec += (usec % 1000000);

  /* check for underflow */
  if(tv->tv_usec < 0)
    {
      tv->tv_sec--;
      tv->tv_usec += 1000000;
    }
  /* check for overflow */
  else if(tv->tv_usec >= 1000000)
    {
      tv->tv_sec++;
      tv->tv_usec -= 1000000;
    }

  return;
}

/*
 * timeval_add_tv
 *
 * given a timeval struct, add the time in the second timeval struct to
 * it.
 */
void timeval_add_tv(struct timeval *tv, const struct timeval *add)
{
  assert(add->tv_sec >= 0);
  assert(add->tv_usec >= 0);

  tv->tv_usec += add->tv_usec;

  /* check for overflow */
  if(tv->tv_usec > 1000000)
    {
      tv->tv_sec++;
      tv->tv_usec -= 1000000;
    }

  tv->tv_sec += add->tv_sec;

  return;
}

/*
 * timeval_diff_usec
 * return the microsecond difference between the two timevals.
 * a - b
 */
int64_t timeval_diff_usec(const struct timeval *a, const struct timeval *b)
{
  int64_t temp, a_sec, b_sec;

  a_sec  = (int64_t)a->tv_sec  * (int64_t)1000000;
  b_sec  = (int64_t)b->tv_sec  * (int64_t)1000000;

  temp = a_sec - b_sec + a->tv_usec - b->tv_usec;
  return temp;
}

void timeval_rtt(struct timeval *rtt,
		 const struct timeval *from, const struct timeval *to)
		 
{
  rtt->tv_sec  = to->tv_sec  - from->tv_sec;
  rtt->tv_usec = to->tv_usec - from->tv_usec;

  if(rtt->tv_usec < 0)
    {
      rtt->tv_sec--;
      rtt->tv_usec += 1000000;
    }

  return;
}

void timeval_cpy(struct timeval *dst, const struct timeval *src)
{
  memcpy(dst, src, sizeof(struct timeval));
  return;
}

int fcntl_unset(const int fd, const int flags)
{
  int i;

  if((i = fcntl(fd, F_GETFL, 0)) == -1)
    {
      return -1;
    }

  if(fcntl(fd, F_SETFL, i & (~flags)) == -1)
    {
      return -1;
    }

  return 0;
}

int fcntl_set(const int fd, const int flags)
{
  int i;

  if((i = fcntl(fd, F_GETFL, 0)) == -1)
    {
      return -1;
    }

  if(fcntl(fd, F_SETFL, i | flags) == -1)
    {
      return -1;
    }

  return 0;
}

int string_isprint(const char *str, const int len)
{
  int i = 0;

  for(i=0; i<len; i++)
    {
      if(isprint((int)str[i]) != 0)
	{
	  continue;
	}
      else if(str[i] == '\0')
	{
	  break;
	}
      else return 0;
    }

  return 1;
}

int string_tolong(const char *str, long *l)
{
  char *endptr;

  *l = strtol(str, &endptr, 0);
  if(*l == 0)
    {
      if(errno == EINVAL) return -1;
    }
  else if(*l == LONG_MIN || *l == LONG_MAX)
    {
      if(errno == ERANGE) return -1;
    }

  return 0;
}

/*
 * string_isnumber
 *
 * scan the word to establish if it an integer.
 */
int string_isnumber(const char *str)
{
  int i = 1;

  if(str[0] != '-' && str[0] != '+' && isdigit((int)str[0]) == 0)
    {
      return 0;
    }

  while(str[i] != '\0')
    {
      if(isdigit((int)str[i]) != 0)
	{
	  i++;
	  continue;
	}

      return 0;
    }

  return 1;
}

/*
 * string_isfloat
 *
 * scan the word to establish if it is a float.
 */
int string_isfloat(const char *str)
{
  int seen_dp = 0;
  int i = 1;

  if(str[0] != '-' && str[0] != '+' && isdigit((int)str[0]) == 0)
    {
      if(str[0] == '.')
	{
	  seen_dp = 1;
	}
      else return 0;
    }

  while(str[i] != '\0')
    {
      if(isdigit((int)str[i]) != 0)
	{
	  i++;
	  continue;
	}
      else if(str[i] == '.')
	{
	  /* if the decimal point has already been seen */
	  if(seen_dp == 1)
	    {
	      return 0;
	    }

	  i++;
	  seen_dp = 1;
	  continue;
	}
      return 0;
    }

  return 1;
}

/*
 * string_nextword
 *
 * scan for the next word occurance in the string, after the current
 * word.  if there is no other word after this one, return NULL.
 */
char *string_nextword(char *buf)
{
  /* scan for a start of a word */
  while(*buf != '\0' && isspace((int)*buf) == 0)
    {
      buf++;
    }

  if(*buf == '\0')
    {
      return NULL;
    }
  *buf = '\0';
  buf++;

  /* now scan for the end of the word */
  while(*buf != '\0' && isspace((int)*buf) != 0)
    {
      buf++;
    }

  if(*buf == '\0')
    {
      return NULL;
    }

  return buf;
}

/*
 * string_nullterm
 *
 * null terminate the string pointed to by buf at the first occurance of
 * a character in the delim string.
 */
char *string_nullterm(char *buf, const char *delim)
{
  const char *dtmp;
  char *tmp;

  if(delim == NULL || *delim == '\0' || (tmp = buf) == NULL)
    {
      return NULL;
    }

  while(*tmp != '\0')
    {
      dtmp = delim;

      while(*dtmp != '\0')
	{
	  if(*tmp != *dtmp++)
	    {
	      continue;
	    }

	  *tmp = '\0';
	  return buf;
	}

      tmp++;
    }

  return NULL;
}

int read_wrap(const int fd, void *ptr, size_t *rc_out, const size_t rt)
{
  uint8_t *buf;
  int      ret = 0;
  ssize_t  r;
  size_t   rc;

  assert(rt > 0);
  assert(ptr != NULL);

  buf = (uint8_t *)ptr;

  for(rc = 0; rc < rt; rc += r)
    {
      if((r = read(fd, buf+rc, rt-rc)) == -1)
	{
	  if(errno == EINTR)
	    {
	      r = 0;
	      continue;
	    }
	  ret = -1;
	  break;
	}
      else if(r == 0)
	{
	  ret = -2;
	  break;
	}
    }

  if(rc_out != NULL)
    {
      *rc_out = rc;
    }

  return ret;
}

int write_wrap(const int fd, const void *ptr, size_t *wc_out, const size_t wt)
{
  uint8_t *buf;
  int      ret = 0;
  ssize_t  w;
  size_t   wc;

  assert(wt > 0);
  assert(ptr != NULL);

  buf = (uint8_t *)ptr;

  for(wc = 0; wc < wt; wc += w)
    {
      if((w = write(fd, buf+wc, wt-wc)) == -1)
	{
	  if(errno == EINTR)
	    {
	      w = 0;
	      continue;
	    }
	  ret = -1;
	  break;
	}
    }

  if(wc_out != NULL)
    {
      *wc_out = wc;
    }

  return ret;
}

#if !defined(__sun__)
int sysctl_wrap(int *mib, int len, void **buf, size_t *size)
{
  if(sysctl(mib, len, NULL, size, NULL, 0) == -1)
    {
      return -1;
    }

  if(*size == 0)
    {
      *buf = NULL;
      return 0;
    }

  if((*buf = malloc(*size)) == NULL)
    {
      return -1;
    }

  if(sysctl(mib, len, *buf, size, NULL, 0) == -1)
    {
      free(*buf);
      return -1;
    }

  return 0;
}
#endif

uint16_t in_cksum(const void *buf, const size_t len)
{
  uint16_t *w = (uint16_t *)buf;
  size_t l = len;
  int sum = 0;

  while(l > 1)
    {
      sum += *w++;
      l   -= 2;
    }

  if(l != 0)
    {
      sum += ((uint8_t *)w)[0];
    }

  sum  = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return ~sum;
}
