#if defined(__APPLE__)
#include <stdint.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

#include "scamper_addr.h"
#include "scamper_list.h"
#include "scamper_tlv.h"
#include "scamper_trace.h"
#include "scamper_file.h"

static void usage()
{
  fprintf(stderr, "usage: warts-dump file <lists|traces>");
  return;
}

static void hop_dump(scamper_trace_hop_t *hop)
{
  scamper_tlv_t *tlv;
  char addr[256];

  scamper_addr_tostr(hop->addr, addr, sizeof(addr));

  fprintf(stdout, "probe_ttl %d addr %s probe_id %d probe_size %d "
	  "reply_ttl %d icmp_type %d icmp_code %d rtt %d.%06d flags 0x%02x",
	  hop->probe_ttl, addr, hop->probe_id, hop->probe_size,
	  hop->reply_ttl,
	  hop->icmp_type, hop->icmp_code,
	  (int)hop->rtt.tv_sec, (int)hop->rtt.tv_usec, hop->flags);

  for(tlv = hop->tlvs; tlv != NULL; tlv = tlv->tlv_next)
    {
      switch(tlv->tlv_type)
	{
	case SCAMPER_TRACE_HOP_TLV_REPLY_IPID:
	  fprintf(stdout, " ipid outer 0x%04x", tlv->tlv_val_16);
	  break;

	case SCAMPER_TRACE_HOP_TLV_REPLY_IPTOS:
	  fprintf(stdout, " iptos outer 0x%02x", tlv->tlv_val_8);
	  break;

	case SCAMPER_TRACE_HOP_TLV_NHMTU:
	  fprintf(stdout, " nhmtu %d", tlv->tlv_val_16);
	  break;

	case SCAMPER_TRACE_HOP_TLV_INNER_IPTTL:
	  fprintf(stdout, " turn ttl %d", tlv->tlv_val_8);
	  break;

	case SCAMPER_TRACE_HOP_TLV_INNER_IPLEN:
	  fprintf(stdout, " iplen inner %d", tlv->tlv_val_16);
	  break;
	}
    }

  fprintf(stdout, "\n");

  return;
}

static void dump_traces(scamper_file_t *file)
{
  scamper_trace_t     *trace;
  scamper_trace_hop_t *hop;
  char                 buf[256];
  int                  i;

  while((trace = scamper_file_read_trace(file)) != NULL)
    {
      scamper_addr_tostr(trace->src, buf, sizeof(buf));
      fprintf(stdout, "%s to ", buf);
      scamper_addr_tostr(trace->dst, buf, sizeof(buf));
      fprintf(stdout, "%s ", buf);

      fprintf(stdout,
	      "start %ld attempts %d hlim %d sport %d dport %d flags %d\n",
	      trace->start.tv_sec,
	      trace->attempts, trace->hoplimit, trace->sport, trace->dport,
	      trace->flags);

      for(i=0; i<trace->hop_count; i++)
	{
	  for(hop = trace->hops[i]; hop != NULL; hop = hop->next)
	    {
	      hop_dump(hop);
	    }
	}

      if(trace->pmtud != NULL)
	{
	  printf("pmtud: ifmtu %d pmtu %d\n",
		 trace->pmtud->ifmtu, trace->pmtud->pmtu);
	  for(hop = trace->pmtud->hops; hop != NULL; hop = hop->next)
	    {
	      hop_dump(hop);
	    }
	}

      scamper_trace_free(trace);

      fprintf(stdout, "\n");
    }

  return;
}

static void dump_lists(scamper_file_t *file)
{
  scamper_list_t *list;

  while((list = scamper_file_read_list(file)) != NULL)
    {
      printf("list: %12s", list->name);
      if(list->descr != NULL) printf(" %s", list->descr);
      printf("\n");
      scamper_list_free(list);
    }
  return;
}

int main(int argc, char *argv[])
{
  scamper_file_t  *file;

  if(argc != 3)
    {
      usage();
      return -1;
    }

  if((file = scamper_file_open(argv[1], 'r', NULL)) == NULL)
    {
      usage();
      fprintf(stderr, "could not open %s\n", argv[1]);
      return -1;
    }

  if(strcmp(argv[2], "lists") == 0)
    {
      dump_lists(file);
    }
  else if(strcmp(argv[2], "traces") == 0)
    {
      dump_traces(file);
    }
  else
    {
      usage();
    }

  scamper_file_close(file);
  return 0;
}
